#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#ifndef MCP_EXTERNAL
#define MCP_EXTERNAL
#endif

#include "ali.h"

#define FRAC_POP
#define HPE_CNST  2.0e-26
#define CH2_CNST  2.34e-21
#define CH2_EXP   4.54

double Cul[MAX_VOLUME][MAX_LEVEL][MAX_LEVEL];

int gaussj();
double **matrix(), *vector();
#ifdef COOLING
double VelRad(double);
void col_init(), tra_init();
#endif

void NgAccel(VOLUME *v, int order)
{
    static int first=1;
    int i, j, l, ierr;
    double w, d0, di, dj, new , s;
    static double **A, *B;

    if (first) {
        A = matrix(order, order);
        B = vector(order);
        first = 0;
    }

    for (i=0; i<order; i++) {
        for (j=i; j<order; j++) {
            A[i][j] = 0.0;
            for (l=0; l<mcp.Nlevel; l++) {
                w = 1.0;
                d0 = v->pop[l] - v->p[0][l];
                di = v->p[i][l] - v->p[i+1][l];
                dj = v->p[j][l] - v->p[j+1][l];
                A[i][j] += w * (d0 - di) * (d0 - dj);
            }
            if (j > i) A[j][i] = A[i][j];
        }
        B[i] = 0.0;
        for (l=0; l<mcp.Nlevel; l++) {
            w = 1.0;
            d0 = v->pop[l] - v->p[0][l];
            di = v->p[i][l] - v->p[i+1][l];
            B[i] += w * d0 * (d0 - di);
        }
    }

    ierr = gaussj(A, order, B);
    if (ierr) {
        fprintf(stderr, "Singular Ng-matrix (ierr=%d).\n", ierr);
        return;
    }
    
    for (l=0; l<mcp.Nlevel; l++) {
        s = 1.0;
        for (i=0; i<order; i++) s -= B[i];
        new = s * v->pop[l];
        for (i=0; i<order; i++) new += v->p[i][l] * B[i];
        /* if (v == &vol[1])
        printf("p[%2d]= %10.3e,%10.3e,%10.3e,%10.3e -> %10.3e\n", l,
               v->p[2][l]-new, v->p[1][l]-new, v->p[0][l]-new, v->pop[l]-new,
	       new); */
        v->pop[l] = new;
    }
}

#ifdef COOLING
double DustGasHeating(VOLUME *v, double r)
{
    double vd, x, vrad, vtherm, Hdg;
    /* IRAS 15194-5115 data, Ryde et al. 1999 */
    /* double Q = 3.0e-2;
    double Mdot = 1.0e-5;
    double d = 600.0, d0 = 600.0;
    double L, L0 = 8.77e3;
    double rho = 2.00, a = 0.05; */
    /* W Hya guess */
    /* double Q = 3.0e-2;
    double Mdot = 7.5e-8;
    double d = 63.0, d0 = 63.0;
    double L, L0 = 4.0e3;
    double rho = 2.00, a = 0.05; */
    /* IRC+10216 data, Crosas & Menten 1997 */
    double Q, Mdot = 3.25e-5;
    double d = 150.0, d0 = 200.0;
    double L, L0 = 2.1e4;
    double rho = 1.85, a = 0.05;
    
    if (r*gSph.Rmax < 1.0e16)
        Q = 2.5e-2;
    else
        Q = 1.8e-2;
    
    L = L0 * d * d/ d0 /d0;
    
    vrad =  1.439e-4 * sqrt(L * VelRad(r) * Q / Mdot);
    vtherm = 0.083454 * sqrt(v->Tkin);
    
    x = 0.5 * vtherm * vtherm / vrad / vrad;
    vd = vrad * sqrt(sqrt(1.0 + x*x) - x);
    
    Hdg = 4.19664e-29 * v->nH2 * v->nH2 * vd*vd*vd / v->Rgd / rho / a;
    
    return Hdg;
}

double DeltaTHeating(VOLUME *v, double r)
{
   double alpha_c, Hdt;
   
   alpha_c = 0.35 * exp(-sqrt((v->Tdust + v->Tkin)/500.0)) + 0.1;
   
   Hdt = 2.171e-30 * v->nH2 * v->nH2 / v->Rgd *
         sqrt(v->Tkin) * (v->Tdust - v->Tkin) * alpha_c;
   
   return Hdt;
}
#endif

void adjust()
{
  static int first=1, Ng=0;
  int n, m, u, l, ierr, uu, ll, olap, o, mo, oo;
  int l_max=0, n_max=1;
  double Rul, Rlu, pup, plow, Tex[MAX_RTRANS], Tau[MAX_RTRANS];
  double Snu, Jeff, kll, kmm;
  static double **a, *b;
  double T[MAX_VOLUME], dTau[MAX_VOLUME], diff=0.0, d, d_max=0.0;
#ifdef FRAC_POP
  double f[MAX_LEVEL];
#endif
#ifdef COOLING
  int i, j, Cn_max = 0;
  double adia, h, dT, r0, r1, r2, beta, Cd_max=0.0;
  /* double Cool[MAX_VOLUME][MAX_RTRANS], CCool[MAX_VOLUME][MAX_RTRANS]; */
  double Temp[MAX_VOLUME], DeltaT[MAX_VOLUME];
  double fTemp[MAX_VOLUME], Deltaf[MAX_VOLUME];
#endif
  VOLUME *v = NULL;
  RTRANS *t, *tt;
  
  void d_fprintf();
  void SavePopulations(char *);
#ifdef COOLING
  void SaveTkin(char *);
#endif
  
  if (first) {
    a = matrix(mcp.Nlevel, mcp.Nlevel);
    b = vector(mcp.Nlevel);
    first = 0;
  }
  
  if (mcp.Iter == 1 && mcp.PopFrac != 1.0)
      printf("Always calc. pop as %.2f p_new + %.2f p_old.\n",
             mcp.PopFrac, 1.0 - mcp.PopFrac);
  
  if (mcp.Iter == mcp.Nlambda)
      printf("Activating local Lambda acceleration.\n");
  
  for (n=1; n<mcp.Nvol; n++) {
    v = &vol[n];
    for (u=0; u<mcp.Nlevel; u++) {
      for (l=0; l<mcp.Nlevel; l++) {
        a[l][u] = Cul[n][l][u];
      }
    }
#ifdef COOLING
    v->CradPrev = v->Crad;
    v->Crad = 0.0;
    v->CradC = 0.0;
    v->CradCC = 0.0;
#endif
    for (m=0; m<mcp.NrTrans; m++) {
      t = &rTra[m];
      l = t->l; u = t->u;
      olap = 0;
      if (mcp.Iter >= mcp.Nlambda && mcp.Nlambda > 0) {
          if (mcp.overlap && t->n > 0) olap = 1;
          Snu = 1.0/(v->pop[l]/v->pop[u]*t->Gul - 1.0);
          Jeff = v->Jbar[m] - v->Lbar[m] * Snu;
          if (olap) {
              kll = t->aconst * (v->pop[l]*t->Gul - v->pop[u]);
              for (o=0; o<t->n; o++) {
                  tt = &rTra[t->m[o]];
                  ll = tt->l; uu = tt->u;
                  Snu = 1.0/(v->pop[ll]/v->pop[uu]*tt->Gul - 1.0);
                  kmm = tt->aconst * (v->pop[ll]*tt->Gul - v->pop[uu]);
                  Jeff -= v->LbarOver[m][o] * Snu * kmm / kll;
              }
          }
          Rul = t->Aul * (1.0 + Jeff - v->Lbar[m]);
          Rlu = t->Aul * t->Gul * Jeff;
      } else {
          Rul = t->Aul * (1.0 + v->Jbar[m]);
          Rlu = t->Aul * t->Gul * v->Jbar[m];
      }
      a[l][l] -= Rlu;
      a[l][u] += Rul;
      /* if (u < mcp.Nlevel-1) { */
          a[u][u] -= Rul;
          a[u][l] += Rlu;
      /* } */
      if (olap) {
          /* if (u == 5 && l == 1) {
              printf("n=%2d m=%3d  %6.3f", n, m, v->Lbar[m]);
          } */
          for (o=0; o<t->n; o++) {
              mo = t->m[o];
              tt = &rTra[mo];
              for (oo=0; oo<tt->n; oo++) {
                  if (tt->m[oo] != m) continue;
                  Rul = t->Aul * v->LbarOver[mo][oo];
                  a[l][u] += -Rul;
                  /* if (u < mcp.Nlevel-1) */  a[u][u] -= -Rul;
                  /* if (u == 5 && l == 1) {
                      printf("%3d:%6.3f,%6.3f  ", mo, v->Lbar[mo], v->LbarOver[mo][oo]);
                  } */
                  break;
              }
          }
          /* if (u == 5 && l == 1) {
              printf("\n");
          } */
      }
#ifdef COOLING
      v->CradC += PLANCK_CNST * t->Fre * v->pop[u] * t->Aul * v->Beta[m];
#endif
    }
    
    for (l=0; l<mcp.Nlevel; l++) a[mcp.Nlevel-1][l] = 1.0;
    for (l=0; l<mcp.Nlevel-1; l++) b[l] = 0.0;
    b[mcp.Nlevel-1] = v->Xmol * v->nH2;
    
    ierr = gaussj(a, mcp.Nlevel, b);
    if (ierr) {
      fprintf(stderr, "Singular a-matrix (ierr=%d) in volume %d.\n",
              ierr, n);
      exit(1);
    }
    
 /* Save the old populations */
    for (l=0; l<mcp.Nlevel; l++) {
        m = MAX_NG;
        while (m > 0) {
           v->p[m][l] = v->p[m-1][l];
           m--;
        }
        v->p[0][l] = v->pop[l];
	v->pop[l] = mcp.PopFrac * b[l] + (1.0-mcp.PopFrac) * v->p[0][l];
#ifdef FRAC_POP
        f[l] = v->pop[l] / v->Xmol / v->nH2;
#endif
    }
    
#ifdef COOLING
    for (m=0; m<mcp.NrTrans; m++) {
        t = &rTra[m];
        u = t->u;
        d = 0.0;
        for (i=0; i<u; i++) {
            for (j=u; j<mcp.Nlevel; j++) {
                d += vol[n].pop[i] * Cul[n][j][i] -
                               vol[n].pop[j] * Cul[n][i][j];
            }
        }
        v->CradCC += d * PLANCK_CNST * t->Fre;
    }
    for (l=0; l<mcp.Nlevel-1; l++) {
      for (u=l+1; u<mcp.Nlevel; u++) {
          d = (lev[u].Egy - lev[l].Egy) * BOLTZ_CNST;
          v->Crad += (v->pop[l] * Cul[n][u][l] -
                      v->pop[u] * Cul[n][l][u]) * d ;
      }
    }
    /* Approximately include cooling due to 13CO by increasing the
       12CO cooling rates by 20 % */
    /* IRAS 15194-5115 data, Ryde et al. 1999 X[12CO]/X[13CO] = 5.5 */
    /* v->Crad *= 1.2;
    v->CradC *= 1.2;
    v->CradCC *= 1.2; */
#endif
  
    if (mcp.Iter >= mcp.Nng && mcp.Iter > mcp.NgOrder && mcp.Nng > 0 && !Ng) {
        if (n == 1) printf("Using Ng acceleration of order %d.\n",
                           mcp.NgOrder);
        NgAccel(v, mcp.NgOrder);
        Ng = mcp.Iter;
    } else if (Ng && ((mcp.Iter - Ng) % mcp.Nng1) == 0) {
        if (n == 1) printf("Using Ng acceleration of order %d.\n",
                           mcp.NgOrder);
        NgAccel(v, mcp.NgOrder);
    }

 /* Calculate the changes from the previous iteration */
    l = 0;
    while (l < mcp.LimitLevel && lev[l].Egy < 3.0 * v->Tkin) {
      d = fabs(v->pop[l]-v->p[0][l])/fabs(v->p[0][l]);
      if (d > d_max) {
        d_max = d;
        l_max = l;
        n_max = n;
      }
      diff += d;
      l++;
    }
    
#ifdef COOLING
    if (n == 1 && gSph.Rmin == 0.0) {
        r0 = 0.0;
    } else {
        r0 = (gSph.Rad[n] + gSph.Rad[n-1])/2.0;
    }
    beta = pow(r0*gSph.Rmax/1.0e14, 4.0);
    h = 2.0/3.0 / v->nH2 / VelRad(r0) * gSph.Rmax / BOLTZ_CNST / 1.0e5;
    if (v->CradPrev != 0.0) {
        d = fabs(v->Crad - v->CradPrev)/fabs(v->CradPrev);
        if (d > Cd_max) {
          Cd_max = d;
          Cn_max = n;
        }
    }
    adia = 4.0/3.0 * v->Tkin / r0 ;
    v->Hdg = DustGasHeating(v, r0);
    v->Hdt = DeltaTHeating(v, r0);
    v->Hpe = HPE_CNST * v->nH2;
    v->CH2 = CH2_CNST * v->nH2 * exp(CH2_EXP * log(v->Tkin/1000.0));
    dT = -adia + (v->Hdg + v->Hpe - v->Crad - v->CH2)*h;
    DeltaT[n] = dT;
    Deltaf[n] = r0*dT + v->Tkin;
    dT *= (gSph.Rad[n] - gSph.Rad[n-1]);
    printf(
"%3d %9.2e C=%9.2e(%8.4f,%8.4f) H2=%9.2e Hdg=%9.2e Hdt=%9.2e Hpe=%9.2e A=%9.3e T=%9.3f dT=%9.2e\n",
           n, r0*gSph.Rmax,
           v->Crad /* * beta */, v->CradC/v->Crad, v->CradCC/v->Crad,
	   v->CH2 * beta, v->Hdg * beta, v->Hdt * beta, v->Hpe * beta,
	   adia/h * beta, v->Tkin, Deltaf[n] * (gSph.Rad[n] - gSph.Rad[n-1]));
#endif
  
    for (m=0; m<mcp.NrTrans; m++) {
      l = rTra[m].l; u = rTra[m].u;
      pup = v->pop[u]; plow = v->pop[l];
      Tex[m] = 0.0;
      Tau[m] = 0.0;
      if (pup > 0.0 && plow > 0.0) {
	  Tex[m] = -PLOBO_CNST * rTra[m].Fre / log(pup/plow/rTra[m].Gul);
	  Tau[m] = rTra[m].aconst * gSph.Rmax * (plow*rTra[m].Gul - pup)/v->Vdop *
	           (gSph.Rad[n] - gSph.Rad[n-1]);
      }
    }
#ifdef FRAC_POP
    d_fprintf(logs.fp, " %10.3e", f, mcp.Nlevel);
#else
    d_fprintf(logs.fp, " %10.3e", b, mcp.Nlevel);
#endif
    d_fprintf(logs.fp, " %10.3e", Tex, mcp.NrTrans);
    if (mcp.TrInfo >= 0 && mcp.TrInfo < mcp.NrTrans) {
        T[n] = Tex[mcp.TrInfo];
	dTau[n] = Tau[mcp.TrInfo];
    }
  }
  
  diff /= (double)(mcp.Nlevel*(mcp.Nvol - 1));
#ifdef COOLING
  printf("I= %3d, ave. %9.3e (max %9.3e at lev=%d in vol=%d, %9.3e %d)\n",
         mcp.Iter, diff, d_max, l_max, n_max, Cd_max, Cn_max);
  fprintf(logs.fp,
         "I= %3d, ave. %9.3e (max %9.3e at lev=%d in vol=%d, %9.3e %d)\n",
         mcp.Iter, diff, d_max, l_max, n_max, Cd_max, Cn_max);
#else
  printf("I= %3d, average rel. pop. change %9.3e (max %9.3e at lev=%d in vol=%d)\n",
         mcp.Iter, diff, d_max, l_max, n_max);
  fprintf(logs.fp, "I= %3d, average rel. pop. change %9.3e (max %9.3e at lev=%d in vol=%d)\n",
         mcp.Iter, diff, d_max, l_max, n_max);
#endif
         
  mcp.maxrel = d_max;
  
#ifdef COOLING
/* Adjust the kinetic temperatures only if rel. pop. change is small */
  if (d_max < 0.001 && Cd_max < 0.001) {
    for (i=0; i<40; i++) {
      for (n=1; n<mcp.Nvol; n++) {
          v = &vol[n];
          if (n == 1 && gSph.Rmin == 0.0) {
              r0 = 0.0;
          } else {
              r0 = (gSph.Rad[n] + gSph.Rad[n-1])/2.0;
          }
          beta = pow(r0*gSph.Rmax/1.0e14, 4.0);
          h = 2.0/3.0 / v->nH2 / VelRad(r0) * gSph.Rmax / BOLTZ_CNST / 1.0e5;
          adia = 4.0/3.0 * v->Tkin / r0 ;
          v->Hdg = DustGasHeating(v, r0);
          v->Hdt = DeltaTHeating(v, r0);
          v->Hpe = HPE_CNST * v->nH2;
          v->CH2 = CH2_CNST * v->nH2 * exp(CH2_EXP * log(v->Tkin/1000.0));
          DeltaT[n] = -adia + (v->Hdg + v->Hpe - v->Crad - v->CH2)*h;
          Deltaf[n] = r0*DeltaT[n] + v->Tkin;
      }
      for (n=2; n<mcp.Nvol; n++) {
          v = &vol[n-1];
          r1 = (gSph.Rad[n-1] + gSph.Rad[n-2])/2.0;
          r2 = (gSph.Rad[n] + gSph.Rad[n-1])/2.0;
          h = r2 - r1;
          Temp[n] = v->Tkin + h*(DeltaT[n-1] + DeltaT[n])/2.0;
          fTemp[n] = r1 * v->Tkin + h*(Deltaf[n-1] + Deltaf[n])/2.0;
      }
      if (mcp.Iter <= mcp.Cooling) {
          r0 = (gSph.Rad[1] + gSph.Rad[0])/2.0;
          r1 = (gSph.Rad[2] + gSph.Rad[1])/2.0;
          r2 = (gSph.Rad[3] + gSph.Rad[2])/2.0;
          beta = log(Temp[3]/Temp[2])/log(r2/r1);
          Temp[1] = Temp[2] * pow(r0/r1, beta); 
          beta = log(fTemp[3]/fTemp[2])/log(r2/r1);
          fTemp[1] = fTemp[2] * pow(r0/r1, beta); 
          for (n=1; n<mcp.Nvol; n++) {
              if (n == 1 && gSph.Rmin == 0.0) {
                  r0 = 0.0;
              } else {
                  r0 = (gSph.Rad[n] + gSph.Rad[n-1])/2.0;
              }
              v = &vol[n];
              beta = fabs(Temp[n] - v->Tkin)/v->Tkin;
              beta = fabs(fTemp[n] - r0*v->Tkin)/r0/v->Tkin;
              if (beta < 0.25) {
                  v->Crad /= (v->Tkin * v->Tkin);
                  if (beta > 0.01) {
                      v->Tkin = (v->Tkin + fTemp[n]/r0)/2.0;
                  } else {
                      v->Tkin = fTemp[n]/r0;
                  }
                  v->Crad *= (v->Tkin * v->Tkin);
              } else if (fTemp[n]/r0 > v->Tkin) {
                  v->Crad *= 1.1*1.1;
                  v->Tkin *= 1.1;
              } else if (0.9 * v->Tkin > 3.0) {
                  v->Crad *= 0.9*0.9;
                  v->Tkin *= 0.9;
              }
          }
      }
    }
    if (mcp.Iter <= mcp.Cooling) {
        printf("New kinetic temps. calculated.\n");
        col_init();
        SaveTkin(".tkin");
    }
  } 
#endif
  
  SavePopulations(mcp.popFile);
  
  logs.rad = fopen(logs.radfile, "w");
  if (mcp.TrInfo >= 0 && mcp.TrInfo < mcp.NrTrans) {
    for (n=1; n<mcp.Nvol; n++) {
      if (n == 1 && gSph.Rmin == 0.0) {
        d = 0.0;
      } else {
        d = (gSph.Rad[n] + gSph.Rad[n-1])/2.0;
      }
      fprintf(logs.fp, "r= %10.4e Tex[%d]= %13.6e %10.5f Tau[%d]= %13.6e\n",
              d, mcp.TrInfo, T[n], T[n]/vol[n].Tkin, mcp.TrInfo, dTau[n]);
      fprintf(logs.rad, "%3d r= %10.4e Tex[%d]= %13.6e %13.6e Tau[%d]= %13.6e\n",
              n, d, mcp.TrInfo, T[n], T[n]/vol[n].Tkin, mcp.TrInfo, dTau[n]);
    }
  } else if (mcp.TrInfo < 0 && -mcp.TrInfo < mcp.NrTrans) {
    for (n=1; n<mcp.Nvol; n++) {
      v = &vol[n];
      if (n == 1 && gSph.Rmin == 0.0) {
        d = 0.0;
      } else {
        d = (gSph.Rad[n] + gSph.Rad[n-1])/2.0;
      }
      fprintf(logs.fp, "%11.5e", d);
      fprintf(logs.rad, "%3d %11.5e %13.6e", n, d, v->Tkin);
      for (m=0; m<=-mcp.TrInfo; m++) {
        l = rTra[m].l; u = rTra[m].u;
        pup = v->pop[u]; plow = v->pop[l];
        Tex[m] = Tau[m] = 0.0;
        if (pup > 0.0 && plow > 0.0) {
	    Tex[m] = -PLOBO_CNST * rTra[m].Fre / log(pup/plow/rTra[m].Gul);
	    Tau[m] = rTra[m].aconst * gSph.Rmax * (plow*rTra[m].Gul - pup) / v->Vdop *
	             (gSph.Rad[n] - gSph.Rad[n-1]);
	}
        fprintf(logs.fp, " %13.6e (%10.5f) %13.6e", Tex[m], Tex[m]/v->Tkin, Tau[m]);
        fprintf(logs.rad, " %13.6e %13.6e", Tex[m] , Tau[m]);
      }
      fprintf(logs.fp, "\n");
      fprintf(logs.rad, "\n");
    }
  }
  fclose(logs.rad);
}
