#define MAX_RTRANS	300
#define MAX_CTRANS	1600
#define MAX_LEVEL	100
#define MAX_CTEMP	15
#define MAX_VOLUME	220
#define MAX_BOUND	(2*MAX_VOLUME)
#define MAX_ANGLES	130
#define MAX_VEL		70
#define MAX_NG		10
#define MAX_OLAP	40

#define MAX_ZERO    MAX_VEL

#define CLIGHT          2.99792458e8    /* c [m/s]                     */
#define TAU_CNST        0.01072068      /* c[km/s] c[cm/s]^2 / 8 pi 10^27    */
#define TAUD_CNST		299792.458
#define PLOBO_CNST      0.04799216      /* h/k 10^9                     */
#define PLANCK_CNST     6.626076e-18    /* h 10^7 10^9                  */
#define BOLTZ_CNST      1.3806586e-16   /* k 10^7                       */
#define INT_CNST        1.4745e-23      /* 2 h 10^27 / c^2              */
#define LOG_PC_CNST     18.48935063     /* log(3.08*10^18 cm)           */
#define INVCM_TO_K      1.43877
#define K_TO_GHZ        20.83674
#define MH2_CNST        1.92e-19        /* (4*PI/3 mH2/Msun)^(1/3)      */
#define MDOT_CNST       2.58e-22
#define LSUN_CNST       1.818e-37       /* cm-2 K-4                    */
#ifndef PI
#define PI              3.141592654
#endif

#define GEO_SPHERE	0
#define GEO_CUBE	1
#define GEO_CLUMPS	2

#ifndef VECTOR
#include "vector.h"
#endif

typedef struct {
    char file[120], prf[120];
    FILE *fp, *pfp;
} LOGS;

typedef struct {
    int Nshell;
    double Rmax, Rmin;
    double Rad[MAX_VOLUME];
    double Tstar;
    double Vrad0, Vrad1, Vrad2, Vrad3, Vp[3];
    double Trad0, Trad1, Trad2, Trad3, Tp[3];
    double Nrad0, Nrad1, Nrad2, Nrad3, Np[3];
    double Wrad0, Wrad1, Wrad2, Wrad3, Wp[3];
    double Xrad0, Xrad1, Xrad2, Xrad3, Xp[3];
    double Tdrad0, Tdrad1, Tdrad2, Tdrad3, Tdp[3];
} G_SPHERE, *gSphPtr;

typedef struct {
    int Ncube;
    double Lx, Ly, Lz;
    double Tstar;
} G_CUBE, *gCubPtr;

typedef struct {
    int Nclump;
    double Rclump;
    double Tstar;
} G_CLUMP, *gCluPtr;

typedef struct {
    int Iter;
    char Molecule[120], datName[120], popFile[120];
    int Niter, Nlambda, Nng, Nng1, NgOrder, Cooling, Nstep;
    double RelLimit, maxrel, Gamma, PopFrac;
    double passBand, molWeight, TexGuess, Elimit;
    int Nlevel;
    int NrTrans, NcTrans;
    int NcTemp;
    double cTemp[MAX_CTEMP];
    int Nvol, Nang, NangMin, Nvel;
    int geometry, TrInfo, LimitLevel;
    gSphPtr gptr;
    double Tbg;
    double pVel[MAX_RTRANS];
    int Nbnd;
    int overlap;
    double dust;  /* dust is 3Q_250 / 4\rho  [cm-2/gram]  typically 10 */
    double Mdot, Lsun;
} MCP;

typedef struct {
    int n;
    double t;
} BOUND;

typedef struct {
    int u, l;
    double Gul;
    double Fre, Aul;
    int n, m[MAX_OLAP];
    double v_min, v_max;
    double dv[MAX_OLAP];
    double st, aconst;
    double dconst[MAX_VOLUME];
} RTRANS;

typedef struct {
    int u, l;
    double Gul;
    double Egy, Cul[MAX_CTEMP];
} CTRANS;

typedef struct {
    double g, Egy;
} LEVEL;

typedef struct {
    double Tkin, nH2, Xmol;
    double Vdop, Vturb;
    double Tdust, Rgd, beta;
    /*
      CARTESIAN Vel;
      CARTESIAN Pos;
      CARTESIAN Siz;
    */
    double volume;
    double pop[MAX_LEVEL];
    double p[MAX_NG+1][MAX_LEVEL];
    double Jbar[MAX_RTRANS];
    double Jold[MAX_RTRANS];
    double Lbar[MAX_RTRANS];
    double LbarOver[MAX_RTRANS][MAX_OLAP];
#ifdef COOLING
    double Hdg, Hpe, Hdt, Crad, CradC, CradCC, CradPrev, CH2;
    double Beta[MAX_RTRANS];
#endif
} VOLUME;

typedef struct {
    int n, Nang;
    BOUND b[MAX_BOUND];
    CARTESIAN org;
    CARTESIAN dir;
    double dmy, Tbg;
} GRID;

typedef struct {
    int n;
    double z[MAX_ZERO];
    double w[MAX_ZERO];
} HERMITE;

#ifndef MCP_EXTERNAL

volatile MCP 		mcp;
volatile LOGS 		logs;

volatile RTRANS     rTra[MAX_RTRANS];
volatile LEVEL      lev[MAX_LEVEL];
volatile VOLUME     vol[MAX_VOLUME];
volatile G_SPHERE   gSph;
volatile G_CUBE     gCub;
volatile G_CLUMP    gClu;
BOUND      bnd[MAX_BOUND];
volatile GRID       grid[MAX_VOLUME][MAX_ANGLES][2];
volatile HERMITE    Hz;

#else

extern MCP 			mcp;
extern LOGS			logs;

extern RTRANS 		rTra[MAX_RTRANS];
extern LEVEL 		lev[MAX_LEVEL];
extern VOLUME 		vol[MAX_VOLUME];
extern G_SPHERE		gSph;
extern G_CUBE		gCub;
extern G_CLUMP		gClu;
extern BOUND		bnd[MAX_BOUND];
extern GRID         grid[MAX_VOLUME][MAX_ANGLES][2];
extern HERMITE      Hz;

#endif
