#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include <assert.h>

#ifndef MCP_EXTERNAL
#define MCP_EXTERNAL
#endif

#include "ali.h"
#include "init.h"
#include "dat.h"

CTRANS cTra[MAX_CTRANS];

void dump()
{
    int i, j;

    printf("MCP\n");
    printf("%d\n", mcp.Iter);
    printf("'%s'\n", mcp.Molecule);
    printf("'%s'\n", mcp.datName);
    printf("%d %d %d %d %d %d %d\n", mcp.Niter, mcp.Nlambda, mcp.Nng, mcp.Nng1, mcp.NgOrder, mcp.Cooling, mcp.Nstep);
    printf("%lf %lf %lf %lf\n", mcp.RelLimit, mcp.maxrel, mcp.Gamma, mcp.PopFrac);
    printf("%lf %lf %lf %lf\n", mcp.passBand, mcp.molWeight, mcp.TexGuess, mcp.Elimit);
    printf("%d %d %d\n", mcp.Nlevel, mcp.NrTrans, mcp.NcTrans);
    printf("%d\n", mcp.NcTemp);
    for (i = 0; i < mcp.NcTemp; i++) printf("%lf ", mcp.cTemp[i]);
    printf("\n");
    printf("%d %d %d\n", mcp.Nvol, mcp.Nang, mcp.NangMin);
    printf("%d %d %d\n", mcp.geometry, mcp.TrInfo, mcp.LimitLevel);
    for (i = 0; i < mcp.NrTrans; i++) printf("%lf ", mcp.pVel[i]);
    printf("\n");
    printf("%lf\n", mcp.Tbg);
    /* printf("\n", mcp.pOrg, mcp.pDir); */
    printf("%d\n", mcp.Nbnd);
    printf("%d\n", mcp.overlap);
    printf("%lf\n", mcp.dust);
    printf("%lf %lf\n", mcp.Mdot, mcp.Lsun);

    printf("\nRTRANS\n");
    for (j = 0; j < mcp.NrTrans; j++) {
        printf("[%d] %d %d %lf %lf %lf %lf %lf %d %lf %lf\n", j,
               rTra[j].u, rTra[j].l, rTra[j].Gul, rTra[j].Fre, rTra[j].Aul, rTra[j].aconst, 
               rTra[j].st, rTra[j].n, rTra[j].v_min, rTra[j].v_max);
    }

    printf("\nLEVEL\n");
    for (j = 0; j < mcp.Nlevel; j++) {
        printf("[%d] %lf %lf\n", j, lev[j].g, lev[j].Egy);
    }

    printf("\nVOLUME\n");
    for (j = 0; j < mcp.Nvol; j++) {
    }

    printf("\nSPHERE\n");
    printf("%d\n", gSph.Nshell);
    printf("%lf %lf\n", gSph.Rmax, gSph.Rmin);
    for (i = 0; i < gSph.Nshell; i++) printf("%lf ", gSph.Rad[i]);
    printf("\n");
    printf("%lf\n", gSph.Tstar);
    printf("%lf %lf %lf %lf (%lf %lf %lf)\n", gSph.Vrad0, gSph.Vrad1, gSph.Vrad2, gSph.Vrad3, gSph.Vp[0], gSph.Vp[1], gSph.Vp[2]);
    printf("%lf %lf %lf %lf (%lf %lf %lf)\n", gSph.Trad0, gSph.Trad1, gSph.Trad2, gSph.Trad3, gSph.Tp[0], gSph.Tp[1], gSph.Tp[2]);
    printf("%lf %lf %lf %lf (%lf %lf %lf)\n", gSph.Nrad0, gSph.Nrad1, gSph.Nrad2, gSph.Nrad3, gSph.Np[0], gSph.Np[1], gSph.Np[2]);
    printf("%lf %lf %lf %lf (%lf %lf %lf)\n", gSph.Wrad0, gSph.Wrad1, gSph.Wrad2, gSph.Wrad3, gSph.Wp[0], gSph.Wp[1], gSph.Wp[2]);
    printf("%lf %lf %lf %lf (%lf %lf %lf)\n", gSph.Xrad0, gSph.Xrad1, gSph.Xrad2, gSph.Xrad3, gSph.Xp[0], gSph.Xp[1], gSph.Xp[2]);
    printf("%lf %lf %lf %lf (%lf %lf %lf)\n", gSph.Tdrad0, gSph.Tdrad1, gSph.Tdrad2, gSph.Tdrad3, gSph.Tdp[0], gSph.Tdp[1], gSph.Tdp[2]);
}

void init(char *molecule)
{
    void read_input_file();
    void read_dat_file();
    void init_dat_data();
    void pop_init();
    void col_init();
    void tra_init();
    void overlap_init();
    void weight_init(), grid_init(), hermite_init(int);
    void init_log_info();
    void ReadPopulations(char *);
  
    strcpy(mcp.Molecule, molecule);
  
    strcpy(logs.file, molecule);
    strcat(logs.file, LOG_SUFFIX);
  
    logs.fp = fopen(logs.file, "w");
  
    if (!logs.fp) {
        fprintf(stderr, "Fatal error: Cannot open '%s'\n", logs.file);
        exit(1);
    }
  
    strcpy(logs.prf, molecule);
    strcat(logs.prf, PRF_SUFFIX);
  
    logs.pfp = fopen(logs.prf, "w");
  
    if (!logs.pfp) {
        fprintf(stderr, "Fatal error: Cannot open '%s'\n", logs.prf);
        exit(1);
    }
  
    read_input_file();
  
    read_dat_file();
    init_dat_data();
  
    if (mcp.TexGuess < 0.0) {
        ReadPopulations(mcp.popFile);
    } else {
        pop_init(mcp.TexGuess);
    }
  
    col_init();
  
    tra_init();
  
    if (mcp.overlap) overlap_init();
  
    weight_init();
  
    hermite_init(mcp.Nvel);
  
    grid_init();
  
    init_log_info();
}

void read_input_file()
{
    int n=0, i, narg, nstep;
    char str[LEN_ROW], InputFile[LEN_FILENAME], type[20];
    double dv[MAX_VOLUME], r, r1, rat_iso;
    FILE *ifp;
  
    int is_comment();
    void get_dvector(), get_ivector(), init_rad();
    double Trad(), Nrad(), Xrad(), Xexp(), Xexp1(), Wrad(), Tdrad();
#ifdef COOLING
    void ReadTkin(char *);
#endif
  
    strcpy(InputFile, mcp.Molecule);
    strcat(InputFile, INP_SUFFIX);
  
    ifp = fopen(InputFile, "r");
  
    if (!ifp) {
        fprintf(stderr, "Fatal error: Cannot open file '%s'\n", InputFile);
        exit(1);
    }
  
    mcp.Gamma = mcp.PopFrac = 1.0;
    mcp.Nstep = 4;
  
    while (fgets(str, LEN_ROW, ifp) != NULL) {
        if (is_comment(str)) continue;
        n++;
        if (n==1) {
            narg = sscanf(str, "%d %d %d %d %d %lf %lf %d %d",
                          &mcp.Niter, &mcp.Nlambda,
                          &mcp.Nng, &mcp.Nng1, &mcp.NgOrder, &mcp.RelLimit, &mcp.PopFrac,
                          &mcp.Cooling, &nstep);
            if (narg == 9) mcp.Nstep=nstep;
        } else if (n==2) {
            sscanf(str, "%s", mcp.datName);
        } else if (n==3) {
            sscanf(str, "%s", mcp.popFile);
        } else if (n==4) {
            sscanf(str, "%lf %lf %lf %lf %d %lf",
                   &mcp.passBand, &mcp.TexGuess, &mcp.Tbg, &mcp.Elimit,
                   &mcp.overlap, &mcp.dust);
        } else if (n==5) {
            sscanf(str, "%d %d %d", &mcp.geometry, &mcp.TrInfo, &mcp.LimitLevel);
            switch (mcp.geometry) {
              case GEO_SPHERE:
              case GEO_CUBE:
              case GEO_CLUMPS:
                break;
              default:
                fprintf(stderr, "Warning: Invalid geometry parameter (%d)\n", mcp.geometry);
                fprintf(stderr, "         Assuming spherical geometry.\n");
                mcp.geometry = GEO_SPHERE;
                break;
            }
        } else if (n==6 && mcp.geometry==GEO_SPHERE) {
            sscanf(str, "%d %lf %lf %d %d %d",
                   &gSph.Nshell, &gSph.Rmax, &gSph.Tstar,
                   &mcp.Nang, &mcp.NangMin, &mcp.Nvel);
            if (gSph.Nshell > MAX_VOLUME) {
                fprintf(stderr, "Fatal error: Nshell (%d) > MAX_VOLUME (%d)\n",
                        gSph.Nshell, MAX_VOLUME);
                exit(1);
            }
            if (mcp.Nang % 2  == 1) {
                fprintf(stderr, "Warning: Nang (%d) must be even. Using %d.\n",
                        mcp.Nang, mcp.Nang+1);
                ++mcp.Nang;
            }
            if (mcp.NangMin % 2  == 1) {
                fprintf(stderr, "Warning: NangMin (%d) must be even. Using %d.\n",
                        mcp.NangMin, mcp.NangMin+1);
                ++mcp.NangMin;
            }
            if (mcp.NangMin >= MAX_ANGLES) {
                fprintf(stderr, "Fatal error: NangMin (%d) >= MAX_ANGLES (%d)\n",
                        mcp.NangMin, MAX_ANGLES);
                exit(1);
            }
            if (mcp.Nang >= MAX_ANGLES) {
                fprintf(stderr, "Fatal error: Nang (%d) >= MAX_ANGLES (%d)\n",
                        mcp.Nang, MAX_ANGLES);
                exit(1);
            }
            /* if (mcp.Nvel % 2  == 1) {
               fprintf(stderr, "Warning: Nvel (%d) must be even. Using %d.\n",
               mcp.Nvel, ++mcp.Nvel);
               } */
            if (mcp.Nvel > MAX_VEL) {
                fprintf(stderr, "Fatal error: Nvel (%d) > MAX_VEL (%d)\n",
                        mcp.Nvel, MAX_VEL);
                exit(1);
            }
            mcp.Nvol = gSph.Nshell;
        } else if (n==7 && mcp.geometry==GEO_SPHERE) {
            if (str[0] == 'F') {
                if (sscanf(str, "F LOG %lf", &gSph.Rmin) == 1) {
                    init_rad("log", gSph.Rad, gSph.Rmin, mcp.Nvol, MAX_VOLUME);
                } else if (sscanf(str, "F VOL %lf", &gSph.Rmin) == 1) {
                    init_rad("vol", gSph.Rad, gSph.Rmin, mcp.Nvol, MAX_VOLUME);
                } else if (sscanf(str, "F LIN %lf", &gSph.Rmin) == 1) {
                    init_rad("lin", gSph.Rad, gSph.Rmin, mcp.Nvol, MAX_VOLUME);
                } else {
                    sscanf(str, "F %s %lf", type, &gSph.Rmin);
                    init_rad(type, gSph.Rad, gSph.Rmin, mcp.Nvol, MAX_VOLUME);
                }
                if (gSph.Tstar < 0.0) {
                    mcp.Lsun = fabs(gSph.Tstar);
                    gSph.Tstar = pow(mcp.Lsun/LSUN_CNST, 0.25)/
                        sqrt(gSph.Rmin * gSph.Rmax);
                }
            } else if (str[0] == 'L') {
                if (sscanf(str, "L LOG %lf", &mcp.Lsun) == 1) {
                    strcpy(type, "log");
                } else if (sscanf(str, "L VOL %lf", &mcp.Lsun) == 1) {
                    strcpy(type, "vol");
                } else if (sscanf(str, "L LIN %lf", &mcp.Lsun) == 1) {
                    strcpy(type, "lin");
                } else {
                    sscanf(str, "L %s %lf", type, &mcp.Lsun);
                }
                gSph.Rmin = sqrt(mcp.Lsun/LSUN_CNST)/gSph.Tstar/gSph.Tstar/gSph.Rmax;
                init_rad(type, gSph.Rad, gSph.Rmin, mcp.Nvol, MAX_VOLUME);
            } else {
                get_dvector(str, gSph.Rad, MAX_VOLUME);
                gSph.Rmin = gSph.Rad[0];
            }
        } else if (n==8 && mcp.geometry==GEO_SPHERE) {
            if (str[0] == 'Q') {
                sscanf(str, "Q %lf %lf %lf", &gSph.Vrad0, &gSph.Vrad1, &gSph.Vrad2);
            } else {
                sscanf(str, "%lf %lf %lf", &gSph.Vrad0, &gSph.Vrad1, &gSph.Vrad2);
            }
        } else if (n==9) {
            if (str[0] == 'F') {
                sscanf(str, "F %lf %lf %lf", &gSph.Trad0, &gSph.Trad1, &gSph.Trad2);
                for (i=1; i<mcp.Nvol; i++) {
                    r = (gSph.Rad[i] + gSph.Rad[i-1])/2.0;
                    if (i == 1 && gSph.Rmin == 0.0) r = 0.0;
                    vol[i].Tkin = Trad(r);
                }
            } else if (str[0] == 'D') {
                sscanf(str, "D %lf %lf %lf", &gSph.Trad0, &gSph.Trad1, &gSph.Trad2);
                for (i=1; i<mcp.Nvol; i++) {
                    r = (gSph.Rad[i] + gSph.Rad[i-1])/2.0;
                    if (i == 1 && gSph.Rmin == 0.0) r = 0.0;
                    if (r < gSph.Trad1)
                        vol[i].Tkin = gSph.Trad0;
                    else
                        vol[i].Tkin = gSph.Trad2;
                }
            } else if (str[0] == 'Q') {
                narg = sscanf(str, "Q %lf %lf %lf %lf %lf %lf %lf",
                              &gSph.Trad0, &gSph.Tp[0], &gSph.Trad1, &gSph.Tp[1],
                              &gSph.Trad2, &gSph.Tp[2], &gSph.Trad3);
                for (i=1; i<mcp.Nvol; i++) {
                    r = (gSph.Rad[i] + gSph.Rad[i-1])/2.0;
                    if (i == 1 && gSph.Rmin == 0.0) r = 0.0;
                    if (r < gSph.Tp[0])
                        vol[i].Tkin = gSph.Trad0;
                    else if (r >= gSph.Tp[0] && r < gSph.Tp[1])
                        vol[i].Tkin = gSph.Trad1;
                    else if (r >= gSph.Tp[1] && r < gSph.Tp[2])
                        vol[i].Tkin = gSph.Trad2;
                    else
                        vol[i].Tkin = gSph.Trad3;
                }
#ifdef COOLING
            } else if (str[0] == '.') {
                ReadTkin(str);
#endif
            } else if (str[0] == 'i') {
                sscanf(str, "i %lf %lf %lf %lf %lf %lf",
                       &gSph.Trad0, &gSph.Trad1, &gSph.Trad2,
                       &gSph.Tp[0], &gSph.Tp[1], &gSph.Tp[2]);
                for (i=1; i<mcp.Nvol; i++) {
                    r = (gSph.Rad[i] + gSph.Rad[i-1])/2.0;
                    if (i == 1 && gSph.Rmin == 0.0) r = 0.0;
                    if (r <= gSph.Tp[2]) {
                        vol[i].Tkin = Trad(r/gSph.Tp[0]);
                    } else {
                        vol[i].Tkin = gSph.Tp[1];
                    }
                }
            } else {
                get_dvector(str, dv, MAX_VOLUME);
                for (i=0; i<mcp.Nvol; i++) vol[i].Tkin = dv[i];
            }
        } else if (n==10) {
            if (str[0] == 'F') {
                sscanf(str, "F %lf %lf %lf", &gSph.Nrad0, &gSph.Nrad1, &gSph.Nrad2);
                gSph.Nrad3 = 0.0;
                for (i=1; i<mcp.Nvol; i++) {
                    r = (gSph.Rad[i] + gSph.Rad[i-1])/2.0;
                    if (i == 1 && gSph.Rmin == 0.0) r = 0.0;
                    vol[i].nH2 = Nrad(r);
                }
            } else if (str[0] == 'D') {
                sscanf(str, "D %lf %lf %lf", &gSph.Nrad0, &gSph.Nrad1, &gSph.Nrad2);
                gSph.Nrad3 = 0.0;
                for (i=1; i<mcp.Nvol; i++) {
                    r = (gSph.Rad[i] + gSph.Rad[i-1])/2.0;
                    if (i == 1 && gSph.Rmin == 0.0) r = 0.0;
                    if (r < gSph.Nrad1)
                        vol[i].nH2 = gSph.Nrad0;
                    else
                        vol[i].nH2 = gSph.Nrad2;
                }
            } else if (str[0] == 'Q') {
                narg = sscanf(str, "Q %lf %lf %lf %lf %lf %lf %lf",
                              &gSph.Nrad0, &gSph.Np[0], &gSph.Nrad1, &gSph.Np[1],
                              &gSph.Nrad2, &gSph.Np[2], &gSph.Nrad3);
                for (i=1; i<mcp.Nvol; i++) {
                    r = (gSph.Rad[i] + gSph.Rad[i-1])/2.0;
                    if (i == 1 && gSph.Rmin == 0.0) r = 0.0;
                    if (r < gSph.Np[0])
                        vol[i].nH2 = gSph.Nrad0;
                    else if (r >= gSph.Np[0] && r < gSph.Np[1])
                        vol[i].nH2 = gSph.Nrad1;
                    else if (r >= gSph.Np[1] && r < gSph.Np[2])
                        vol[i].nH2 = gSph.Nrad2;
                    else
                        vol[i].nH2 = gSph.Nrad3;
                }
            } else if (str[0] == 'f') {
                sscanf(str, "f %lf %lf %lf %lf",
                       &gSph.Nrad0, &gSph.Nrad1, &gSph.Nrad2, &gSph.Nrad3);
                for (i=1; i<mcp.Nvol; i++) {
                    r = (gSph.Rad[i] + gSph.Rad[i-1])/2.0;
                    if (i == 1 && gSph.Rmin == 0.0) r = 0.0;
                    vol[i].nH2 = Nrad(r);
                }
            } else if (str[0] == 'i') {
                sscanf(str, "i %lf %lf %lf %lf %lf %lf",
                       &gSph.Nrad0, &gSph.Nrad1, &gSph.Nrad2,
                       &gSph.Np[0], &gSph.Np[1], &gSph.Np[2]);
                for (i=1; i<mcp.Nvol; i++) {
                    r = (gSph.Rad[i] + gSph.Rad[i-1])/2.0;
                    if (i == 1 && gSph.Rmin == 0.0) r = 0.0;
                    if (r <= gSph.Np[2])
                        vol[i].nH2 = Nrad(r/gSph.Np[0]);
                    else
                        vol[i].nH2 = gSph.Np[1];
                }
            } else if (str[0] == 'M' && str[1] == '.') {
                sscanf(str, "M. %lf", &mcp.Mdot);
                vol[mcp.Nvol-1].nH2 = mcp.Mdot / pow(MDOT_CNST*gSph.Rmax, 2.0) /
                    gSph.Vrad0;
                r1 = (gSph.Rad[mcp.Nvol-1] + gSph.Rad[mcp.Nvol-1-2])/2.0;
                for (i=1; i<mcp.Nvol-1; i++) {
                    r = (gSph.Rad[i] + gSph.Rad[i-1])/2.0;
                    vol[i].nH2 = vol[mcp.Nvol-1].nH2 * r1*r1/r/r;;
                }
            } else {
                get_dvector(str, dv, MAX_VOLUME);
                for (i=0; i<mcp.Nvol; i++) vol[i].nH2 = dv[i];
            }
        } else if (n==11) {
            if (str[0] == 'F') {
                sscanf(str, "F %lf %lf %lf", &gSph.Xrad0, &gSph.Xrad1, &gSph.Xrad2);
                gSph.Xrad3 = 0.0;
                for (i=1; i<mcp.Nvol; i++) {
                    r = (gSph.Rad[i] + gSph.Rad[i-1])/2.0;
                    if (i == 1 && gSph.Rmin == 0.0) r = 0.0;
                    vol[i].Xmol = Xrad(r);
                }
            } else if (str[0] == 'D') {
                narg = sscanf(str, "D %lf %lf %lf %lf",
                              &gSph.Xrad0, &gSph.Xrad1, &gSph.Xrad2, &rat_iso);
                if (narg != 4) rat_iso = 1.0;
                gSph.Xrad3 = 0.0;
                for (i=1; i<mcp.Nvol; i++) {
                    r = (gSph.Rad[i] + gSph.Rad[i-1])/2.0;
                    if (i == 1 && gSph.Rmin == 0.0) r = 0.0;
                    if (r < gSph.Xrad1)
                        vol[i].Xmol = gSph.Xrad0/rat_iso;
                    else
                        vol[i].Xmol = gSph.Xrad2/rat_iso;
                }
            } else if (str[0] == 'Q') {
                narg = sscanf(str, "Q %lf %lf %lf %lf %lf %lf %lf %lf",
                              &gSph.Xrad0, &gSph.Xp[0], &gSph.Xrad1, &gSph.Xp[1],
                              &gSph.Xrad2, &gSph.Xp[2], &gSph.Xrad3, &rat_iso);
                if (narg != 8) rat_iso = 1.0;
                for (i=1; i<mcp.Nvol; i++) {
                    r = (gSph.Rad[i] + gSph.Rad[i-1])/2.0;
                    if (i == 1 && gSph.Rmin == 0.0) r = 0.0;
                    if (r < gSph.Xp[0])
                        vol[i].Xmol = gSph.Xrad0/rat_iso;
                    else if (r >= gSph.Xp[0] && r < gSph.Xp[1])
                        vol[i].Xmol = gSph.Xrad1/rat_iso;
                    else if (r >= gSph.Xp[1] && r < gSph.Xp[2])
                        vol[i].Xmol = gSph.Xrad2/rat_iso;
                    else
                        vol[i].Xmol = gSph.Xrad3/rat_iso;
                }
            } else if (str[0] == 'i') {
                sscanf(str, "i %lf %lf %lf %lf %lf %lf",
                       &gSph.Xrad0, &gSph.Xrad1, &gSph.Xrad2,
                       &gSph.Xp[0], &gSph.Xp[1], &gSph.Xp[2]);
                for (i=1; i<mcp.Nvol; i++) {
                    r = (gSph.Rad[i] + gSph.Rad[i-1])/2.0;
                    if (i == 1 && gSph.Rmin == 0.0) r = 0.0;
                    if (r <= gSph.Xp[2])
                        vol[i].Xmol = Xrad(r/gSph.Xp[0]);
                    else
                        vol[i].Xmol = gSph.Xp[1];
                }
            } else if (str[0] == 'M') {
                sscanf(str, "M %lf %lf %lf %lf",
                       &gSph.Xrad0, &gSph.Xrad1, &gSph.Xrad2, &gSph.Xrad3);
                for (i=1; i<mcp.Nvol; i++) {
                    r = (gSph.Rad[i] + gSph.Rad[i-1])/2.0;
                    if (i == 1 && gSph.Rmin == 0.0) r = 0.0;
                    vol[i].Xmol = Xexp1(r);
                }
            } else if (str[0] == 'E') {
                sscanf(str, "E %lf %lf %lf", &gSph.Xrad0, &gSph.Xrad1, &gSph.Xrad2);
                gSph.Xrad3 = 0.0;
                for (i=1; i<mcp.Nvol; i++) {
                    r = (gSph.Rad[i] + gSph.Rad[i-1])/2.0;
                    if (i == 1 && gSph.Rmin == 0.0) r = 0.0;
                    vol[i].Xmol = Xexp(r);
                }
            } else if (str[0] == 'e') {
                sscanf(str, "e %lf %lf %lf %lf",
                       &gSph.Xrad0, &gSph.Xrad1, &gSph.Xrad2, &gSph.Xrad3);
                for (i=1; i<mcp.Nvol; i++) {
                    r = (gSph.Rad[i] + gSph.Rad[i-1])/2.0;
                    if (i == 1 && gSph.Rmin == 0.0) r = 0.0;
                    vol[i].Xmol = Xexp(r);
                }
            } else {
                get_dvector(str, dv, MAX_VOLUME);
                for (i=0; i<mcp.Nvol; i++) vol[i].Xmol = dv[i];
            }
        } else if (n==12) {
            if (str[0] == 'F') {
                sscanf(str, "F %lf %lf %lf", &gSph.Wrad0, &gSph.Wrad1, &gSph.Wrad2);
                for (i=1; i<mcp.Nvol; i++) {
                    r = (gSph.Rad[i] + gSph.Rad[i-1])/2.0;
                    if (i == 1 && gSph.Rmin == 0.0) r = 0.0;
                    vol[i].Vturb = Wrad(r, 'F');
                }
            } else if (str[0] == 'D') {
                sscanf(str, "D %lf %lf %lf", &gSph.Wrad0, &gSph.Wrad1, &gSph.Wrad2);
                for (i=1; i<mcp.Nvol; i++) {
                    r = (gSph.Rad[i] + gSph.Rad[i-1])/2.0;
                    if (i == 1 && gSph.Rmin == 0.0) r = 0.0;
                    if (r < gSph.Wrad1)
                        vol[i].Vturb = gSph.Wrad0;
                    else
                        vol[i].Vturb = gSph.Wrad2;
                }
            } else if (str[0] == 'Q') {
                narg = sscanf(str, "Q %lf %lf %lf %lf %lf %lf %lf",
                              &gSph.Wrad0, &gSph.Wp[0], &gSph.Wrad1, &gSph.Wp[1],
                              &gSph.Wrad2, &gSph.Wp[2], &gSph.Wrad3);
                for (i=1; i<mcp.Nvol; i++) {
                    r = (gSph.Rad[i] + gSph.Rad[i-1])/2.0;
                    if (i == 1 && gSph.Rmin == 0.0) r = 0.0;
                    if (r < gSph.Wp[0])
                        vol[i].Vturb = gSph.Wrad0;
                    else if (r >= gSph.Wp[0] && r < gSph.Wp[1])
                        vol[i].Vturb = gSph.Wrad1;
                    else if (r >= gSph.Wp[1] && r < gSph.Wp[2])
                        vol[i].Vturb = gSph.Wrad2;
                    else
                        vol[i].Vturb = gSph.Wrad3;
                }
            } else if (str[0] == 'i') {
                sscanf(str, "i %lf %lf %lf %lf %lf %lf",
                       &gSph.Wrad0, &gSph.Wrad1, &gSph.Wrad2,
                       &gSph.Wp[0], &gSph.Wp[1], &gSph.Wp[2]);
                for (i=1; i<mcp.Nvol; i++) {
                    r = (gSph.Rad[i] + gSph.Rad[i-1])/2.0;
                    if (i == 1 && gSph.Rmin == 0.0) r = 0.0;
                    if (r <= gSph.Wp[2])
                        vol[i].Vturb = Wrad(r/gSph.Wp[0], 'F');
                    else
                        vol[i].Vturb = gSph.Wp[1];
                }
            } else if (str[0] == 'G') {
                sscanf(str, "G %lf %lf %lf %lf %lf",
                       &gSph.Wrad0, &gSph.Wrad1, &gSph.Wp[0],
                       &gSph.Wp[1], &gSph.Wp[2]);
                for (i=1; i<mcp.Nvol; i++) {
                    r = (gSph.Rad[i] + gSph.Rad[i-1])/2.0;
                    if (i == 1 && gSph.Rmin == 0.0) r = 0.0;
                    vol[i].Vturb = Wrad(r, 'G');
                }
            } else {
                get_dvector(str, dv, MAX_VOLUME);
                for (i=0; i<mcp.Nvol; i++) vol[i].Vturb = dv[i];
            }
        } else if (n==13) {
            if (str[0] == 'F') {
                sscanf(str, "F %lf %lf %lf", &gSph.Tdrad0, &gSph.Tdrad1, &gSph.Tdrad2);
                for (i=1; i<mcp.Nvol; i++) {
                    r = (gSph.Rad[i] + gSph.Rad[i-1])/2.0;
                    if (i == 1 && gSph.Rmin == 0.0) r = 0.0;
                    vol[i].Tdust = Tdrad(r);
                }
            } else if (str[0] == 'T') {
                for (i=1; i<mcp.Nvol; i++) vol[i].Tdust = vol[i].Tkin;
            } else {
                get_dvector(str, dv, MAX_VOLUME);
                for (i=0; i<mcp.Nvol; i++) vol[i].Tdust = dv[i];
            }
        } else if (n==14) {
            get_dvector(str, dv, MAX_VOLUME);
            for (i=0; i<mcp.Nvol; i++) vol[i].Rgd = dv[i];
        } else if (n==15) {
            get_dvector(str, dv, MAX_VOLUME);
            for (i=0; i<mcp.Nvol; i++) vol[i].beta = dv[i];
        } else {
            fprintf(stderr, "Skipping line '%s' in file '%s'\n", str, InputFile);
        }
    }
  
    fclose(ifp);
}

char *trimwhitespace(char *str)
{
  char *end;

  while(isspace(*str)) str++;

  if(*str == 0) return str;

  end = str + strlen(str) - 1;
  while(end > str && isspace(*end)) end--;

  *(end+1) = 0;

  return str;
}

void read_dat_file()
{
    int i, k, n=0, nn=-1, key=0, nkeys;
    char str[LEN_ROW], DatFile[LEN_FILENAME], *ptr, *tok;
    char key_word[LEN_KEY];
    /* char tmp[LEN_ROW]; */
    FILE *dfp;
  
    int is_comment(), is_keyword();
  
    strcpy(DatFile, DAT_DIR);
    strcat(DatFile, mcp.datName);
    strcat(DatFile, DAT_SUFFIX);
  
    dfp = fopen(DatFile, "r");
  
    if (!dfp) {
        fprintf(stderr, "Fatal error: Cannot open file '%s'\n", DatFile);
        exit(1);
    }
  
    i = 0;
    nkeys = sizeof(kw)/sizeof(KW_STRUCT);
  
    while (fgets(str, LEN_ROW, dfp) != NULL) {
        ptr = trimwhitespace(str);
        if (is_comment(ptr)) continue;
        if (is_keyword(ptr)) {
#ifdef DEBUG
            printf("'%s'\n", ptr);
#endif
            sscanf(ptr, "!%s", key_word);
            n = 0;
            key = K_UNDEF;
            for (k = 0; k < nkeys; k++) {
                if (strcmp(key_word, kw[k].label) == 0) {
                    key = k;
                    if (kw[k].type == KW_REAL_ARR) {
                        nn++;
                    } else {
                        nn = -1;
                    }
                    break;
                }
            }
            if (key == K_UNDEF) {
                fprintf(stderr, "Warning: Undefined keyword: %s\n", key_word);
            }
        } else if (key != K_UNDEF) {
            switch (kw[key].type) {
              case KW_STRING:
                strcpy(kw[key].p.c, ptr);
                break;
              case KW_INT:
                kw[key].p.i[0] = atoi(ptr);
#ifdef DEBUG
                printf("KW_INT: %d\n", kw[key].p.i[0]);
#endif
                break;
              case KW_REAL:
                kw[key].p.d[0] = atof(ptr);
#ifdef DEBUG
                printf("KW_REAL: %lf\n", kw[key].p.d[0]);
#endif
                break;
              case KW_INT_VEC:
                tok = strtok(ptr, " \t");
                while (tok != NULL) {
                    kw[key].p.i[n] = atoi(tok);
#ifdef DEBUG
                    printf("KW_INT_VEC: [%d] %d\n", n, kw[key].p.i[n]);
#endif
                    kw[key].p.i[n]--;
                    n++;
                    tok = strtok(NULL, " \t");
                }
                break;
              case KW_REAL_VEC:
                tok = strtok(ptr, " \t");
                while (tok != NULL) {
                    kw[key].p.d[n] = atof(tok);
#ifdef DEBUG
                    printf("KW_REAL_VEC: [%d] %lf\n", n, kw[key].p.d[n]);
#endif
                    n++;
                    tok = strtok(NULL, " \t");
                }
                break;
              case KW_REAL_ARR:
                tok = strtok(ptr, " \t");
                while (tok != NULL) {
                    if (strcmp(kw[key].label, "C_COEFFS") == 0) {
                        c_coeffs[nn][n] = atof(tok);
#ifdef DEBUG
                        printf("KW_REAL_ARR: [%d][%d] %lf\n", nn, n, c_coeffs[nn][n]);
#endif
                    } else if (strcmp(kw[key].label, "C1_COEFFS") == 0) {
                        c1_coeffs[nn][n] = atof(tok);
#ifdef DEBUG
                        printf("KW_REAL_ARR: [%d][%d] %lf\n", nn, n, c1_coeffs[nn][n]);
#endif
                    }
                    n++;
                    tok = strtok(NULL, " \t");
                }
                break;
            }
        }
        i++;
    }
  
    fclose(dfp);
}

void init_dat_data()
{
    double e0;
    int n, nn, u, l, lim=0, m=0;
  
    mcp.molWeight = molw;
  
    if (mcp.Elimit > 0.0) lim = 1;

    if (n_levs > MAX_LEVEL) {
        fprintf(stderr, "Increase MAX_level to %d.\n", n_levs);
        exit(1);
    }
    mcp.Nlevel = n_levs;

    if (n_rtrans > MAX_RTRANS) {
        fprintf(stderr, "Increase MAX_RTRANS to %d.\n", n_rtrans);
        exit(1);
    }
    mcp.NrTrans = n_rtrans;
  
#ifdef DEBUG
    printf("No. of levels: %d   No. of radiative transitions: %d\n",
           n_levs, n_rtrans);
#endif

    e0 = egy[0];
    if (lim) m = 0;
    for (n=0; n<mcp.Nlevel; n++) {
        if (lim && (egy[n]-e0)*INVCM_TO_K > mcp.Elimit) continue;
        lev[n].g = g[n];
        lev[n].Egy = (egy[n]-e0)*INVCM_TO_K;
#ifdef DEBUG
        printf("i=%d  g=%lf  egy=%lf\n", n, lev[n].g, lev[n].Egy);
#endif
        if (lim) m++;
    }
    if (lim) mcp.Nlevel = m;
  
    if (mcp.LimitLevel <= 0 || mcp.LimitLevel > mcp.Nlevel)
        mcp.LimitLevel = mcp.Nlevel;
  
    if (lim) m = 0;
    for (n=0; n<mcp.NrTrans; n++) {
        if (lim && (egy[ul[n]]-e0)*INVCM_TO_K > mcp.Elimit) continue;
        if (!lim) m = n;
        u = rTra[m].u = ul[n];
        l = rTra[m].l = ll[n];
        if (u >= n_levs || u < 0) {
            printf("WARNING!  RAD_UPPER is corrupted: %d %d\n", n, u);
        }
        if (l >= n_levs || l < 0) {
            printf("WARNING!  RAD_LOWER is corrupted: %d %d\n", n, l);
        }
        rTra[m].Aul = aul[n];
        rTra[m].Gul = lev[u].g/lev[l].g;
        rTra[m].Fre = (lev[u].Egy - lev[l].Egy)*K_TO_GHZ;
#ifdef DEBUG
        printf("%3d ful=%10.3e Aul=%10.3e Gul=%10.3e\n", m,
               rTra[m].Fre, rTra[m].Aul, rTra[m].Gul);
#endif
        if (lim) m++;
    }
    if (lim) mcp.NrTrans = m;

    if (n_ctrans > MAX_CTRANS) {
        fprintf(stderr, "Increase MAX_CTRANS to %d.\n", n_ctrans);
        exit(1);
    }
    mcp.NcTrans = n_ctrans;
  
    if (n_tk > MAX_CTEMP) {
        fprintf(stderr, "Increase MAX_CTEMP to %d.\n", n_tk);
        exit(1);
    }
    mcp.NcTemp = n_tk;
#ifdef DEBUG
    printf("No. of collisional transitions: %d  No. coll. temps: %d\n",
           n_ctrans, n_tk);
#endif

    for (nn=0; nn<mcp.NcTemp; nn++) {
        mcp.cTemp[nn] = c_tk[nn];
    }
  
    if (lim) m = 0;
    for (n=0; n<mcp.NcTrans; n++) {
        if (lim && (egy[c_ul[n]]-e0)*INVCM_TO_K > mcp.Elimit) continue;
        if (!lim) m = n;
        u = cTra[m].u = c_ul[n];
        l = cTra[m].l = c_ll[n];
        cTra[m].Gul = lev[u].g/lev[l].g;
        cTra[m].Egy = lev[u].Egy - lev[l].Egy;
#ifdef DEBUG
        printf("n=%3d %3d->%3d  Gul=%7.3lf   Ediff=%10.3lf\n", n, u, l,
               cTra[n].Gul, cTra[n].Egy);
#endif
        if (u >= n_levs || u < 0) {
            printf("WARNING!  COL_UPPER is corrupted: %d %d\n", n, u);
        }
        if (l >= n_levs || l < 0) {
            printf("WARNING!  COL_LOWER is corrupted: %d %d\n", n, l);
        }
        for (nn=0; nn<mcp.NcTemp; nn++) {
            cTra[m].Cul[nn] = c_coeffs[nn][n];
        }
        if (lim) m++;
    }
  
    if (lim) mcp.NcTrans = m;
  
    if (lim) {
        printf("Warning: The energy limit of %f K has resulted in\n", mcp.Elimit);
        printf("         No of energy levs = %d (%d)\n", mcp.Nlevel, n_levs);
        printf("         No of rad. trans  = %d (%d)\n", mcp.NrTrans, n_rtrans);
        printf("         No of col. trans  = %d (%d)\n", mcp.NcTrans, n_ctrans);
    }
  
#ifdef DEBUG
    for (n=0; n<mcp.Nvol; n++) {
        printf("%3d Tkin=%6.1lf nh2=%10.3e\n", n, vol[n].Tkin, vol[n].nH2);
    }
#endif
}

int is_comment(char *str)
{
    int i;
    char c;
  
    if ((int)strlen(str) == 0) return 1;
  
    i = 0;
    c = str[0];
  
    while (i < (int)strlen(str)-1 && (c == ' ' || c == '\t') ) {
        i++;
        c = str[i];
    }
  
    if (i >= (int)strlen(str)) return 1;	/* Blank line == comment */
  
    if (c == '%' || c == '#' || c == '*' || c == '\0') return 1;
  
    return 0;
}

int is_keyword(char *str)
{
    int i;
    char c;
  
    if ((int)strlen(str) == 0) return 0;
  
    i = 0;
    c = str[0];
  
    while (i < (int)strlen(str)-1 && (c == ' ' || c == '\t') ) {
        i++;
        c = str[i];
    }
  
    if (i >= (int)strlen(str)) return 0;	/* Blank line == comment */
  
    if (c == '!') return 1;
  
    return 0;
}

void get_dvector(char *str, double *d, int size)
{
    int i=0, n;
    char *delims = " ";
    char *token;
  
    token = strtok(str, delims);
  
    while (token != NULL && i < size) {
        d[i] = atof(token);
        token = strtok(NULL, delims);
        i++;
    }
    if (i > 0) {
        for (n=i; n<size; n++) d[n] = d[n-1];
    }
}

void get_ivector(char *str, int *d, int size)
{
    int i=0;
    char *delims = " ";
    char *token;
  
    token = strtok(str, delims);
  
    while (token != NULL && i < size) {
        d[i] = atoi(token);
        token = strtok(NULL, delims);
        i++;
    }
}

void init_log_info()
{
    int n, m;
    double f, r, NH2=0.0, Nmol=0.0, H2mass=0.0, TauDust=0.0, wSum=0.0;
    double Lum;
    FILE *fp;
  
    void i_fprintf(), d_fprintf();
    double VelRad(double);
  
    fp = logs.fp;
  
    if (!fp) return;
  
    fprintf(fp, "Molecule: %s (%5.1f)\n", mcp.Molecule, mcp.molWeight);
  
    fprintf(fp, "\nMax iter=%3d  Accel. i=%d  Ng-acc i=%d  Conv. lim=%10.3e  Nstep=%d\n",
            mcp.Niter, mcp.Nlambda, mcp.Nng, mcp.RelLimit, mcp.Nstep);
  
    fprintf(fp, "\nPopFrac=%7.3f  Passband=%7.2f km/s  TexGuess=%7.2f K   Tbg=%8.3f K\n",
            mcp.PopFrac, mcp.passBand, mcp.TexGuess, mcp.Tbg);
          
    fprintf(fp, "\nNAng=%3d NVel=%3d  NVol=%3d  Rmax=%12.5e cm  Tstar=%7.1f K\n",
            mcp.Nang, mcp.Nvel, mcp.Nvol, gSph.Rmax, gSph.Tstar);
          
    if (gSph.Tstar > 0.0 && gSph.Rad[0] > 0.0) {
        r = gSph.Rad[0]*gSph.Rmax;
        Lum = LSUN_CNST * r * r * gSph.Tstar*gSph.Tstar*gSph.Tstar*gSph.Tstar;
        fprintf(fp,
                "Luminosity of central source: %e Lsun (T=%f K, r=%12.5e cm)\n",
                Lum, gSph.Tstar, gSph.Rad[0]*gSph.Rmax);
        mcp.Lsun = Lum;
    } else {
        mcp.Lsun = 0.0;
    }
          
    fprintf(fp, "\n------------------------------------------------------");
    fprintf(fp, "-----------------------------------------\n");
    fprintf(fp, " n       Rad       Tkin      nH2       Xmol       Vturb");
    fprintf(fp, "      N[H2]       N[mol]    Vrad    Nang\n");
    fprintf(fp, "                    K       cm-3                   km/s");
    fprintf(fp, "      cm-2        cm-2      km/s\n");
    fprintf(fp, "------------------------------------------------------");
    fprintf(fp, "-----------------------------------------\n");
    for (n=1; n<mcp.Nvol; n++) {
        r = (gSph.Rad[n] + gSph.Rad[n-1])/2.0;
        if (n == 1 && gSph.Rmin == 0.0) r = 0.0;
        H2mass += (pow(gSph.Rad[n], 3.0) - pow(gSph.Rad[n-1], 3.0))*
            pow(MH2_CNST*gSph.Rmax, 3.0)*vol[n].nH2;
        NH2  += (gSph.Rad[n] - gSph.Rad[n-1])*gSph.Rmax*vol[n].nH2;
        Nmol += (gSph.Rad[n] - gSph.Rad[n-1])*gSph.Rmax*vol[n].nH2*vol[n].Xmol;
        fprintf(fp,
                "%3d  %5.3f-%5.3f %7.2f %10.3e  %10.3e  %7.3e %10.3e %10.3e %7.2f %5d\n",
                n, gSph.Rad[n-1], gSph.Rad[n], vol[n].Tkin, vol[n].nH2,
                vol[n].Xmol, vol[n].Vturb, NH2, Nmol, VelRad(r), grid[n][0][0].Nang);
    }
    fprintf(fp, "------------------------------------------------------");
    fprintf(fp, "-----------------------------------------\n\n");
    
    mcp.Mdot = pow(MDOT_CNST*gSph.Rmax, 2.0) * vol[mcp.Nvol-1].nH2 * gSph.Vrad0;
  
    if (gSph.Vrad0 != 0.0)
        fprintf(fp, "H2-mass=%e Msun\n\n", H2mass);
    else
        fprintf(fp, "H2-mass=%e Msun   Mdot=%e Msun/yr\n\n", H2mass, mcp.Mdot);
        
    fprintf(fp, "Hermite integration points and weights:\n");
    fprintf(fp, "-----------------------------------------\n");
    for (n=0; n<mcp.Nvel; n++) {
        fprintf(fp, "Vel[%2d]/Vdop = %12.7f w=%12.5e\n",
                n, Hz.z[n], Hz.w[n]);
        wSum += Hz.w[n];
    }
    fprintf(fp, "-----------------------------------------\n");
    fprintf(fp, "Sum of Hermite weights: %f (should be very close to 1)\n\n",
            wSum);
  
    if (mcp.dust) {
        fprintf(fp,
                "Dust component is included: 3 Q(250mu)/4\\rho=%f cm-2/gm\n",
                mcp.dust);
        fprintf(fp, "------------------------------------------------------");
        fprintf(fp, "--------------------------\n\n");
        fprintf(fp, " n       Rad       Tdust     Rgd    beta  Tau[100um]\n");
        fprintf(fp, "                     K\n");
        fprintf(fp, "------------------------------------------------------");
        fprintf(fp, "--------------------------\n\n");
        /* Freq at 100 um: f_GHz = CLIGHT/1.0e3/100.0 */
        for (n=1; n<mcp.Nvol; n++) {
            f = pow(CLIGHT/1.0e3/100.0/rTra[0].Fre,  vol[n].beta);
            TauDust += (gSph.Rad[n] - gSph.Rad[n-1])*rTra[0].dconst[n]*f;
            fprintf(fp, "%3d  %5.3f-%5.3f %7.2f %7.2f %7.3f  %10.3e\n",
                    n, gSph.Rad[n-1], gSph.Rad[n], vol[n].Tdust, vol[n].Rgd,
                    vol[n].beta, TauDust);
        }
        fprintf(fp, "------------------------------------------------------");
        fprintf(fp, "--------------------------\n\n");
    } else
        fprintf(fp, "Dust component is not included.\n\n");
  
    fprintf(fp, "No of energy levels: %d\n", mcp.Nlevel);
    fprintf(fp, "No of radiative transitions: %d\n", mcp.NrTrans);
    fprintf(fp, "No of collisional transitions: %d\n", mcp.NcTrans);
  
    fprintf(fp, "\n-----------------------------------------------------------\n");
    fprintf(fp, "  n    u -> l      f           gu/gl     Aul      Eu\n");
    fprintf(fp, "                                         1/s       K\n");
    fprintf(fp, "-----------------------------------------------------------\n");
    for (n=0; n<mcp.NrTrans; n++) {
        f = rTra[n].Fre;
        if (f < 1000.0) {
            fprintf(fp, "%3d  %3d  %3d %9.5f GHz  %7.3f %10.3e %10.3e\n", n, rTra[n].u, rTra[n].l,
                    f, rTra[n].Gul, rTra[n].Aul, lev[rTra[n].u].Egy);
        } else {
            f = CLIGHT/1.0e3/f;
            fprintf(fp, "%3d  %3d  %3d %9.5f um   %7.3f %10.3e %10.3e\n", n, rTra[n].u, rTra[n].l,
                    f, rTra[n].Gul, rTra[n].Aul, lev[rTra[n].u].Egy);
        }
    }
    fprintf(fp, "-----------------------------------------------------------\n\n");
  
    if (mcp.overlap) {
        fprintf(fp, "Line overlap option is on. Resulting overlaps:\n");
        for (n=0; n<mcp.NrTrans; n++) {
            fprintf(fp, "%3d:", n);
            for (m=0; m<rTra[n].n; m++) {
                fprintf(fp, " %3d", rTra[n].m[m]);
            }
            if (rTra[n].n == 0) {
                fprintf(fp, " none\n");
            } else {
                fprintf(fp, "\n");
            }
        }
        fprintf(fp, "\n");
    }
  
    fp = logs.pfp;
  
    fprintf(fp, "%s\n", mcp.Molecule);
  
    fprintf(fp, "%d %d %d %e %e\n",
            mcp.NrTrans, mcp.Nlevel, mcp.Nvol, mcp.passBand, gSph.Rmax);
          
    fprintf(fp, "%e %e\n", mcp.Tbg, gSph.Tstar);
  
    fprintf(fp, "%e %e %e\n", gSph.Vrad0, gSph.Vrad1, gSph.Vrad2);
   
    for (n=0; n<mcp.Nvol; n++) {
        if (n==0) {
            fprintf(fp, "%e", gSph.Rad[n]);
        } else {
            fprintf(fp, " %e", gSph.Rad[n]);
        }
    }
    fprintf(fp, "\n");
 
    for (n=0; n<mcp.Nvol; n++) {
        if (n==0) {
            fprintf(fp, "%e", vol[n].Vdop);
        } else {
            fprintf(fp, " %e", vol[n].Vdop);
        }
    }
    fprintf(fp, "\n");
  
    for (n=0; n<mcp.NrTrans; n++) {
        fprintf(fp, "%d %d %e %e %e %e\n", rTra[n].l,  rTra[n].u,
                rTra[n].aconst, rTra[n].st, rTra[n].Gul, rTra[n].Fre);
    }
  
    fprintf(fp, "%d\n", mcp.overlap);
  
    if (mcp.overlap) {
        for (n=0; n<mcp.NrTrans; n++) {
            fprintf(fp, "%d %e %e\n", rTra[n].n, rTra[n].v_min, rTra[n].v_max);
            for (m=0; m<rTra[n].n; m++) {
                fprintf(fp, "%d %e\n", rTra[n].m[m], rTra[n].dv[m]);
            }
        }
    }
  
    fprintf(fp, "%f\n", mcp.dust);
  
    if (mcp.dust) {
        for (n=0; n<mcp.Nvol; n++) {
            if (n==0)
                fprintf(fp, "%e", vol[n].Tdust);
            else
                fprintf(fp, " %e", vol[n].Tdust);
        }
        fprintf(fp, "\n");
        for (m=0; m<mcp.NrTrans; m++) {
            for (n=0; n<mcp.Nvol; n++) {
                if (n==0)
                    fprintf(fp, "%e", rTra[m].dconst[n]);
                else
                    fprintf(fp, " %e", rTra[m].dconst[n]);
            }
            fprintf(fp, "\n");
        }
    }
}

void d_fprintf(FILE *fp, char *format, double *d, int size)
{
    int n;
  
    for (n=0; n<size; n++) fprintf(fp, format, d[n]);
  
    fprintf(fp, "\n");
}

void m_fprintf(FILE *fp, char *format, double *d, int size)
{
    int i, j;
  
    for (i=0; i<size; i++) {
        for (j=0; j<size; j++) {
            fprintf(fp, format, *d);
            d++;
        }
        fprintf(fp, "\n");
    }
}

void a_fprintf(FILE *fp, char *format, double **a, int size)
{
    int i, j;
  
    for (i=0; i<size; i++) {
        for (j=0; j<size; j++) {
            fprintf(fp, format, a[i][j]);
        }
        fprintf(fp, "\n");
    }
}

void i_fprintf(FILE *fp, char *format, int *i, int size)
{
    int n;
  
    for (n=0; n<size; n++) fprintf(fp, format, i[n]);
  
    fprintf(fp, "\n");
}
