#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#ifndef MCP_EXTERNAL
#define MCP_EXTERNAL
#endif

#include "ali.h"
#include "init.h"
#include "dat.h"

CTRANS 		cTra[MAX_CTRANS];

void init(char *molecule)
{
  void read_input_file();
  void read_dat_file();
  void init_dat_data();
  void pop_init();
  void col_init();
  void tra_init();
  void overlap_init();
  void weight_init(), grid_init(), hermite_init(int);
  void init_log_info();
  void ReadPopulations(char *);
  
  strcpy(mcp.Molecule, molecule);
  
  strcpy(logs.file, molecule);
  strcat(logs.file, LOG_SUFFIX);
  
  logs.fp = fopen(logs.file, "w");
  
  if (!logs.fp) {
    fprintf(stderr, "Fatal error: Cannot open '%s'\n", logs.file);
    exit(1);
  }
  
  strcpy(logs.prf, molecule);
  strcat(logs.prf, PRF_SUFFIX);
  
  logs.pfp = fopen(logs.prf, "w");
  
  if (!logs.pfp) {
    fprintf(stderr, "Fatal error: Cannot open '%s'\n", logs.prf);
    exit(1);
  }
  
  read_input_file();
  
  read_dat_file();
  init_dat_data();
  
  if (mcp.TexGuess < 0.0) {
      ReadPopulations(mcp.popFile);
  } else {
      pop_init(mcp.TexGuess);
  }
  
  col_init();
  
  tra_init();
  
  if (mcp.overlap) overlap_init();
  
  weight_init();
  
  hermite_init(mcp.Nvel);
  
  grid_init();
  
  init_log_info();
}

void read_input_file()
{
  int n=0, i;
  char str[LEN_ROW], InputFile[LEN_FILENAME], type[20];
  double dv[MAX_VOLUME], r;
  FILE *ifp;
  
  int is_comment();
  void get_dvector(), get_ivector(), init_rad();
  double Trad(), Nrad(), Xrad(), Xexp(), Xexp1(), Wrad();
#ifdef COOLING
  void ReadTkin(char *);
#endif
  
  strcpy(InputFile, mcp.Molecule);
  strcat(InputFile, INP_SUFFIX);
  
  ifp = fopen(InputFile, "r");
  
  if (!ifp) {
    fprintf(stderr, "Fatal error: Cannot open file '%s'\n", InputFile);
    exit(1);
  }
  
  while (fgets(str, LEN_ROW, ifp) != NULL) {
    if (is_comment(str)) continue;
    n++;
    if (n==1) {
      sscanf(str, "%d %d %d %d %d %lf %lf %d", &mcp.Niter, &mcp.Nlambda,
             &mcp.Nng, &mcp.Nng1, &mcp.NgOrder, &mcp.RelLimit, &mcp.Gamma,
             &mcp.Cooling);
    } else if (n==2) {
      sscanf(str, "%s", mcp.datName);
    } else if (n==3) {
      sscanf(str, "%s", mcp.popFile);
    } else if (n==4) {
      sscanf(str, "%lf %lf %lf %lf %d %lf",
             &mcp.passBand, &mcp.TexGuess, &mcp.Tbg, &mcp.Elimit,
             &mcp.overlap, &mcp.dust);
    } else if (n==5) {
      sscanf(str, "%d %d %d", &mcp.geometry, &mcp.TrInfo, &mcp.LimitLevel);
      switch (mcp.geometry) {
        case GEO_SPHERE:
        case GEO_CUBE:
        case GEO_CLUMPS:
          break;
        default:
          fprintf(stderr, "Warning: Invalid geometry parameter (%d)\n", mcp.geometry);
          fprintf(stderr, "         Assuming spherical geometry.\n");
          mcp.geometry = GEO_SPHERE;
          break;
      }
    } else if (n==6 && mcp.geometry==GEO_SPHERE) {
      sscanf(str, "%d %lf %lf %d %d",
             &gSph.Nshell, &gSph.Rmax, &gSph.Tstar, &mcp.Nang, &mcp.Nvel);
      if (gSph.Nshell > MAX_VOLUME) {
        fprintf(stderr, "Fatal error: Nshell (%d) > MAX_VOLUME (%d)\n",
                gSph.Nshell, MAX_VOLUME);
        exit(1);
      }
      if (mcp.Nang % 2  == 1) {
        fprintf(stderr, "Warning: Nang (%d) must be even. Using %d.\n",
                mcp.Nang, ++mcp.Nang);
      }
      if (mcp.Nang >= MAX_ANGLES) {
        fprintf(stderr, "Fatal error: Nang (%d) >= MAX_ANGLES (%d)\n",
                mcp.Nang, MAX_ANGLES);
        exit(1);
      }
      /* if (mcp.Nvel % 2  == 1) {
        fprintf(stderr, "Warning: Nvel (%d) must be even. Using %d.\n",
                mcp.Nvel, ++mcp.Nvel);
      } */
      if (mcp.Nvel > MAX_VEL) {
        fprintf(stderr, "Fatal error: Nvel (%d) > MAX_VEL (%d)\n",
                mcp.Nvel, MAX_VEL);
        exit(1);
      }
      mcp.Nvol = gSph.Nshell;
    } else if (n==7 && mcp.geometry==GEO_SPHERE) {
      if (str[0] == 'F') {
        if (sscanf(str, "F LOG %lf", &gSph.Rmin) == 1) {
          init_rad("log", gSph.Rad, gSph.Rmin, mcp.Nvol, MAX_VOLUME);
        } else if (sscanf(str, "F VOL %lf", &gSph.Rmin) == 1) {
          init_rad("vol", gSph.Rad, gSph.Rmin, mcp.Nvol, MAX_VOLUME);
        } else if (sscanf(str, "F LIN %lf", &gSph.Rmin) == 1) {
          init_rad("lin", gSph.Rad, gSph.Rmin, mcp.Nvol, MAX_VOLUME);
        } else {
          sscanf(str, "F %s %lf", type, &gSph.Rmin);
          init_rad(type, gSph.Rad, gSph.Rmin, mcp.Nvol, MAX_VOLUME);
        }
      } else {
        get_dvector(str, gSph.Rad, MAX_VOLUME);
        gSph.Rmin = gSph.Rad[0];
      }
    } else if (n==8 && mcp.geometry==GEO_SPHERE) {
      sscanf(str, "%lf %lf %lf", &gSph.Vrad0, &gSph.Vrad1, &gSph.Vrad2);
    } else if (n==9) {
      if (str[0] == 'F') {
        sscanf(str, "F %lf %lf %lf", &gSph.Trad0, &gSph.Trad1, &gSph.Trad2);
        for (i=1; i<mcp.Nvol; i++) {
          r = (gSph.Rad[i] + gSph.Rad[i-1])/2.0;
          if (i == 1 && gSph.Rmin == 0.0) r = 0.0;
          vol[i].Tkin = Trad(r);
        }
#ifdef COOLING
      } else if (str[0] == '.') {
        ReadTkin(str);
#endif
      } else {
        get_dvector(str, dv, MAX_VOLUME);
        for (i=0; i<mcp.Nvol; i++) vol[i].Tkin = dv[i];
      }
    } else if (n==10) {
      if (str[0] == 'F') {
        sscanf(str, "F %lf %lf %lf", &gSph.Nrad0, &gSph.Nrad1, &gSph.Nrad2);
        gSph.Nrad3 = 0.0;
        for (i=1; i<mcp.Nvol; i++) {
          r = (gSph.Rad[i] + gSph.Rad[i-1])/2.0;
          if (i == 1 && gSph.Rmin == 0.0) r = 0.0;
          vol[i].nH2 = Nrad(r);
        }
      } else if (str[0] == 'f') {
        sscanf(str, "f %lf %lf %lf %lf",
               &gSph.Nrad0, &gSph.Nrad1, &gSph.Nrad2, &gSph.Nrad3);
        for (i=1; i<mcp.Nvol; i++) {
          r = (gSph.Rad[i] + gSph.Rad[i-1])/2.0;
          if (i == 1 && gSph.Rmin == 0.0) r = 0.0;
          vol[i].nH2 = Nrad(r);
        }
      } else {
        get_dvector(str, dv, MAX_VOLUME);
        for (i=0; i<mcp.Nvol; i++) vol[i].nH2 = dv[i];
      }
    } else if (n==11) {
      if (str[0] == 'F') {
        sscanf(str, "F %lf %lf %lf", &gSph.Xrad0, &gSph.Xrad1, &gSph.Xrad2);
        gSph.Xrad3 = 0.0;
        for (i=1; i<mcp.Nvol; i++) {
          r = (gSph.Rad[i] + gSph.Rad[i-1])/2.0;
          if (i == 1 && gSph.Rmin == 0.0) r = 0.0;
          vol[i].Xmol = Xrad(r);
        }
      } else if (str[0] == 'M') {
        sscanf(str, "M %lf %lf %lf %lf",
               &gSph.Xrad0, &gSph.Xrad1, &gSph.Xrad2, &gSph.Xrad3);
        for (i=1; i<mcp.Nvol; i++) {
          r = (gSph.Rad[i] + gSph.Rad[i-1])/2.0;
          if (i == 1 && gSph.Rmin == 0.0) r = 0.0;
          vol[i].Xmol = Xexp1(r);
        }
      } else if (str[0] == 'E') {
        sscanf(str, "E %lf %lf %lf", &gSph.Xrad0, &gSph.Xrad1, &gSph.Xrad2);
        gSph.Xrad3 = 0.0;
        for (i=1; i<mcp.Nvol; i++) {
          r = (gSph.Rad[i] + gSph.Rad[i-1])/2.0;
          if (i == 1 && gSph.Rmin == 0.0) r = 0.0;
          vol[i].Xmol = Xexp(r);
        }
      } else if (str[0] == 'e') {
        sscanf(str, "e %lf %lf %lf %lf",
               &gSph.Xrad0, &gSph.Xrad1, &gSph.Xrad2, &gSph.Xrad3);
        for (i=1; i<mcp.Nvol; i++) {
          r = (gSph.Rad[i] + gSph.Rad[i-1])/2.0;
          if (i == 1 && gSph.Rmin == 0.0) r = 0.0;
          vol[i].Xmol = Xexp(r);
        }
      } else {
        get_dvector(str, dv, MAX_VOLUME);
        for (i=0; i<mcp.Nvol; i++) vol[i].Xmol = dv[i];
      }
    } else if (n==12) {
      if (str[0] == 'F') {
        sscanf(str, "F %lf %lf %lf", &gSph.Wrad0, &gSph.Wrad1, &gSph.Wrad2);
        for (i=1; i<mcp.Nvol; i++) {
          r = (gSph.Rad[i] + gSph.Rad[i-1])/2.0;
          if (i == 1 && gSph.Rmin == 0.0) r = 0.0;
          vol[i].Vturb = Wrad(r, 'F');
        }
      } else if (str[0] == 'G') {
        sscanf(str, "G %lf %lf %lf %lf %lf",
               &gSph.Wrad0, &gSph.Wrad1, &gSph.Wp[0],
               &gSph.Wp[1], &gSph.Wp[2]);
        for (i=1; i<mcp.Nvol; i++) {
          r = (gSph.Rad[i] + gSph.Rad[i-1])/2.0;
          if (i == 1 && gSph.Rmin == 0.0) r = 0.0;
          vol[i].Vturb = Wrad(r, 'G');
        }
      } else {
        get_dvector(str, dv, MAX_VOLUME);
        for (i=0; i<mcp.Nvol; i++) vol[i].Vturb = dv[i];
      }
    } else if (n==13) {
      if (str[0] == 'F') {
        sscanf(str, "F %lf %lf %lf", &gSph.Trad0, &gSph.Trad1, &gSph.Trad2);
        for (i=1; i<mcp.Nvol; i++) {
          r = (gSph.Rad[i] + gSph.Rad[i-1])/2.0;
          if (i == 1 && gSph.Rmin == 0.0) r = 0.0;
          vol[i].Tdust = Trad(r);
        }
      } else if (str[0] == 'T') {
        for (i=1; i<mcp.Nvol; i++) vol[i].Tdust = vol[i].Tkin;
      } else {
        get_dvector(str, dv, MAX_VOLUME);
        for (i=0; i<mcp.Nvol; i++) vol[i].Tdust = dv[i];
      }
    } else if (n==14) {
      get_dvector(str, dv, MAX_VOLUME);
      for (i=0; i<mcp.Nvol; i++) vol[i].Rgd = dv[i];
    } else if (n==15) {
      get_dvector(str, dv, MAX_VOLUME);
      for (i=0; i<mcp.Nvol; i++) vol[i].beta = dv[i];
    } else {
      fprintf(stderr, "Skipping line '%s' in file '%s'\n", str, InputFile);
    }
  }
  
  fclose(ifp);
}

void read_dat_file()
{
  int i, j, k, n=0, nn=-1, key=0, nkeys;
  char str[LEN_ROW], DatFile[LEN_FILENAME];
  char key_word[LEN_KEY], tmp[LEN_ROW];
  FILE *dfp;
  
  int is_comment(), is_keyword();
  
  strcpy(DatFile, DAT_DIR);
  strcat(DatFile, mcp.datName);
  strcat(DatFile, DAT_SUFFIX);
  
  dfp = fopen(DatFile, "r");
  
  if (!dfp) {
    fprintf(stderr, "Fatal error: Cannot open file '%s'\n", DatFile);
    exit(1);
  }
  
  i = 0;
  nkeys = sizeof(kw)/sizeof(KW_STRUCT);
  
  while (fgets(str, LEN_ROW, dfp) != NULL) {
    if (is_comment(str)) continue;
    if (is_keyword(str)) {
      sscanf(str, "!%s", key_word);
      n = 0;
      key = K_UNDEF;
      for (k=0; k<nkeys; k++) {
        if (strcmp(key_word, kw[k].label)==0) {
          key = k;
          if (kw[k].type == KW_REAL_ARR) {
            nn++;
          } else {
            nn = -1;
          }
          break;
        }
      }
      if (key == K_UNDEF) {
        fprintf(stderr, "Warning: Undefined keyword: %s\n", key_word);
      }
    } else if (key != K_UNDEF) {
      switch (kw[key].type) {
        case KW_STRING:
          strcpy(kw[key].p.c, str);
          break;
        case KW_INT:
          sscanf(str, "%d", kw[key].p.i);
          break;
        case KW_REAL:
          sscanf(str, "%lf", kw[key].p.d);
          break;
        case KW_INT_VEC:
          for (j=0; j<(int)strlen(str)-1; j += kw[key].len) {
            strncpy(tmp, &str[j], kw[key].len);
            tmp[kw[key].len] = '\0';
            sscanf(tmp, kw[key].form, &kw[key].p.i[n]);
            kw[key].p.i[n]--;
            n++;
          }
          break;
        case KW_REAL_VEC:
          for (j=0; j<(int)strlen(str)-1; j += kw[key].len) {
            strncpy(tmp, &str[j], kw[key].len);
            tmp[kw[key].len] = '\0';
            sscanf(tmp, kw[key].form, &kw[key].p.d[n]);
            n++;
          }
          break;
        case KW_REAL_ARR:
          for (j=0; j<(int)strlen(str)-1; j += kw[key].len) {
            if (strcmp(kw[key].label, "C_COEFFS")==0) {
              sscanf(&str[j], kw[key].form, &c_coeffs[nn][n]);
            } else if (strcmp(kw[key].label, "C1_COEFFS")==0) {
              sscanf(&str[j], kw[key].form, &c1_coeffs[nn][n]);
            }
            n++;
          }
          break;
      }
    }
    i++;
  }
  
  fclose(dfp);
}

void init_dat_data()
{
  int n, nn, u, l, lim=0, m=0;
  
  mcp.molWeight = molw;
  
  if (mcp.Elimit > 0.0) lim = 1;

  if (n_levs > MAX_LEVEL) {
    fprintf(stderr, "Increase MAX_level to %d.\n", n_levs);
    exit(1);
  }
  mcp.Nlevel = n_levs;

  if (n_rtrans > MAX_RTRANS) {
    fprintf(stderr, "Increase MAX_RTRANS to %d.\n", n_rtrans);
    exit(1);
  }
  mcp.NrTrans = n_rtrans;
  
#ifdef DEBUG
  printf("No. of levels: %d   No. of radiative transitions: %d\n",
         n_levs, n_rtrans);
#endif

  if (lim) m = 0;
  for (n=0; n<mcp.Nlevel; n++) {
    if (lim && egy[n]*INVCM_TO_K > mcp.Elimit) continue;
    lev[n].g = g[n];
    lev[n].Egy = egy[n]*INVCM_TO_K;
#ifdef DEBUG
    printf("i=%d  g=%lf  egy=%lf\n", n, lev[n].g, lev[n].Egy);
#endif
    if (lim) m++;
  }
  if (lim) mcp.Nlevel = m;
  
  if (mcp.LimitLevel <= 0 || mcp.LimitLevel > mcp.Nlevel)
      mcp.LimitLevel = mcp.Nlevel;
  
  if (lim) m = 0;
  for (n=0; n<mcp.NrTrans; n++) {
    if (lim && egy[ul[n]]*INVCM_TO_K > mcp.Elimit) continue;
    if (!lim) m = n;
    u = rTra[m].u = ul[n];
    l = rTra[m].l = ll[n];
    if (u >= n_levs || u < 0) {
      printf("WARNING!  RAD_UPPER is corrupted: %d %d\n", n, u);
    }
    if (l >= n_levs || l < 0) {
      printf("WARNING!  RAD_LOWER is corrupted: %d %d\n", n, l);
    }
    rTra[m].Aul = aul[n];
    rTra[m].Gul = lev[u].g/lev[l].g;
    rTra[m].Fre = (lev[u].Egy - lev[l].Egy)*K_TO_GHZ;
#ifdef DEBUG
    printf("%3d ful=%10.3e Aul=%10.3e Gul=%10.3e\n", m,
           rTra[m].Fre, rTra[m].Aul, rTra[m].Gul);
#endif
    if (lim) m++;
  }
  if (lim) mcp.NrTrans = m;

  if (n_ctrans > MAX_CTRANS) {
    fprintf(stderr, "Increase MAX_CTRANS to %d.\n", n_ctrans);
    exit(1);
  }
  mcp.NcTrans = n_ctrans;
  
  if (n_tk > MAX_CTEMP) {
    fprintf(stderr, "Increase MAX_CTEMP to %d.\n", n_tk);
    exit(1);
  }
  mcp.NcTemp = n_tk;
#ifdef DEBUG
  printf("No. of collisional transitions: %d  No. coll. temps: %d\n",
         n_ctrans, n_tk);
#endif

  for (nn=0; nn<mcp.NcTemp; nn++) {
    mcp.cTemp[nn] = c_tk[nn];
  }
  
  if (lim) m = 0;
  for (n=0; n<mcp.NcTrans; n++) {
    if (lim && egy[c_ul[n]]*INVCM_TO_K > mcp.Elimit) continue;
    if (!lim) m = n;
    u = cTra[m].u = c_ul[n];
    l = cTra[m].l = c_ll[n];
    cTra[m].Gul = lev[u].g/lev[l].g;
    cTra[m].Egy = lev[u].Egy - lev[l].Egy;
#ifdef DEBUG
    printf("n=%3d %3d->%3d  Gul=%7.3lf   Ediff=%10.3lf\n", n, u, l,
           cTra[n].Gul, cTra[n].Egy);
#endif
    if (u >= n_levs || u < 0) {
      printf("WARNING!  COL_UPPER is corrupted: %d %d\n", n, u);
    }
    if (l >= n_levs || l < 0) {
      printf("WARNING!  COL_LOWER is corrupted: %d %d\n", n, l);
    }
    for (nn=0; nn<mcp.NcTemp; nn++) {
      cTra[m].Cul[nn] = c_coeffs[nn][n];
    }
    if (lim) m++;
  }
  
  if (lim) mcp.NcTrans = m;
  
  if (lim) {
    printf("Warning: The energy limit of %f K has resulted in\n", mcp.Elimit);
    printf("         No of energy levs = %d (%d)\n", mcp.Nlevel, n_levs);
    printf("         No of rad. trans  = %d (%d)\n", mcp.NrTrans, n_rtrans);
    printf("         No of col. trans  = %d (%d)\n", mcp.NcTrans, n_ctrans);
  }
  
#ifdef DEBUG
  for (n=0; n<mcp.Nvol; n++) {
    printf("%3d Tkin=%6.1lf nh2=%10.3e\n", n, vol[n].Tkin, vol[n].nH2);
  }
#endif
}

int is_comment(char *str)
{
  int i;
  char c;
  
  if ((int)strlen(str) == 0) return 1;
  
  i = 0;
  c = str[0];
  
  while (i < (int)strlen(str)-1 && (c == ' ' || c == '\t') ) {
    i++;
    c = str[i];
  }
  
  if (i >= (int)strlen(str)) return 1;	/* Blank line == comment */
  
  if (c == '%' || c == '#' || c == '*' || c == '\0') return 1;
  
  return 0;
}

int is_keyword(char *str)
{
  int i;
  char c;
  
  if ((int)strlen(str) == 0) return 0;
  
  i = 0;
  c = str[0];
  
  while (i < (int)strlen(str)-1 && (c == ' ' || c == '\t') ) {
    i++;
    c = str[i];
  }
  
  if (i >= (int)strlen(str)) return 0;	/* Blank line == comment */
  
  if (c == '!') return 1;
  
  return 0;
}

void get_dvector(char *str, double *d, int size)
{
  int i=0, n;
  char *delims = " ";
  char *token;
  
  token = strtok(str, delims);
  
  while (token != NULL && i < size) {
    d[i] = atof(token);
    token = strtok(NULL, delims);
    i++;
  }
  if (i > 0) {
    for (n=i; n<size; n++) d[n] = d[n-1];
  }
}

void get_ivector(char *str, int *d, int size)
{
  int i=0;
  char *delims = " ";
  char *token;
  
  token = strtok(str, delims);
  
  while (token != NULL && i < size) {
    d[i] = atoi(token);
    token = strtok(NULL, delims);
    i++;
  }
}

void init_log_info()
{
  int n, m;
  double f, r, NH2=0.0, Nmol=0.0, H2mass=0.0, Mdot=0.0, TauDust=0.0, wSum=0.0;
  FILE *fp;
  
  void i_fprintf(), d_fprintf();
  double VelRad(double);
  
  fp = logs.fp;
  
  if (!fp) return;
  
  fprintf(fp, "Molecule: %s (%5.1f)\n", mcp.Molecule, mcp.molWeight);
  
  fprintf(fp, "\nMax iter=%3d  Accel. i=%d  Ng-accel i=%d  Conv. limit=%10.3e\n",
          mcp.Niter, mcp.Nlambda, mcp.Nng, mcp.RelLimit);
  
  fprintf(fp, "\nGamma=%7.1f  Passband=%7.2f km/s  TexGuess=%7.2f K   Tbg=%8.3f K\n",
          mcp.Gamma, mcp.passBand, mcp.TexGuess, mcp.Tbg);
          
  fprintf(fp, "\nNAng=%3d NVel=%3d  NVol=%3d  Rmax=%12.5e cm  Tstar=%7.1f K\n",
          mcp.Nang, mcp.Nvel, mcp.Nvol, gSph.Rmax, gSph.Tstar);
          
  fprintf(fp, "\n------------------------------------------------------");
  fprintf(fp, "-----------------------------------------\n");
  fprintf(fp, " n       Rad       Tkin      nH2       Xmol       Vturb");
  fprintf(fp, "      N[H2]       N[mol]    Vrad    Nang\n");
  fprintf(fp, "                    K       cm-3                   km/s");
  fprintf(fp, "      cm-2        cm-2      km/s\n");
  fprintf(fp, "------------------------------------------------------");
  fprintf(fp, "-----------------------------------------\n");
  for (n=1; n<mcp.Nvol; n++) {
     r = (gSph.Rad[n] + gSph.Rad[n-1])/2.0;
     if (n == 1 && gSph.Rmin == 0.0) r = 0.0;
     H2mass += (pow(gSph.Rad[n], 3.0) - pow(gSph.Rad[n-1], 3.0))*
               pow(MH2_CNST*gSph.Rmax, 3.0)*vol[n].nH2;
     NH2  += (gSph.Rad[n] - gSph.Rad[n-1])*gSph.Rmax*vol[n].nH2;
     Nmol += (gSph.Rad[n] - gSph.Rad[n-1])*gSph.Rmax*vol[n].nH2*vol[n].Xmol;
     fprintf(fp,
     "%3d  %5.3f-%5.3f %7.2f %10.3e  %10.3e  %7.3e %10.3e %10.3e %7.2f %5d\n",
             n, gSph.Rad[n-1], gSph.Rad[n], vol[n].Tkin, vol[n].nH2,
             vol[n].Xmol, vol[n].Vturb, NH2, Nmol, VelRad(r), grid[n][0].Nang);
  }
  fprintf(fp, "------------------------------------------------------");
  fprintf(fp, "-----------------------------------------\n\n");
  
  Mdot = pow(MDOT_CNST*gSph.Rmax, 2.0) * vol[mcp.Nvol-1].nH2 * gSph.Vrad0;
  
  if (gSph.Vrad1 != 0.0)
      fprintf(fp, "H2-mass=%e Msun\n\n", H2mass);
  else
      fprintf(fp, "H2-mass=%e Msun   Mdot=%e Msun/yr\n\n", H2mass, Mdot);
        
  fprintf(fp, "Hermite integration points and weights:\n");
  fprintf(fp, "-----------------------------------------\n");
  for (n=0; n<mcp.Nvel; n++) {
      fprintf(fp, "Vel[%2d]/Vdop = %12.7f w=%12.5e\n",
              n, Hz.z[n], Hz.w[n]);
      wSum += Hz.w[n];
  }
  fprintf(fp, "-----------------------------------------\n");
  fprintf(fp, "Sum of Hermite weights: %f (should be very close to 1)\n\n",
          wSum);
  
  if (mcp.dust) {
      fprintf(fp,
              "Dust component is included: 3 Q(250mu)/4\\rho=%f cm-2/gm\n",
              mcp.dust);
      fprintf(fp, "------------------------------------------------------");
      fprintf(fp, "--------------------------\n\n");
      fprintf(fp, " n       Rad       Tdust     Rgd    beta  Tau[100um]\n");
      fprintf(fp, "                     K\n");
      fprintf(fp, "------------------------------------------------------");
      fprintf(fp, "--------------------------\n\n");
/* Freq at 100 um: f_GHz = CLIGHT/1.0e3/100.0 */
      for (n=1; n<mcp.Nvol; n++) {
          f = pow(CLIGHT/1.0e3/100.0/rTra[0].Fre,  vol[n].beta);
          TauDust += (gSph.Rad[n] - gSph.Rad[n-1])*rTra[0].dconst[n]*f;
          fprintf(fp, "%3d  %5.3f-%5.3f %7.2f %7.2f %7.3f  %10.3e\n",
                  n, gSph.Rad[n-1], gSph.Rad[n], vol[n].Tdust, vol[n].Rgd,
                  vol[n].beta, TauDust);
      }
      fprintf(fp, "------------------------------------------------------");
      fprintf(fp, "--------------------------\n\n");
  } else
      fprintf(fp, "Dust component is not included.\n\n");
  
  fprintf(fp, "No of energy levels: %d\n", mcp.Nlevel);
  fprintf(fp, "No of radiative transitions: %d\n", mcp.NrTrans);
  fprintf(fp, "No of collisional transitions: %d\n", mcp.NcTrans);
  
  fprintf(fp, "\n-----------------------------------------------------------\n");
  fprintf(fp, "  n    u -> l      f           gu/gl     Aul      Eu\n");
  fprintf(fp, "                                         1/s       K\n");
  fprintf(fp, "-----------------------------------------------------------\n");
  for (n=0; n<mcp.NrTrans; n++) {
    f = rTra[n].Fre;
    if (f < 1000.0) {
      fprintf(fp, "%3d  %3d  %3d %9.5f GHz  %7.3f %10.3e %10.3e\n", n, rTra[n].u, rTra[n].l,
              f, rTra[n].Gul, rTra[n].Aul, lev[rTra[n].u].Egy);
    } else {
      f = CLIGHT/1.0e3/f;
      fprintf(fp, "%3d  %3d  %3d %9.5f um   %7.3f %10.3e %10.3e\n", n, rTra[n].u, rTra[n].l,
              f, rTra[n].Gul, rTra[n].Aul, lev[rTra[n].u].Egy);
    }
  }
  fprintf(fp, "-----------------------------------------------------------\n\n");
  
  if (mcp.overlap) {
    fprintf(fp, "Line overlap option is on. Resulting overlaps:\n");
    for (n=0; n<mcp.NrTrans; n++) {
      fprintf(fp, "%3d:", n);
      for (m=0; m<rTra[n].n; m++) {
        fprintf(fp, " %3d", rTra[n].m[m]);
      }
      if (rTra[n].n == 0) {
        fprintf(fp, " none\n");
      } else {
        fprintf(fp, "\n");
      }
    }
    fprintf(fp, "\n");
  }
  
  fp = logs.pfp;
  
  fprintf(fp, "%s\n", mcp.Molecule);
  
  fprintf(fp, "%d %d %d %e %e\n",
          mcp.NrTrans, mcp.Nlevel, mcp.Nvol, mcp.passBand, gSph.Rmax);
          
  fprintf(fp, "%e %e\n", mcp.Tbg, gSph.Tstar);
  
  fprintf(fp, "%e %e %e\n", gSph.Vrad0, gSph.Vrad1, gSph.Vrad2);
   
  for (n=0; n<mcp.Nvol; n++) {
    if (n==0) {
      fprintf(fp, "%e", gSph.Rad[n]);
    } else {
      fprintf(fp, " %e", gSph.Rad[n]);
    }
  }
  fprintf(fp, "\n");
 
  for (n=0; n<mcp.Nvol; n++) {
    if (n==0) {
      fprintf(fp, "%e", vol[n].Vdop);
    } else {
      fprintf(fp, " %e", vol[n].Vdop);
    }
  }
  fprintf(fp, "\n");
  
  for (n=0; n<mcp.NrTrans; n++) {
    fprintf(fp, "%d %d %e %e %e %e\n", rTra[n].l,  rTra[n].u,
            rTra[n].aconst, rTra[n].st, rTra[n].Gul, rTra[n].Fre);
  }
  
  fprintf(fp, "%d\n", mcp.overlap);
  
  if (mcp.overlap) {
    for (n=0; n<mcp.NrTrans; n++) {
      fprintf(fp, "%d %e %e\n", rTra[n].n, rTra[n].v_min, rTra[n].v_max);
      for (m=0; m<rTra[n].n; m++) {
        fprintf(fp, "%d %e\n", rTra[n].m[m], rTra[n].dv[m]);
      }
    }
  }
  
  fprintf(fp, "%f\n", mcp.dust);
  
  if (mcp.dust) {
    for (n=0; n<mcp.Nvol; n++) {
      if (n==0)
        fprintf(fp, "%e", vol[n].Tdust);
      else
        fprintf(fp, " %e", vol[n].Tdust);
    }
    fprintf(fp, "\n");
    for (m=0; m<mcp.NrTrans; m++) {
      for (n=0; n<mcp.Nvol; n++) {
        if (n==0)
          fprintf(fp, "%e", rTra[m].dconst[n]);
        else
          fprintf(fp, " %e", rTra[m].dconst[n]);
      }
      fprintf(fp, "\n");
    }
  }
}

void d_fprintf(FILE *fp, char *format, double *d, int size)
{
  int n;
  
  for (n=0; n<size; n++) fprintf(fp, format, d[n]);
  
  fprintf(fp, "\n");
}

void m_fprintf(FILE *fp, char *format, double *d, int size)
{
  int i, j;
  
  for (i=0; i<size; i++) {
    for (j=0; j<size; j++) {
      fprintf(fp, format, *d);
      d++;
    }
    fprintf(fp, "\n");
  }
}

void a_fprintf(FILE *fp, char *format, double **a, int size)
{
  int i, j;
  
  for (i=0; i<size; i++) {
    for (j=0; j<size; j++) {
      fprintf(fp, format, a[i][j]);
    }
    fprintf(fp, "\n");
  }
}

void i_fprintf(FILE *fp, char *format, int *i, int size)
{
  int n;
  
  for (n=0; n<size; n++) fprintf(fp, format, i[n]);
  
  fprintf(fp, "\n");
}
