#include <math.h>

#include "vector.h"

CARTESIAN
S_to_C(s)
SPHERICAL s;
{
  static CARTESIAN c;

  c.x = s.r*sin(s.theta)*cos(s.phi);
  c.y = s.r*sin(s.theta)*sin(s.phi);
  c.z = s.r*cos(s.theta);

  return c;
}

SPHERICAL
C_to_S(c)
CARTESIAN c;
{
  double r2;
  static SPHERICAL s;

  r2 = c.x*c.x + c.y*c.y + c.z*c.z;

  if (r2 == 0.0) {
    s.r = 0.0;
    s.theta = 0.0;
    s.phi = 0.0;
    return s;
  }

  s.r = sqrt(r2);
  s.theta = acos(c.z / s.r);
  s.phi = atan2(c.y, c.x);

  return s;
}

SPHERICAL
P_to_S(p)
POLAR p;
{
  static SPHERICAL s;

  s.phi = p.phi;
  s.r = sqrt(p.r*p.r + p.z*p.z);
  if (s.r > 0.0)
    s.theta = acos(p.z/s.r);
  else
    s.theta = 0.0;

  return s;
}

POLAR
S_to_P(s)
SPHERICAL s;
{
  static POLAR p;

  p.phi = s.phi;
  p.z = s.r * cos(s.theta);
  p.r = sqrt(s.r*s.r - p.z*p.z);

  return p;
}

CARTESIAN
P_to_C(p)
POLAR p;
{
  static CARTESIAN c;

  c.x = p.r*cos(p.phi);
  c.y = p.r*sin(p.phi);
  c.z = p.z;

  return c;
}

POLAR
C_to_P(c)
CARTESIAN c;
{
  static POLAR p;

  p.r = sqrt(c.x*c.x + c.y*c.y);
  if (p.r == 0.0) {
    p.phi = 0.0;
  } else {
    p.phi = atan2(c.y, c.x);
  }
  p.z = c.z;

  return p;
}

/* Here starts Cartesian functions */

CARTESIAN
C_neg(c)
CARTESIAN c;
{
  static CARTESIAN a;

  a.x = -c.x;
  a.y = -c.y;
  a.z = -c.z;

  return a;
}

CARTESIAN
C_line(o, v, t)
CARTESIAN o, v;
double t;
{
  static CARTESIAN c;

  c.x = o.x + t*v.x;
  c.y = o.y + t*v.y;
  c.z = o.z + t*v.z;

  return c;
}

CARTESIAN
C_add(c1, c2)
CARTESIAN c1, c2;
{
  static CARTESIAN c;

  c.x = c1.x + c2.x;
  c.y = c1.y + c2.y;
  c.z = c1.z + c2.z;

  return c;
}

CARTESIAN
C_sub(c1, c2)
CARTESIAN c1, c2;
{
  static CARTESIAN c;

  c.x = c1.x - c2.x;
  c.y = c1.y - c2.y;
  c.z = c1.z - c2.z;

  return c;
}

double
C_dot(c1, c2)
CARTESIAN c1, c2;
{
  return (c1.x*c2.x + c1.y*c2.y + c1.z*c2.z);
}

CARTESIAN
C_cross(a, b)
CARTESIAN a, b;
{
  static CARTESIAN c;

  c.x = a.y*b.z - a.z*b.y;
  c.y = a.z*b.x - a.x*b.z;
  c.z = a.x*b.y - a.y*b.x;

  return c;
}

double
C_magn(c)
CARTESIAN c;
{
  return sqrt(c.x*c.x + c.y*c.y + c.z*c.z);
}

CARTESIAN
C_norm(c)
CARTESIAN c;
{
  double r;
  static CARTESIAN a;

  r = sqrt(c.x*c.x + c.y*c.y + c.z*c.z);

  if (r != 0.0) {
    a.x = c.x/r;
    a.y = c.y/r;
    a.z = c.z/r;
  }

  return a;
}

/* Here starts spherical functions */

SPHERICAL
S_neg(s)
SPHERICAL s;
{
  static SPHERICAL a;
  CARTESIAN c;

  c = S_to_C(s);
  c = C_neg(c);
  a = C_to_S(c);

  return a;
}

SPHERICAL
S_add(s1, s2)
SPHERICAL s1, s2;
{
  static SPHERICAL s;
  CARTESIAN c1, c2;

  c1 = S_to_C(s1);
  c2 = S_to_C(s2);

  c1 = C_add(c1, c2);

  s = C_to_S(c1);

  return s;
}

SPHERICAL
S_sub(s1, s2)
SPHERICAL s1, s2;
{
  static SPHERICAL s;
  CARTESIAN c1, c2;

  c1 = S_to_C(s1);
  c2 = S_to_C(s2);

  c1 = C_sub(c1, c2);

  s = C_to_S(c1);

  return s;
}

double
S_dot(s1, s2)
SPHERICAL s1, s2;
{
  double d;
  CARTESIAN c1, c2;

  c1 = S_to_C(s1);
  c2 = S_to_C(s2);

  d = C_dot(c1, c2);

  return d;
}

SPHERICAL
S_cross(s1, s2)
SPHERICAL s1, s2;
{
  static SPHERICAL s;
  CARTESIAN c1, c2;

  c1 = S_to_C(s1);
  c2 = S_to_C(s2);

  c1 = C_cross(c1, c2);

  s = C_to_S(c1);

  return s;
}

double
S_magn(s)
SPHERICAL s;
{
  double d;

  d = s.r;

  return d;
}

SPHERICAL
S_norm(s)
SPHERICAL s;
{
  static SPHERICAL a;

  a = s;
  a.r = 1.0;

  return a;
}

/* Here starts cylindrical functions */

POLAR
P_neg(p)
POLAR p;
{
  static POLAR a;
  CARTESIAN c;

  c = P_to_C(p);
  c = C_neg(c);
  a = C_to_P(c);

  return a;
}

POLAR
P_add(p1, p2)
POLAR p1, p2;
{
  static POLAR p;
  CARTESIAN c1, c2;

  c1 = P_to_C(p1);
  c2 = P_to_C(p2);

  c1 = C_add(c1, c2);

  p = C_to_P(c1);

  return p;
}

POLAR
P_sub(p1, p2)
POLAR p1, p2;
{
  static POLAR p;
  CARTESIAN c1, c2;

  c1 = P_to_C(p1);
  c2 = P_to_C(p2);

  c1 = C_sub(c1, c2);

  p = C_to_P(c1);

  return p;
}

double
P_dot(p1, p2)
POLAR p1, p2;
{
  double d;
  CARTESIAN c1, c2;

  c1 = P_to_C(p1);
  c2 = P_to_C(p2);

  d = C_dot(c1, c2);

  return d;
}

POLAR
P_cross(p1, p2)
POLAR p1, p2;
{
  static POLAR p;
  CARTESIAN c1, c2;

  c1 = P_to_C(p1);
  c2 = P_to_C(p2);

  c1 = C_cross(c1, c2);

  p = C_to_P(c1);

  return p;
}

double
P_magn(p)
POLAR p;
{
  double d;

  d = sqrt(p.r*p.r + p.z*p.z);

  return d;
}

POLAR
P_norm(p)
POLAR p;
{
  double d;
  static POLAR a;

  a = p;
  d = sqrt(p.r*p.r + p.z*p.z);

  if (d != 0.0) {
    a.r /= d;
    a.z /= d;
  }

  return a;
}
