#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#ifndef MCP_EXTERNAL
#define MCP_EXTERNAL
#endif

/* #define DEBUG */

#include "ali.h"

extern CTRANS   cTra[MAX_CTRANS];
extern double   Cul[MAX_VOLUME][MAX_LEVEL][MAX_LEVEL];

void pop_init(double Tex)
{
    int n, l;
    VOLUME *v;
    double s, t = Tex;

#ifdef DEBUG
    printf("TexGuess=%lf\n", t);
#endif

    for (n=0; n<mcp.Nvol; n++) {
        v = &vol[n];
        v->pop[0] = 1.0;
        s = 1.0;
        if (Tex <= 1.0) t = v->Tkin * Tex;
        for (l=1; l<mcp.Nlevel; l++) {
            v->pop[l] = v->pop[0] * lev[l].g/lev[0].g *
                exp((lev[0].Egy-lev[l].Egy)/t);
            s += v->pop[l];
        }
        for (l=0; l<mcp.Nlevel; l++) v->pop[l] *= v->Xmol * v->nH2 / s;
#ifdef DEBUG
        printf("Vol=%3d s=%lf\n", n, s);
        printf("Pop=");
        for (l=0; l<mcp.Nlevel; l++) printf(" %10.3e", v->pop[l]);
        printf("\n");
#endif
    }
}

void col_init()
{
    int n, m;
    int i, l, u;
    VOLUME *v;
    double t, t1, t2;
    double c1, c2;
    double s;
    static double a[MAX_LEVEL][MAX_LEVEL];

    for (n=1; n<mcp.Nvol; n++) {
        v = &vol[n];
        t = v->Tkin;
        m = mcp.NcTemp - 1;
        if (m > 0) {
            while (t < mcp.cTemp[m] && m > 0) m--;
            if (m == mcp.NcTemp - 1) m = mcp.NcTemp-2;
        } else {
            m = 0;
        }
        for (u=0; u<mcp.Nlevel; u++) {
            for (l=0; l<mcp.Nlevel; l++) {
                a[l][u] = 0.0;
            }
        }
        for (i=0; i<mcp.NcTrans; i++) {
            l = cTra[i].l;
            u = cTra[i].u;
            if (mcp.NcTemp == 1) {
                a[l][u] = cTra[i].Cul[m] * sqrt(t/mcp.cTemp[m]);
            } else if (t < mcp.cTemp[0]) {
                a[l][u] = cTra[i].Cul[0] * sqrt(t/mcp.cTemp[0]);
            } else {
                c1 = cTra[i].Cul[m]; c2 = cTra[i].Cul[m+1];
                if (c1 > 0.0 && c2 > 0.0) {
                    c1 = log(c1);      c2 = log(c2);
                    t1 = mcp.cTemp[m]; t2 = mcp.cTemp[m+1];
                    a[l][u] = exp((c2 - c1)*(t - t1)/(t2 - t1) + c1);
                } else {
                    a[l][u] = 0.0;
                }
            }
            a[u][l] = a[l][u] * cTra[i].Gul * exp(-cTra[i].Egy/t);
        }
        for (u=0; u<mcp.Nlevel; u++) {
            a[u][u] = 0.0;
            s = 0.0;
            for (l=0; l<mcp.Nlevel; l++) {
                /* if (l == u) continue; */
                s += a[l][u];
                Cul[n][u][l] = a[u][l] * v->nH2;
            }
            Cul[n][u][u] = -s * v->nH2;
        }
        /* for (l=0; l<mcp.Nlevel; l++) Cul[n][mcp.Nlevel-1][l] = 1.0; */
        v->Vdop = sqrt(0.01663*v->Tkin/mcp.molWeight + v->Vturb * v->Vturb);
    }
}

void tra_init()
{
    int m, n;
    /* int u, l; */
    /* double down; */
    VOLUME *v;

    for (m=0; m<mcp.NrTrans; m++) {
     /* l = rTra[m].l;
        u = rTra[m].u;
        down = rTra[m].Aul;
        for (n=0; n<mcp.Nvol; n++) {
           v = &vol[n];
           Cul[n][l][u] += down;
           if (u == mcp.Nlevel - 1) continue;
           Cul[n][u][u] -= down;
           } */
        rTra[m].n = 0;
    }

    for (n=0; n<mcp.Nvol; n++) {
        v = &vol[n];
        if (n==0) {
            v->volume = 4.0*PI/3.0*pow(gSph.Rad[0], 3.0);
        } else {
            v->volume = 4.0*PI/3.0*(pow(gSph.Rad[n], 3.0) - pow(gSph.Rad[n-1], 3.0));
        }
#ifdef COOLING
        v->Crad = v->CradC = v->CradCC = v->CradPrev = 0.0;
#endif
    }
}

void overlap_init()
{
    int n, m, olap, nlap=0, found, o;
    int unresolved=1;
    double dv, v_min, v_max;

    int get_olap();

    for (n=0; n<mcp.NrTrans; n++) {
        olap = 0;
        v_max = mcp.passBand/2.0;
        v_min = -v_max;
        for (m=0; m<mcp.NrTrans; m++) {
            if (n == m) continue;
            dv = CLIGHT*(rTra[n].Fre - rTra[m].Fre)/rTra[n].Fre/1.0e3;
            if (fabs(dv) < mcp.passBand) {
                if (olap >= MAX_OLAP) {
                    fprintf(stderr, "Warning: Too many overlaps.\n");
                    fprintf(stderr, "         Increase MAX_OLAP (%d).\n", MAX_OLAP);
                    exit(1);
                }
                rTra[n].m[olap] = m;
                rTra[n].dv[olap] = dv;
                if (dv - mcp.passBand/2.0 < v_min) v_min = dv - mcp.passBand/2.0;
                if (dv + mcp.passBand/2.0 > v_max) v_max = dv + mcp.passBand/2.0;
                olap++;
            }
        }
        rTra[n].n = olap;
        rTra[n].v_min = v_min;
        rTra[n].v_max = v_max;
    }

    while (unresolved) {
        unresolved = 0;
        for (n=0; n<mcp.NrTrans; n++) {
            if (rTra[n].n == 0) continue;
            get_olap(-1);
            while ((m = get_olap(n)) >= 0) {
                if (m == n) continue;
                olap = rTra[n].n;
                found = 0;
                for (o=0; o<olap; o++) {
                    if (m == rTra[n].m[o]) {
                        found = 1;
                        break;
                    }
                }
                if (!found) {
                    unresolved = 1;
                    if (olap >= MAX_OLAP) {
                        fprintf(stderr, "Fatal error: olap (%d) > MAX_OLAP (%d)\n",
                                olap, MAX_OLAP);
                        exit(1);
                    }
                    dv = CLIGHT*(rTra[n].Fre - rTra[m].Fre)/rTra[n].Fre/1.0e3;
                    rTra[n].m[olap] = m;
                    rTra[n].dv[olap] = dv;
                    olap++;
                    rTra[n].n = olap;
                    if (dv - mcp.passBand/2.0 < rTra[n].v_min)
                        rTra[n].v_min = dv - mcp.passBand/2.0;
                    if (dv + mcp.passBand/2.0 > rTra[n].v_max)
                        rTra[n].v_max = dv + mcp.passBand/2.0;
                }
            }
        }
    }

    for (n=0; n<mcp.NrTrans; n++) {
        if (rTra[n].n > 0) nlap++;
    }
    printf("Found %d out of %d lines that participate in overlapping.\n",
           nlap, mcp.NrTrans);
}

int get_olap(int n)
{
    static int o1, o2;
    int n1, n2;

    if (n < 0) {
        o1 = 0;
        o2 = 0;
        return -1;
    }

    if (o1 >= MAX_OLAP || o2 >= MAX_OLAP) {
        fprintf(stderr, "Fatal error: o1 (%d) or o2 (%d) > MAX_OLAP (%d)\n",
                o1, o2, MAX_OLAP);
        exit(1);
    }

    n1 = rTra[n].m[o1];
    n2 = rTra[n1].m[o2];

    o2++;
    if (o2 == rTra[n1].n) {
        o2 = 0;
        o1++;
        if (o1 == rTra[n].n) {
            o1 = 0;
            return -1;
        }
    }

    return n2;
}

void weight_init()
{
    int n, m;
    double /* c1, c2, c3,*/ c4 = 0.0 /*, c5 = 0.0; */;
    double f, fmu;
    /* double edust; */
    RTRANS *t;

    /* c1 = pow(2.708765e7/gSph.Rmax, 3.0); */  /* 4 Pi^2 / (h c[km/s] Rmax^3) */
    /* c2 = pow(gSph.Rmax, 2.0) * mcp.passBand * c1; */
    /* c3 = pow(gSph.Rad[0], 2.0) * c2; */
    if (mcp.dust) {
        /* m_H2[gm] * 3Q_250/4\rho * Rmax */
        c4 = 3.34525e-24 * mcp.dust * gSph.Rmax;
        /* 4 Pi / (h c[km/s] Rmax) * \Delta v */
        /* c5 = 6.3260523e21 / gSph.Rmax * mcp.passBand; */
    }

    for (m=0; m<mcp.NrTrans; m++) {
        t = &rTra[m];
        f = t->Fre;
        /* t->st = pow(2.452098e-7 * f, 3.0); */  /* 2h \nu^3 / c^2[cm/s] */
        t->st = 1.474499e-23 * pow(f, 3.0); /* 2h \nu^3 / c^2[m/s] , W m^-2 Hz^-1*/

        if (mcp.dust) {
            fmu = CLIGHT/1.0e3/f;  /* wavelength in micron */
            for (n=1; n<mcp.Nvol; n++) {
                /* edust = 0.0; */
                /* if (vol[n].Tdust != 0.0) */
                /*     edust = 1.0/(exp(PLOBO_CNST*f/vol[n].Tdust) - 1.0); */
                t->dconst[n] = c4*pow(250.0/fmu, vol[n].beta)*vol[n].nH2/vol[n].Rgd;
            }
        }

        /* c^2[cm/s] c[km/s] / (8 Pi^1.5) */
        t->aconst = 6.0485e-3 * t->Aul / pow(f, 3.0);
    }
}

void grid_init()
{
    int n, m, k, Nang, i, star=0;
    double p, my, dmy, my0=0.0;
    CARTESIAN org, dir;
    GRID *g;

    void find_bndr();

    org.x = org.y = 0.0;
    dir.y = 0.0;
    if (gSph.Tstar > 0.0 && gSph.Rad[0] > 0.0) star=1;

    for (i=0; i<=star; i++) {

        for (n=1; n<mcp.Nvol; n++) {
            if (n == 1 && gSph.Rmin == 0.0) {
                p = 0.0;
            } else {
                p = (gSph.Rad[n] + gSph.Rad[n-1])/2.0;
            }
            org.z = p;
            if (i==0)
                Nang = mcp.NangMin + (int)(p * (double)(mcp.Nang-mcp.NangMin) + 0.5);
            else
                Nang = mcp.NangMin;
            if (Nang % 2 == 1) Nang++;
            if (star) {
                my0 = -sqrt(1.0 - gSph.Rad[0]*gSph.Rad[0]/p/p);
                if (i) {
                    dmy = (my0 + 1.0)/(double)Nang;
                } else {
                    dmy = (1.0 - my0)/(double)Nang;
                }
            } else {
                dmy = 2.0/(double)Nang;
            }
            /* printf("i=%1d  Nang[%2d] = %2d   dmy=%f\n", i, n, Nang, dmy); */
            for (m=0; m<=Nang; m++) {
                g = &grid[n][m][i];
                g->Nang = Nang;
                g->dmy = dmy;
                if (star) {
                    if (i) {
                        my = -1.0 + (double)m * dmy;
                    } else {
                        my = my0 + (double)m * dmy;
                    }
                } else {
                    my = -1.0 + (double)m * dmy;
                }
                my *= 0.99999999;
                dir.x = sqrt(1.0 - my*my);
                dir.z = my;
                find_bndr(org, dir);
                g->n = mcp.Nbnd;
                if (i) {
                    for (k=0; k<mcp.Nbnd; k++) {
                        if (bnd[k].n == 0) {
                            g->n = k;
                            g->Tbg = gSph.Tstar;
                            /* printf("Star at mu=%f %f: %d -> %d\n", my, my0, mcp.Nbnd, g->n); */
                            break;
                        }
                    }
                } else {
                    g->Tbg = mcp.Tbg;
                }
                /* a = -p*my + sqrt(p*p*my*my - p*p + 1.0);
                   b = a*my;
                   z0 = p + b;
                   x0 = sqrt(1.0 - z0*z0); */
                g->org = org;
                g->dir = dir;
                for (k=0; k<g->n; k++) {
                    g->b[k] = bnd[k];
                }
            }
        }

    }

}

void SavePopulations(char *file)
{
    int n, l;
    FILE *fp;
    VOLUME *v;

    fp = fopen(file, "w");
    if (!fp) {
        fprintf(stderr, "Couldn't open %s for writing.\n", file);
        return;
    }
    fprintf(fp, "%d %d\n", mcp.Nvol, mcp.Nlevel);
    for (n=0; n<mcp.Nvol; n++) {
        v = &vol[n];
        for (l=0; l<mcp.Nlevel; l++) {
            fprintf(fp, "%e\n", v->pop[l]);
        }
    }

    fclose(fp);
}

void ReadPopulations(char *file)
{
    int n, l, Nvol, Nlevel;
    double p;
    FILE *fp;
    char buf[100];
    VOLUME *v;

    fp = fopen(file, "r");
    if (!fp) {
        fprintf(stderr, "Couldn't open %s for reading.\n", file);
        exit(1);
    }
    fgets(buf, 100, fp);
    sscanf(buf, "%d %d", &Nvol, &Nlevel);
    if (Nvol != mcp.Nvol || Nlevel != mcp.Nlevel) {
        fprintf(stderr,
                "The population file %s doesn't match the current model cloud.\n",
                file);
        fprintf(stderr, "Nvol = %d, %d, and Nlevel = %d, %d\n",
                Nvol, mcp.Nvol, Nlevel, mcp.Nlevel);
        exit(1);
    }
    for (n=0; n<mcp.Nvol; n++) {
        v = &vol[n];
        for (l=0; l<mcp.Nlevel; l++) {
            fgets(buf, 100, fp);
            sscanf(buf, "%lf", &p);
            v->pop[l] = p;
        }
    }


    fclose(fp);
}

#ifdef COOLING
void SaveTkin(char *file)
{
    int n;
    double r;
    FILE *fp;
    VOLUME *v;

    double Trad(double);

    fp = fopen(file, "w");
    if (!fp) {
        fprintf(stderr, "Couldn't open %s for writing.\n", file);
        return;
    }
    fprintf(fp, "%d\n", mcp.Nvol);
    for (n=1; n<mcp.Nvol; n++) {
        v = &vol[n];
        if (n == 1 && gSph.Rmin == 0.0) {
            r = 0.0;
        } else {
            r = (gSph.Rad[n] + gSph.Rad[n-1])/2.0;
        }
        fprintf(fp, "%8.4f %13.5e %13.5e\n", log10(r*gSph.Rmax), Trad(r), v->Tkin);
    }

    fclose(fp);
}

void ReadTkin(char *file)
{
    int n, Nvol;
    double r, t, p;
    FILE *fp;
    char buf[100];
    VOLUME *v;

    fp = fopen(file, "r");
    if (!fp) {
        fprintf(stderr, "Couldn't open %s for reading.\n", file);
        exit(1);
    }
    fgets(buf, 100, fp);
    sscanf(buf, "%d", &Nvol);
    if (Nvol != mcp.Nvol) {
        fprintf(stderr,
                "The kin. temp. file %s doesn't match the current model cloud.\n",
                file);
        fprintf(stderr, "Nvol = %d, %d\n", Nvol, mcp.Nvol);
        exit(1);
    }
    for (n=1; n<mcp.Nvol; n++) {
        v = &vol[n];
        fgets(buf, 100, fp);
        sscanf(buf, "%lf %lf %lf", &r, &t, &p);
        v->Tkin = p;
    }


    fclose(fp);
}
#endif
