#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>

#ifndef MCP_EXTERNAL
#define MCP_EXTERNAL
#endif

#include "ali.h"

#define MAX_THREADS 32
/* #define USE_FAR_REGION */
#define SMART_VEL
#define NEGLECT_MASERS
/* #define PRINT_LAMBDA 5 */

pthread_mutex_t run_lock;
#ifdef NTHREADS
volatile int n_threads=NTHREADS;    /* How many threads? */
#else
volatile int n_threads=1;           /* How many threads? */
#endif

#ifndef SMART_VEL
volatile double Vel[MAX_VEL];
#endif

void add_con(VOLUME *V, GRID *g, RTRANS *T,
#ifdef SMART_VEL
             double Vel[],
#endif
             double Tau0[], double Tau1[], double TauI[], double TauD[],
             double **TauOver, double Tau[], double Inu[])
{
    int u, l, n, k, i, j, o, ll, uu;
    double t1, t2, ts, t, arg, vdir;
    int olap=0, dust=0, nstep=mcp.Nstep;
#ifdef USE_FAR_REGION
    int active = -1;
#endif
    double dpt, dTau, dTau0, Snu, Tcnst, Snu_sum;
    double d_dTau=0.0, d_Snu=0.0;
    double o_dpt[MAX_OLAP], o_dTau[MAX_OLAP], o_Snu[MAX_OLAP], o_Tcnst[MAX_OLAP];
    RTRANS *TT;
    VOLUME *v;
    BOUND *b;
  
    double VelCos3(GRID *, double, double);

    l = T->l; u = T->u;
    Tcnst = T->aconst * gSph.Rmax;
  
    if (mcp.overlap && T->n > 0) olap = 1;
    if (mcp.dust > 0.0) dust = 1;
  
    if (olap) {
        for (o=0; o<T->n; o++)
            o_Tcnst[o] = rTra[T->m[o]].aconst * gSph.Rmax;
    }

    for (j=0; j<mcp.Nvel; j++) {
        Inu[j] = 0.0;
        if (g->Tbg > 0.0) Inu[j] = T->st/(exp(PLOBO_CNST * T->Fre/g->Tbg) - 1.0);
        Tau[j] = Tau0[j] = Tau1[j] = TauI[j] = TauD[j] = 0.0;
        if (olap) {
            for (o=0; o<T->n; o++) TauOver[j][o] = 0.0;
        }
    }
  
    for (k=g->n-1; k>=0; k--) {
        b = &(g->b[k]);
        n = b->n;
        if (n <= 0) continue;
    
        if (k == 0) {
            t1 = 0.0;
        } else {
            t1 = g->b[k-1].t;
        }
        t2 = b->t;
        if (t1 == t2) continue;
    
        ts = (t2 - t1)/(double)(nstep);
        v = &vol[n];

        dpt = Tcnst * (v->pop[l]*T->Gul - v->pop[u])/v->Vdop * ts;
        Snu = T->st/(v->pop[l]/v->pop[u]*T->Gul - 1.0);
        if (dust) {
            d_dTau = T->dconst[n] * ts;
            d_Snu = 0.0;
            if (v->Tdust != 0.0)
                d_Snu = T->st / (exp(PLOBO_CNST*T->Fre/v->Tdust) - 1.0);
        }
        if (olap) {
            for (o=0; o<T->n; o++) {
                TT = &rTra[T->m[o]];
                ll = TT->l; uu = TT->u;
                o_dpt[o] = o_Tcnst[o] * (v->pop[ll]*TT->Gul - v->pop[uu])/v->Vdop * ts;
                o_Snu[o] = TT->st/(v->pop[ll]/v->pop[uu]*TT->Gul - 1.0);
            }
        }
        for (i=0; i<nstep; i++) {
            t = t2 - (double)i * ts;
            vdir = -VelCos3(g, t, -ts);
            for (j=0; j<mcp.Nvel; j++) {
                arg = (Vel[j] + vdir)/v->Vdop;
                dTau = dTau0 = dpt * exp(-arg * arg);
                if (olap) {
                    Snu_sum = Snu * dTau;
                    if (dust) {
                        Snu_sum += d_Snu * d_dTau;
                        dTau += d_dTau;
                    }
                    for (o=0; o<T->n; o++) {
                        arg = (Vel[j] + vdir - T->dv[o])/v->Vdop;
                        o_dTau[o] = o_dpt[o] * exp(-arg * arg);
                        Snu_sum += o_Snu[o] * o_dTau[o];
                        dTau += o_dTau[o];
                    }
                    if (dTau != 0.0) {
                        Inu[j] += (Snu_sum/dTau - Inu[j]) * (1.0 - exp(-dTau));
                    } else {
                        Inu[j] += Snu_sum;
                    }
                } else {
                    if (dust) {
                        Snu_sum = Snu * dTau + d_Snu * d_dTau;
                        dTau += d_dTau;
                        if (dTau != 0.0) {
                            Inu[j] += (Snu_sum/dTau - Inu[j]) * (1.0 - exp(-dTau));
                        } else {
                            Inu[j] += Snu_sum;
                        }
                    } else {
                        Inu[j] += (Snu - Inu[j]) * (1.0 - exp(-dTau));
                    }
                }
                Tau[j] += dTau;
                if (k == 0) {
                    Tau0[j] += dTau0;
                    if (dust) TauD[j] += d_dTau;
                    if (olap) {
                        for (o=0; o<T->n; o++) {
                            arg = (Vel[j] + vdir - T->dv[o])/v->Vdop;
                            TauOver[j][o] += o_dTau[o];
                        }
                    }
                }
#ifdef USE_FAR_REGION
                if (k > 0 && v == V) {
                    Tau1[j] += dTau;
                    active = k;
                }
                if (active >= 0 && k != active) {
                    TauI[j] += dTau;
                }
#endif
            }
        }
    }
}

void *process(void *number)
{
    int n = *(int *)number, m, k, j, olap, dust=0, o, Nang, i, star=0;
#ifdef NEGLECT_MASERS
    int maser=-1;
#endif
    double dmy, wmy, wnu, Jmy, Lambda, e, TauTot, arg;
    double Jbar[MAX_RTRANS], Lbar[MAX_RTRANS], LbarOver[MAX_RTRANS][MAX_OLAP];
#ifdef COOLING
    double b, Beta[MAX_RTRANS];
#endif
    double Tau[MAX_VEL], Inu[MAX_VEL];
    double Tau0[MAX_VEL], Tau1[MAX_VEL], TauI[MAX_VEL], TauD[MAX_VEL];
    double **TauOver = NULL;
    double o_Lambda[MAX_OLAP];
    RTRANS *T;
    GRID *g;
    VOLUME *v;
#ifdef SMART_VEL
    double Vel[MAX_VEL];
#else
    double dv;
#endif

    double VelCos(GRID *, double);
    double **matrix(int, int);
    void free_matrix(double **, int);
    
    v = &vol[n];
#ifndef SMART_VEL
    dv = mcp.passBand/(double)(mcp.Nvel - 1);
#endif
    
    if (mcp.overlap) {
        TauOver = (double **) matrix(mcp.Nvel, MAX_OLAP);
        if (!TauOver) {
            fprintf(stderr, "process: Out of memory (%d).\n", n);
            exit(1);
        }
    }
    
    if (mcp.dust > 0.0) dust = 1;
    
    for (k=0; k<mcp.NrTrans; k++) {
        Jbar[k] = Lbar[k] = 0.0;
        if (mcp.overlap) {
            for (o=0; o<MAX_OLAP; o++) LbarOver[k][o] = 0.0;
        }
#ifdef COOLING
        Beta[k] = 0.0;
#endif
    }
    
    if (gSph.Tstar > 0.0 && gSph.Rad[0] > 0.0) star=1;
    
    for (i=0; i<=star; i++) {
    
        Nang = grid[n][0][i].Nang;
        /* dmy = 2.0/(double)Nang; */
        dmy = grid[n][0][i].dmy;
    
        /* printf("i=%d n=%d dmy=%f Nang=%d\n", i, n, dmy, Nang); */
    
        for (m=0; m<=Nang; m++) {
            g = &grid[n][m][i];
#ifdef SMART_VEL
            for (j=0; j<mcp.Nvel; j++) {
                Vel[j] = v->Vdop * Hz.z[j] + VelCos(g, 0.0);
            }
            wmy = (double)(1 + (m % 2));
#else
            wmy = (double)(1 + (m % 2)) * dv;
#endif
            if (m == 0 || m == Nang) wmy *= 0.5;
            for (k=0; k<mcp.NrTrans; k++) {
                T = &rTra[k];
                olap = 0;
                if (mcp.overlap && T->n > 0) olap = 1;
                add_con(v, g, T,
#ifdef SMART_VEL
                        Vel,
#endif
                        Tau0, Tau1, TauI, TauD, TauOver, Tau, Inu);
                Jmy = Lambda = 0.0;
#ifdef COOLING
                b = 0.0;
#endif
                if (olap) {
                    for (o=0; o<T->n; o++) o_Lambda[o] = 0.0;
                }
                for (j=0; j<mcp.Nvel; j++) {
#ifndef SMART_VEL
                    wnu = (double)(1 + ((j+1) % 2));
                    if (j == 0 || j == mcp.Nvel - 1) wnu *= 0.5;
#else
                    wnu = Hz.w[j];
#endif
                    TauTot = Tau0[j];
                    if (dust) TauTot += TauD[j];
                    if (olap) {
                        for (o=0; o<T->n; o++) TauTot += TauOver[j][o];
                        if (TauTot != 0.0) {
                            for (o=0; o<T->n; o++) {
                                arg = (Vel[j] - VelCos(g, 0.0) - T->dv[o])/v->Vdop;
                                e = exp(-arg * arg) * wmy * wnu * Tau0[j] / TauTot;
#ifdef SMART_VEL
                                e *= exp(Hz.z[j] * Hz.z[j]);
#endif
                                o_Lambda[o] += (1.0 - exp(-TauTot/mcp.Gamma)) * e;
                            }
                        }
                    }
#ifndef SMART_VEL
                    arg = (Vel[j] - VelCos(g, 0.0))/v->Vdop;
                    e = exp(-arg * arg) * wmy * wnu;
#else
                    e = wmy * wnu;
#endif
                    Jmy += Inu[j] * e;
#ifdef COOLING
                    b += exp(-TauTot) * e;
#endif
                    if (TauTot == 0.0) continue;
#ifdef NEGLECT_MASERS
                    if (Tau0[j] < -1.0) {
                        if (maser == -1) {
                            maser = k;
                        }
                        continue;
                    }
#endif
                    e *= Tau0[j]/TauTot;
                    Lambda += (1.0 - exp(-TauTot/mcp.Gamma)) * e;
#ifdef USE_FAR_REGION
                    Lambda += (1.0 - exp(-Tau1[j]/mcp.Gamma)) *
                        exp(-TauI[j]/mcp.Gamma) * e;
#endif
                }
                Jbar[k] += Jmy * dmy;
                Lbar[k] += Lambda * dmy;
#ifdef COOLING
                Beta[k] += b * dmy;
#endif
                if (olap) {
                    for (o=0; o<T->n; o++) LbarOver[k][o] += o_Lambda[o] * dmy;
                }
            }
        }
    
    }
    
#ifdef SMART_VEL
    e = 1.0 / 3.0;
#else
    e = 2.0 * 1.0 / sqrt(M_PI) / v->Vdop / 9.0;
#endif    
    if (TauOver) free_matrix(TauOver, mcp.Nvel);
  
    /* Lock for update, update it, and unlock */
    pthread_mutex_lock(&run_lock);
#ifdef PRINT_LAMBDA
    printf("%2d ", n);
#endif
    for (k=0; k<mcp.NrTrans; k++) {
        T = &rTra[k];
        v->Jbar[k] = Jbar[k] * e / T->st;
        v->Lbar[k] = Lbar[k] * e;
#ifdef COOLING
        v->Beta[k] = Beta[k] * e;
#endif
        if (mcp.overlap && T->n > 0) {
            for (o=0; o<T->n; o++) v->LbarOver[k][o] = LbarOver[k][o] * e;
        }
#ifdef PRINT_LAMBDA
        if (k == PRINT_LAMBDA) {
            printf("%6.3f : ", v->Lbar[k]);
            if (mcp.overlap && T->n > 0) {
                for (o=0; o<T->n; o++)
                    printf("%6.3f ", v->LbarOver[k][o]);
            }
            T = &rTra[k-1];
            printf(" - %6.3f : ", v->Lbar[k-1]);
            if (mcp.overlap && T->n > 0) {
                for (o=0; o<T->n; o++)
                    printf("%6.3f ", v->LbarOver[k-1][o]);
            }
        }
#endif
    }
#ifdef PRINT_LAMBDA
    printf("\n");
#endif
    pthread_mutex_unlock(&run_lock);

    return(NULL);
}

void run()
{
    static int first = 1;
    int n, i;
#ifndef SMART_VEL
    int j;
    double dv;
#endif
    pthread_t thread[MAX_THREADS];
    int t[MAX_THREADS];
    void *retval;
  
  
#ifndef SMART_VEL
    dv = mcp.passBand/(double)(mcp.Nvel - 1);
#endif
    if (first) {
#ifndef SMART_VEL
        for (j=0; j<mcp.Nvel; j++) Vel[j] = -mcp.passBand/2.0 + (double)j * dv;
#endif
        /* Initialize the lock */
        first = 0;
        if (n_threads > MAX_THREADS) {
            fprintf(stderr, "Increase MAX_THREADS to %d\n", n_threads);
            exit(1);
        }
        pthread_mutex_init(&run_lock, NULL);
    }
  
  
    i = 0;
    for (n=1; n<mcp.Nvol; n++) {
    
        if (pthread_create(&thread[i], NULL, process, (void *)&n)) {
	    fprintf(stderr, "%s: cannot make thread %d.\n", "run", i);
	    exit(1);
        } else {
            t[i] = 1;
        }

        i++;
        if (i >= n_threads) {
            for(i=0; i<n_threads; i++) {
                if (pthread_join(thread[i], &retval)) {
	            fprintf(stderr, "%s: thread %d join failed.\n", "run", i);
	            exit(1);
                } else {
                    t[i] = 0;
                }
            }
            i = 0;
        }
    }
  
    if (i == 0) return;
  
    /* Final joining of threads if nthreads doesn't divide mcp.Nvol */
    for (i=0; i<n_threads; i++) {
        if (t[i]) {
            if (pthread_join(thread[i], &retval)) {
	        fprintf(stderr, "%s: thread %d final join failed.\n", "run", i);
	        exit(1);
            } else {
                t[i] = 0;
            }
        }
    }
}

/* Factorial function, n! */
double Factorial(int n)
{
    int i;
    double f=1.0;
    
    if (n == 0 || n == 1) {
        f = 1.0;
    } else {
        for (i=n; i>1; i--) f *= (double)i;
    }
    
    return f;
}

double RatioFact(int n1, int n2)
{
    int i;
    double f = 1.0;
    
    if (n1 == n2) {
        return 1.0;
    } else if (n2 > n1) {
        return 0.0;
    } else if (n2 == 0 || n2 == 1) {
        return Factorial(n1);
    } else {
        for (i=n2+1; i<=n1; i++) f *= (double)i;
    }
    
    return f;
}

/* Hermite polynomial a la Handbook of Mathematical Functions Sect. 22.18
   */
double Hermite(int N, double x)
{
    int n, m;
    double a=1.0;
    
    if (N % 2 == 1) {
        if (N == 1) return 2.0*x;
        if (N == 3) return 4.0*x*(2.0*x*x - 3.0);
        n = (N-1)/2;
        for (m=n; m>0; m--) {
            a = 1.0 - 2.0 * (double)(n-m+1) / (double)(m*(2*m+1)) * x * x * a;
        }
        a *= 2.0 * x * RatioFact(2*n+1, n);
    } else {
        if (N == 0) return 1.0;
        if (N == 2) return 2.0*(2.0*x*x - 1.0);
        n = N/2;
        for (m=n; m>0; m--) {
            a = 1.0 - 2.0 * (double)(n-m+1) / (double)(m*(2*m-1)) * x * x * a;
        }
        a *= RatioFact(2*n, n);
    }
    
    if (n % 2 == 1) a *= -1.0;

    return a;
}

double RatioHermite(int N, double x)
{
    int n, m;
    double a1=1.0, a2=1.0, r;
    
    if (N % 2 == 1) {
        if (N == 1) {
            r = 2.0 * x;
        } else if (N == 3) {
            r = 2.0*x*(2.0*x*x - 3.0)/(2.0*x*x - 1.0);
        } else {
            n = (N-1)/2;
            for (m=n; m>0; m--) {
                a1 = 1.0 - 2.0*(double)(n-m+1)/(double)(m*(2*m+1))*x*x*a1;
                a2 = 1.0 - 2.0*(double)(n-m+1)/(double)(m*(2*m-1))*x*x*a2;
            }
            r = x * a1 / a2;
        }
    } else {
        if (N == 2) {
            r = (2.0*x*x - 1.0)/4.0/x;
        } else {
            n = N/2;
            for (m=n; m>0; m--) {
                a1 = 1.0 - 2.0*(double)(n-m+1)/(double)(m*(2*m-1))*x*x*a1;
                if (m < n)
                    a2 = 1.0 - 2.0*(double)(n-m)/(double)(m*(2*m+1))*x*x*a2;
            }
            r = -a1 / a2 / x / 4.0 / (double)n;
        }
    }

    return r;
}

/* Hermite polynomial H(n,x), recursive version */
double rHermite(int n, double x)
{
    double h;
    
    if (n == 0) {
        h = 1.0;
    } else if (n == 1) {
        h = 2.0*x;
    } else if (n == 2) {
        h = 4.0*x*x - 2.0;
    } else if (n == 3) {
        h = 4.0*x*(2.0*x*x - 3.0);
    } else if (n < 0) {
        h = 0.0;
    } else {
        h = 2.0*(x*Hermite(n-1, x) - (double)(n-1) * Hermite(n-2, x));
    }
    return h;
}

/* The derivative of Hermite polynomials */
double dHermite(int n, double x)
{
    return 2.0*n*Hermite(n-1,x);
}

/* Find the zeros of Hermite polynomials in the interval [x1,x2]
   using the Newton-Raphson method. It can fail to find a zero in
   the interval if there are two or more zeros in [x1,x2].
   However, for n <= 50 it works if the interval length is <=0.05.
   The accuracy is set to 10^(-11). It works up n=74 but there the
   accuracy is only 10^(-7).
 */
static int GetHermiteZero(double x1, double x2, int n, double *x)
{
    int i = 0;
    double h1, h2, xnew, xold;
    double eps = 1.0e-11;
    
    h1 = Hermite(n, x1);
    h2 = Hermite(n, x2);
    
    
    /* Any zero in the interval [x1,x2]? */
    if (h1 > 0.0 && h2 > 0.0) return 1;
    if (h1 < 0.0 && h2 < 0.0) return 1;
    
    /* printf("[%f,%f]   h=[%e,%e]\n", x1,x2,h1,h2); */
    
    if (n > 50) eps *= 10.0;
    if (n > 59) eps *= 10.0;
    if (n > 64) eps *= 10.0;
    if (n > 68) eps *= 10.0;
    
    xold = x2;
    xnew = (x1+x2)/2.0;
    while (fabs(xnew-xold) > eps) {
        xold = xnew;
        xnew = xold - RatioHermite(n, xnew);
        i++;
        if (i > 4000) return 1; /* Let's not get stuck in the loop... */
    }
    /* Is the zero found outside requested interval [x1,x2]? */
    if (xnew < x1 || xnew > x2) return 1;
    
    /* Ok, accept the zero! */
    if (x) *x = xnew;
    
    return 0;
}

static double HermiteWeight(int n, double x)
{
    double w, h;
    
    h = Hermite(n-1, x);
    w = pow(2.0, (double)(n-1)) / (double)n / h;
    w *= Factorial(n-1) / h;
    
    return w;
}

void FindHermiteZeros(int n)
{
    int m, nz=0, NZ=n;
    double x, x1=0.01, x2, xs=0.025, w;
    
    if (n % 2 == 1) { /* Odd Hermite polynomials have a zero at x=0 */
        w = HermiteWeight(n, 0.0);
        Hz.n = nz;
        Hz.z[nz] = 0.0;
        Hz.w[nz] = w;
        nz++;
    }
    
    for (m=0; m<1000; m++) {
        x2 = x1 + xs;
        if (!GetHermiteZero(x1, x2, n, &x)) {
            w = HermiteWeight(n, x);
            if (nz < MAX_ZERO-1) {
                Hz.n = nz+1;
                Hz.z[nz] = x;
                Hz.w[nz] = w;
                nz++;
                Hz.z[nz] = -x;
                Hz.w[nz] = w;
            }
            nz++;
            if (nz == NZ) break;
        }
        x1 = x2;
    }
    
    if (nz == NZ) {
        printf("All %d Hermite-zeros found.\n", nz);
        printf("Doppler range: [%f,%f]\n", Hz.z[nz-1], Hz.z[nz-2]);
#ifdef SMART_VEL
        if (mcp.overlap) {
            w = fabs(Hz.z[nz-2] - Hz.z[nz-1]);
            if (w < mcp.passBand) {
                fprintf(stderr, "Error! Preprocessor #define SMART_VEL in\n");
                fprintf(stderr, "run.c cannot be used together with line\n");
                fprintf(stderr, "overlap if the Doppler range %f does\n", w);
                fprintf(stderr, "not cover the overlap region %f.\n",
                        mcp.passBand);
                fprintf(stderr, "Uncomment and recompile.\n");
                exit(1);
            }
        }
#endif
    } else {
        fprintf(stderr,
                "Error: Only %d Hermite-zeros out of %d were found.\n", nz, NZ);
        exit(1);
    }
}

void hermite_init(int order)
{
#ifdef SMART_VEL
    FindHermiteZeros(order);
#endif
}
