#include <math.h>
#include <stdio.h>
#include <string.h>

#ifndef MCP_EXTERNAL
#define MCP_EXTERNAL
#endif

#include "ali.h"

/* #define INFALL */
/* #define I16 */
/* #define IRC10216 0.30 */
/* #define WHYA */
/* #define ORION */
/* #define O2_INFALL */

void init_rad(char *type, double *r, double Rmin, int n, int nmax)
{
    int i;
    double dr, e;

    if (strcmp(type, "log") == 0) {
        dr = -log(Rmin)/(double)(n - 1);
        for (i=0; i<n; i++) {
            r[i] = exp(log(Rmin) + dr * (double) i);
        }
    } else if (strcmp(type, "vol") == 0) {
        dr = (1.0 - pow(Rmin, 3.))/(double)(n - 1);
        for (i=0; i<n; i++) {
            r[i] = pow(pow(Rmin, 3.) + dr * (double) i, 1./3.0);
        }
    } else if (strcmp(type, "lin") == 0) {
        dr = (1.0 - Rmin)/(double)(n - 1);
        for (i=0; i<n; i++) {
            r[i] = Rmin + dr * (double) i;
        }
    } else {
        sscanf(type, "%lf", &e);
        dr = (1.0 - pow(Rmin, e))/(double)(n - 1);
        for (i=0; i<n; i++) {
            r[i] = pow(pow(Rmin, e) + dr * (double) i, 1./e);
        }
    }
}

double Trad(double r)
{
    double T;

#ifdef INFALL
    if (r >= 0.15) {
        return 11.0 + 3.53*(r - 0.15);
    }
    T = gSph.Trad0 + gSph.Trad1 * pow(r/0.15, gSph.Trad2);
#else
    T = gSph.Trad0 + gSph.Trad1 * pow(r, gSph.Trad2);
#ifdef IRC10216
    if (r > IRC10216) {
        T += (r - IRC10216)/(1.0 - IRC10216) * 6.0;
    }
#endif
#endif

#ifdef WHYA
    if (T > 1300.0) T = 1300.0;
#endif
  
    return T;
}

double Tdrad(double r)
{
    double T;

#ifdef INFALL
    if (r >= 0.15) {
        return 11.0 + 3.53*(r - 0.15);
    }
    T = gSph.Tdrad0 + gSph.Tdrad1 * pow(r/0.15, gSph.Tdrad2);
#else
    T = gSph.Tdrad0 + gSph.Tdrad1 * pow(r, gSph.Tdrad2);
#endif

#ifdef WHYA
    if (T > 1300.0) T = 1300.0;
#endif
  
    return T;
}

double Nrad(double r)
{
    double N;
  
#ifdef INFALL
    if (r >= 0.15) {
        return 2.1e3 * pow(r, -1.6);
    }
    N = gSph.Nrad0 + gSph.Nrad1 * pow(r/0.15, gSph.Nrad2);
#else
    N = gSph.Nrad0 + gSph.Nrad1 * pow(r, gSph.Nrad2);
    if (N < gSph.Nrad3) N = gSph.Nrad3;
#endif
  
    return N;
}

double Wrad(double r, char c)
{
    double W=0.0;
  
    if (c == 'F') {
        W = gSph.Wrad0 + gSph.Wrad1 * pow(r, gSph.Wrad2);
    } else {
        if (r <= gSph.Wrad0)
            W = gSph.Wp[0];
        else if (r > gSph.Wrad0 && r <= gSph.Wrad1)
            W = gSph.Wp[1];
        else
            W = gSph.Wp[2];
    }
  
    return W;
}

double Xrad(double r)
{
    double X;
  
    X = gSph.Xrad0 + gSph.Xrad1 * pow(r, gSph.Xrad2);
  
    return X;
}

double Xexp(double r)
{
    double X, arg;
  
    arg = (r - gSph.Xrad1)*(r - gSph.Xrad1)/gSph.Xrad2/gSph.Xrad2;
    X = gSph.Xrad0 * exp(-arg);
  
    if (X < gSph.Xrad3) X = gSph.Xrad3;
  
    return X;
}

double Xexp1(double r)
{
    double X, arg;
  
    arg = pow(r/gSph.Xrad1, gSph.Xrad2);
    X = gSph.Xrad0 * exp(-log(2.0)*arg);
  
    if (X < gSph.Xrad3) X = gSph.Xrad3;
  
    return X;
}

double VelRad(double r)
{
    double v=0.0;

#ifdef I16  
    if (r >= 0.25) return 0.5;
  
    v = gSph.Vrad0 + gSph.Vrad1 * pow(r/0.25, gSph.Vrad2);
#else
#ifdef WHYA
    if (r > gSph.Vrad1) v = gSph.Vrad0 * sqrt(1.0 - gSph.Vrad1/r);
    if (v < gSph.Vrad2) v = gSph.Vrad2;
#else
#ifdef ORION
    if (r < 0.233) {
        v = 0.0;
    } else {
        v = gSph.Vrad1*r;
    }
#else
#ifdef O2_INFALL
    if (r <= gSph.Vrad0) {
        v = gSph.Vrad1 * sqrt(r/gSph.Vrad0);
    } else {
        v = gSph.Vrad1;
    }
#else
    if (gSph.Vrad2 == 0.0) {
        if (gSph.Vrad1 < 0.0) {
            if (r < -gSph.Vrad1)
                v = gSph.Vrad0;
            else
                v = 0.0;
        } else {
            if (r >= gSph.Vrad1)
                v = gSph.Vrad0;
            else
                v = 0.0;
        }
    } else {
        v = gSph.Vrad0 + gSph.Vrad1 * pow(r, gSph.Vrad2);
    }
#endif
#endif
#endif
#endif  
  
    return v;
}

double VelCos(GRID *g, double t)
{
    double r, vcos;
    CARTESIAN c;
  
    double VelRad();
  
    C_line(g->org, g->dir, t, &c);
    r = C_magn(c);
  
    if (r == 0.0) return 0.0;
  
    vcos = VelRad(r) * C_dot(g->dir, c) / r;
  
    return vcos;
}

double VelCos3(GRID *g, double t, double ts)
{
    double r, vcos, v1, v2, v3;
    CARTESIAN c;
  
    double VelRad();
  
    C_line(g->org, g->dir, t, &c);
    r = C_magn(c);
    if (r == 0.0)
        v1 = 0.0;
    else
        v1 = VelRad(r) * C_dot(g->dir, c) / r;

    C_line(g->org, g->dir, t + ts/2.0, &c);
    r = C_magn(c);
    if (r == 0.0)
        v2 = 0.0;
    else
        v2 = VelRad(r) * C_dot(g->dir, c) / r;

    C_line(g->org, g->dir, t + ts, &c);
    r = C_magn(c);
    if (r == 0.0)
        v3 = 0.0;
    else
        v3 = VelRad(r) * C_dot(g->dir, c) / r;
  
    vcos = (v1 + 2.0*v2 + v3)/4.0;
  
    return vcos;
}

void find_bndr(CARTESIAN org, CARTESIAN dir)
{
    int n, m;
    double a, b, c, t, r, swap;
    CARTESIAN line;
  
    mcp.Nbnd = 0;
  
    a = -C_dot(org, dir);
    b = -C_dot(org, org);
  
    for (n=0; n<mcp.Nvol; n++) {
        r = gSph.Rad[n];
        c = a*a + b + r*r;
        if (c <= 0.0) continue;
        for (m=-1; m<=1; m += 2) {
            t = a + (double)m * sqrt(c);
            if (t < 0.0) continue;
            bnd[mcp.Nbnd].t = t;
            mcp.Nbnd++;
        }
    }
  
    for (n=0; n<mcp.Nbnd-1; n++) {
        for (m=n+1; m<mcp.Nbnd; m++) {
            if (bnd[n].t > bnd[m].t) {
                swap = bnd[n].t;
                bnd[n].t = bnd[m].t;
                bnd[m].t = swap;
            }
        }
    }
  
    for (m=0; m<mcp.Nbnd; m++) {
        if (m == 0) {
            t = 0.0;
        } else {
            t = (bnd[m].t + bnd[m-1].t)/2.0;
        }
        C_line(org, dir, t, &line);
        r = C_magn(line);
        bnd[m].n = -1;
        for (n=0; n<mcp.Nvol; n++) {
            if (r < gSph.Rad[n]) {
                bnd[m].n = n;
                break;
            }
        }
    }
}
