#
#awk script to convert Leiden format to Bernes/ali format (PB 2005)
#
#usage:  awk -f leiden.awk leidendatfile > molecule.dat
#
#Note all occurances with space after ! (i.e. "! MOLECULE")
#in the Leiden files need to be removed (i.e. "!MOLECULE")
#before running the awk script
#
#carefully check the result in molecule.dat since the Leiden keywords vary from file to file
#
BEGIN{
  molecule = "none"
  molweight = 0.0
  nlev = nrtra = 0
  nctra[0] = 0
  ntemp[0] = 0
  colldesc[0] = "none"
  use_freqs = 0
}
{
if ($1 == "!MOLECULE") {
  getline
  molecule = $1
} else if ($1 == "!MOLECULAR" || $1 == "!MASS") {
  getline
  molweight = $1
} else if ($1 == "!NUMBER" && $3 == "ENERGY") {
  getline
  nlev = $1
} else if ($1 == "!LEVEL") {
  for (n=1; n<=nlev; n++) {
    getline
    if ($1 == n) {
        egy[n] = $2
	g[n] = $3
    } else {
      print "*** Level error", n
    }
  }
} else if ($1 == "!NUMBER" && $3 == "RADIATIVE") {
  getline
  nrtra = $1
} else if ($1 == "!TRANS" && ($7 == "EINSTEINA(s^-1)" || $7 == "A(s^-1)") ) {
  for (n=1; n<=nrtra; n++) {
    getline
    if ($1 == n) {
        upper[n] = $2
	lower[n] = $3
        Aul[n] = $4
	ful[n] = $5
    } else {
      print "*** Transition error ", $1, n
      break
    }
  }
} else if ($1 == "!NUMBER" && $4 == "PARTNERS") {
  getline
  npart = $1
  p = 0
} else if ( ( ($1 == "!COLLISIONS" || $1 == "!COLLISION") && $2 != "TEMPERATURES") ||\
              $1 == "!Partner" ) {
  getline
  colldesc[p] = $0
} else if (($1 == "!NUMBER" || $1 == "!Number") &&\
           ($3 == "COLLISIONAL" || $3 == "COLL" || $3 == "collisional") &&\
           ($4 == "TRANS" || $4 == "TRANSITIONS" || $4 == "transitions") ) {
  getline
  nctra[p] = $1
} else if (($1 == "!NUMBER" || $1 == "!Number") &&\
           ($4 == "TEMPS" || $4 == "TEMPERATURES" || $4 == "temperatures")) {
  getline
  ntemp[p] = $1
} else if ($1 == "!COLL" || $1 == "!COLLISION") {
  getline
  n = split($0, tmp)
  for (m=1; m<=n; m++) ctemp[m,p] = tmp[m]
  if (n != ntemp[p]) print "*** ntemp error", ntemp[p], n
} else if (($1 == "!TRANS" || $1 == "!TRANS+") &&\
           ($7 == "COLLRATES(cm^3" || $7 == "RATE" || $7 == "s-1")) {
  for (n=1; n<=nctra[p]; n++) {
    getline
    m = split($0, coll)
    if (m != 3+ntemp[p]) {
        print "*** Coll. error at n,m=", n, m
	break
    }
    if ($1 == n) {
        cupper[n,p] = $2
	clower[n,p] = $3
	for (m=1; m<=ntemp[p]; m++) {
           Cul[n,m,p] = coll[3+m]
	}
    } else {
      print "*** Collision error ", n
      break
    }
  }
  if (npart > 1) p = p + 1
} else if ($1 == "!NOTES") {
}
}
END{
printf("!LABEL1\n")
printf("%s term energies and A-values\n", molecule)
printf("!MOLW\n")
printf("%.1f\n", molweight)
printf("!NLEVS\n")
printf("%d\n", nlev)
printf("!NTRANS\n")
printf("%d\n", nrtra)
printf("!ENERGIES\n")
for (n=1; n<=nlev; n++) {
  printf("%11.6f", egy[n])
  if (n % 7 == 0 || n == nlev) printf("\n")
}
printf("!WEIGHTS\n")
for (n=1; n<=nlev; n++) {
  printf("%5.1f", g[n])
  if (n % 16 == 0 || n == nlev) printf("\n")
}
printf("!RAD_UPPER\n")
for (n=1; n<=nrtra; n++) {
  printf("%3d", upper[n])
  if (n % 24 == 0 || n == nrtra) printf("\n")
}
printf("!RAD_LOWER\n")
for (n=1; n<=nrtra; n++) {
  printf("%3d", lower[n])
  if (n % 24 == 0 || n == nrtra) printf("\n")
}
printf("!A_COEFFS\n")
for (n=1; n<=nrtra; n++) {
  printf("%10.3e", Aul[n])
  if (n % 8 == 0 || n == nrtra) printf("\n")
}
if (use_freqs) {
  printf("!FREQ_MHZ\n")
  for (n=1; n<=nrtra; n++) {
    printf("%16.9e", ful[n])
    if (n % 5 == 0 || n == nrtra) printf("\n")
  }
}
for (i=0; i<npart; i++) {
  printf("!LABEL%d\n",i+2)
  printf("%s\n", colldesc[i])
  if (i == 0)
    printf("!COL_TRANS\n")
  else
    printf("!COL%d_TRANS\n", i)
    
  printf("%d\n", nctra[i])
  if (i == 0)
    printf("!COL_NTK\n")
  else
    printf("!COL%d_NTK\n", i)
  printf("%d\n", ntemp[i])
  
  if (i == 0)
    printf("!COL_TK\n")
  else
    printf("!COL%d_TK\n", i)
  for (n=1; n<=ntemp[i]; n++) {
    printf("%6.1f", ctemp[n,i])
    if (n % 14 == 0 || n == ntemp[i]) printf("\n")
  }
  
  if (i == 0)
    printf("!COL_UPPER\n")
  else
    printf("!COL%d_UPPER\n", i)
  for (n=1; n<=nctra[i]; n++) {
    printf("%3d", cupper[n,i])
    if (n % 24 == 0 || n == nctra[i]) printf("\n")
  }
  
  if (i == 0)
    printf("!COL_LOWER\n")
  else
    printf("!COL%d_LOWER\n", i)
  for (n=1; n<=nctra[i]; n++) {
    printf("%3d", clower[n,i])
    if (n % 24 == 0 || n == nctra[i]) printf("\n")
  }
  
  for (m=1; m<=ntemp[i]; m++) {
    if (i == 0)
      printf("!C_COEFFS\n")
    else
      printf("!C%d_COEFFS\n", i)
    for (n=1; n<=nctra[i]; n++) {
      printf("%8.1e", Cul[n,m,i])
      if (n % 8 == 0 || n == nctra[i]) printf("\n")
    }
  }
}
}
