#define LOGRECLEN     80
#define MAXCARDS	  36
#define PHYSRECLEN	(LOGRECLEN*MAXCARDS)
#define KEYWORDLEN     8
#define VALUELEN	  20
#define COMMENTLEN	  46
#define MAXVARLEN     40

#define NVARTYPE 0
#define BOOLTYPE 1
#define CHARTYPE 2
#define LONGTYPE 3
#define REALTYPE 5
#define CPLXTYPE 6

#define TRUE	1
#define FALSE	0

#define MAXNAXIS 5

#ifdef LINUX
#ifndef BYTESWAP
#define BYTESWAP
#endif
#endif

enum kw_known {
    KW_SIMPLE,
    KW_XTENSION,
    KW_BITPIX,
    KW_NAXIS,
    KW_NAXIS1,
    KW_NAXIS2,
    KW_NAXIS3,
    KW_CTYPE1,
    KW_CRPIX1,
    KW_CRVAL1,
    KW_CDELT1,
    KW_CROTA1,
    KW_CTYPE2,
    KW_CRPIX2,
    KW_CRVAL2,
    KW_CDELT2,
    KW_CROTA2,
    KW_CTYPE3,
    KW_CRPIX3,
    KW_CRVAL3,
    KW_CDELT3,
    KW_CROTA3,
    KW_BSCALE,
    KW_BZERO,
    KW_BUNIT,
    KW_BLANK,
    KW_DATAMAX,
    KW_DATAMIN,
    KW_ORIGIN,
    KW_OBSERVER,
    KW_OBJECT,
    KW_INSTRUME,
    KW_TELESCOP,
    KW_LINE,
    KW_SCAN,
    KW_UTC,
    KW_LST,
    KW_DATE_OBS,
    KW_DATE,
    KW_TIME,
    KW_JDATE,
    KW_TSYS,
    KW_TLOAD,
    KW_TREC,
    KW_TAU,
    KW_DBLOAD,
    KW_EQUINOX,
    KW_RA,
    KW_DEC,
    KW_AZIMUTH,
    KW_ELEVATIO,
    KW_AZOFF,
    KW_ELOFF,
    KW_AZPOINT,
    KW_ELPOINT,
    KW_AZCORR,
    KW_ELCORR,
    KW_COLLIMAT,
    KW_RESTFREQ,
    KW_FREQRES,
    KW_VLSR,
    KW_VHEL,
    KW_VELO_LSR,
    KW_VELO_HEL,
    KW_VELO_GEO,
    KW_VELOCITY,
    KW_DELTAV,
    KW_DELTAVEL,
    KW_OBSTIME,
    KW_INTTIME,
    KW_TAMB,
    KW_TOUTSIDE,
    KW_PRESSURE,
    KW_HUMIDITY,
    KW_COMMENT,
    KW_HISTORY,
    KW_EXTEND,
    KW_PCOUNT,
    KW_GCOUNT,
    KW_TFIELDS,
    KW_EXTNAME,
    KW_EXTVER,
    KW_NMATRIX,
    KW_MAXIS,
    KW_MAXIS1,
    KW_TFORM,
    KW_TTYPE,
    KW_TUNIT,
    KW_TDIM,
    KW_END
};

struct knownkey {
    int key;
    int type;
    char text[KEYWORDLEN+1];
    char comment[COMMENTLEN+1];
};

#ifndef FITSKNOWN
#define FITSKNOWN
struct knownkey known[] = {
    { KW_SIMPLE,   BOOLTYPE, "SIMPLE",   "Basic FITS format" },
    { KW_XTENSION, CHARTYPE, "XTENSION", "Extension indicator" },
    { KW_BITPIX,   LONGTYPE, "BITPIX",   "Number of bits used for pixel value" },
    { KW_NAXIS,    LONGTYPE, "NAXIS",    "Number of axes" },
    { KW_NAXIS1,   LONGTYPE, "NAXIS1",   "Number of values along first axis" },
    { KW_NAXIS2,   LONGTYPE, "NAXIS2",   "Number of values along second axis" },
    { KW_NAXIS3,   LONGTYPE, "NAXIS3",   "Number of values along third axis" },
    { KW_CTYPE1,   CHARTYPE, "CTYPE1",   "Type of first axis" },
    { KW_CRPIX1,   REALTYPE, "CRPIX1",   "Reference of first axis" },
    { KW_CRVAL1,   REALTYPE, "CRVAL1",   "Offset value of first axis" },
    { KW_CDELT1,   REALTYPE, "CDELT1",   "Increment value of first axis" },
    { KW_CROTA1,   REALTYPE, "CROTA1",   "Rotation of first axis" },
    { KW_CTYPE2,   CHARTYPE, "CTYPE2",   "Type of second axis" },
    { KW_CRPIX2,   REALTYPE, "CRPIX2",   "Reference of second axis" },
    { KW_CRVAL2,   REALTYPE, "CRVAL2",   "Offset of second axis" },
    { KW_CDELT2,   REALTYPE, "CDELT2",   "Increment of second axis" },
    { KW_CROTA2,   REALTYPE, "CROTA2",   "Rotation of second axis" },
    { KW_CTYPE3,   CHARTYPE, "CTYPE3",   "Type of third axis" },
    { KW_CRPIX3,   REALTYPE, "CRPIX3",   "Reference of third axis" },
    { KW_CRVAL3,   REALTYPE, "CRVAL3",   "Offset of third axis" },
    { KW_CDELT3,   REALTYPE, "CDELT3",   "Increment of third axis" },
    { KW_CROTA3,   REALTYPE, "CROTA3",   "Rotation of third axis" },
    { KW_BSCALE,   REALTYPE, "BSCALE",   "value = idata*BSCALE + BZERO" },
    { KW_BZERO,    REALTYPE, "BZERO",    "value = idata*BSCALE + BZERO" },
    { KW_BUNIT,    CHARTYPE, "BUNIT",    "Intensity unit" },
    { KW_BLANK,    REALTYPE, "BLANK",    "" },
    { KW_DATAMAX,  REALTYPE, "DATAMAX",  "Maximum data value" },
    { KW_DATAMIN,  REALTYPE, "DATAMIN",  "Minimum data value" },
    { KW_ORIGIN,   CHARTYPE, "ORIGIN",   "Program that generated this file" },
    { KW_OBSERVER, CHARTYPE, "OBSERVER", "Observer" },
    { KW_OBJECT,   CHARTYPE, "OBJECT",   "Object or source name" },
    { KW_INSTRUME, CHARTYPE, "INSTRUME", "Backend instrument" },
    { KW_TELESCOP, CHARTYPE, "TELESCOP", "Telescope" },
    { KW_LINE,     CHARTYPE, "LINE",     "Spectral line" },
    { KW_SCAN,     LONGTYPE, "SCAN-NUM", "Sequence number" },
    { KW_UTC,      CHARTYPE, "UTC",      "Universal time at start of observation" },
    { KW_LST,      CHARTYPE, "LST",      "Sidereal time at start of observation"},
    { KW_DATE_OBS, CHARTYPE, "DATE-OBS", "Date of data acquisition (dd/mm/yy)" },
    { KW_DATE,     CHARTYPE, "DATE",     "Date of this file (dd/mm/yy)" },
    { KW_TIME,     CHARTYPE, "TIME",     "Time of this file (hh:mm:ss)" },
    { KW_JDATE,    REALTYPE, "JDATE",    "Julian date at start of observation" },
    { KW_TSYS,     REALTYPE, "TSYS",     "System temperature (K)" },
    { KW_TLOAD,    REALTYPE, "TLOAD",    "Load temperature (K)" },
    { KW_TREC,     REALTYPE, "TREC",     "Receiver temperature (K)" },
    { KW_TAU,      REALTYPE, "TAU-ATM",  "Zenith optical depth of atm. (neper)" },
    { KW_DBLOAD,   REALTYPE, "DBLOAD",   "Load attenuation (dB)" },
    { KW_EQUINOX,  REALTYPE, "EQUINOX",  "Equinox of coordinates" },
    { KW_RA,       REALTYPE, "RA",       "Apparent right ascension (deg)" },
    { KW_DEC,      REALTYPE, "DEC",      "Apparent declination (deg)" },
    { KW_AZIMUTH,  REALTYPE, "AZIMUTH",  "Telescope azimuth readout (deg)" },
    { KW_ELEVATIO, REALTYPE, "ELEVATIO", "Telescope elevation readout (deg)" },
    { KW_AZOFF,    REALTYPE, "AZOFF",    "Offset, on sky, in azimuth (deg)" },
    { KW_ELOFF,    REALTYPE, "ELOFF",    "Offset, on sky, in elevation (deg)" },
    { KW_AZPOINT,  REALTYPE, "AZPOINT",  "Azimuth pointing offset (deg)" },
    { KW_ELPOINT,  REALTYPE, "ELPOINT",  "Elevation pointing offset (deg)" },
    { KW_AZCORR,   REALTYPE, "AZCORR",   "Azimuth pointing correction (deg)" },
    { KW_ELCORR,   REALTYPE, "ELCORR",   "Elevation pointing correction (deg)" },
    { KW_COLLIMAT, REALTYPE, "COLLIMAT", "???" },
    { KW_RESTFREQ, REALTYPE, "RESTFREQ", "Rest frequency (Hz)" },
    { KW_FREQRES,  REALTYPE, "FREQRES",  "Frequency resolution (Hz)" },
    { KW_VLSR,     REALTYPE, "VLSR",     "Local standard of rest velocity (m/s)" },
    { KW_VHEL,     REALTYPE, "VHEL",     "Heliocentric velocity (m/s)" },
    { KW_VELO_LSR, REALTYPE, "VELO-LSR", "LSR velocity (m/s)" },
    { KW_VELO_HEL, REALTYPE, "VELO-HEL", "Heliocentric velocity (m/s)" },
    { KW_VELO_GEO, REALTYPE, "VELO-GEO", "Geocentric velocity (m/s)" },
    { KW_VELOCITY, REALTYPE, "VELOCITY", "Local standard of rest velocity (m/s)" },
    { KW_DELTAV,   REALTYPE, "DELTAV",   "Velocity increment (m/s)" },
    { KW_DELTAVEL, REALTYPE, "DELTAVEL", "Velocity increment (m/s)" },
    { KW_OBSTIME,  REALTYPE, "OBSTIME",  "Length of observation (s)" },
    { KW_INTTIME,  REALTYPE, "INTTIME",  "Length of observation (s)" },
    { KW_TAMB,     REALTYPE, "TAMB",     "Ambient temperature (K)" },
    { KW_TOUTSIDE, REALTYPE, "TOUTSIDE", "Outside temperature (K)" },
    { KW_PRESSURE, REALTYPE, "PRESSURE", "Pressure (mB)" },
    { KW_HUMIDITY, REALTYPE, "HUMIDITY", "Relative humidity (%)" },
    { KW_COMMENT,  NVARTYPE, "COMMENT",  "Comment field" },
    { KW_HISTORY,  NVARTYPE, "HISTORY",  "History field" },
    { KW_EXTEND,   BOOLTYPE, "EXTEND",   "" },
    { KW_PCOUNT,   LONGTYPE, "PCOUNT",   "" },
    { KW_GCOUNT,   LONGTYPE, "GCOUNT",   "" },
    { KW_TFIELDS,  LONGTYPE, "TFIELDS",  "" },
    { KW_EXTNAME,  CHARTYPE, "EXTNAME",  "" },
    { KW_EXTVER,   LONGTYPE, "EXTVER",   "" },
    { KW_NMATRIX,  LONGTYPE, "NMATRIX",  "" },
    { KW_MAXIS,    LONGTYPE, "MAXIS",    "" },
    { KW_MAXIS1,   LONGTYPE, "MAXIS1",   "Number of values along first axis" },
    { KW_TFORM,    CHARTYPE, "TFORM",    "" },
    { KW_TTYPE,    CHARTYPE, "TTYPE",    "" },
    { KW_TUNIT,    CHARTYPE, "TUNIT",    "" },
    { KW_TDIM,     CHARTYPE, "TDIM",     "" },
    { KW_END,      NVARTYPE, "END",      "End of header" }
};
#else
extern struct knownkey known[];
#endif

typedef union {
  char str[19];
  long l;
  double d;
  double c[2];
} DATA;

typedef struct {
    int key;
    int type;
    DATA val;
} FITS, *FITSptr;
