#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define PRF_EXTERNAL
#include "prf.h"
#include "init.h"

void init(char *molecule)
{
    void read_prf_file(), read_inp_file(), report_defines();
    
    strcpy(prf.Molecule, molecule);
  
    strcpy(logs.file, molecule);
    strcat(logs.file, LOG_SUFFIX);
  
    logs.fp = fopen(logs.file, "w");
  
    if (!logs.fp) {
        fprintf(stderr, "Fatal error: Cannot open '%s'\n", logs.file);
        exit(1);
    }

    read_prf_file();
    read_inp_file(molecule);
  
    report_defines();
}

void read_inp_file(char *molecule)
{
    int n=0, i, iv[MAX_POSITION];
    char str[LEN_ROW], InputFile[LEN_FILENAME];
    double d[MAX_POSITION];
    double r1, r2, dr;
    FILE *ifp;
  
    int is_comment();
    void get_dvector(), get_ivector(), make_string();
  
    strcpy(InputFile, molecule);
    strcat(InputFile, INP_SUFFIX);
  
    ifp = fopen(InputFile, "r");
  
    if (!ifp) {
        fprintf(stderr, "Fatal error: Cannot open file '%s'\n", InputFile);
        exit(1);
    }
  
    while (fgets(str, LEN_ROW, ifp) != NULL) {
        if (is_comment(str)) continue;
        n++;
        if (n==1) {
            make_string(prf.Source, str);
        } else if (n==2) {
            sscanf(str, "%lf %lf", &prf.Vlsr, &prf.Distance);
        } else if (n==3) {
            i = sscanf(str, "%d %d %d %d", &prf.Npos, &prf.Nrad, &prf.Nvel, &prf.Nstep);
            if (i == 3) prf.Nstep = NSTEPS;
            if (prf.Nvel > MAX_VEL) {
                fprintf(stderr, "Fatal error: Nvel too large. Increase MAX_VEL!\n");
                exit(1);
            }
            if (prf.Nrad < 0) {
                prf.Nrad = -prf.Nrad;
                prf.IntMethod = 1;
            }
            if (prf.Nrad % 2 == 1) {
                prf.Nrad += 1;
                fprintf(stderr, "Warning: Nrad odd. Increasing to %d!\n", prf.Nrad);
            }
            if (prf.Npos > MAX_POSITION) {
                fprintf(stderr, "Fatal error: Npos too large. Increase MAX_POSITION!\n");
                exit(1);
            }
        } else if (n==4) {
            if (sscanf(str, "F %lf %lf %lf", &r1, &r2, &dr) == 3) {
                if (dr == 0.0) {
                    fprintf(stderr, "Fatal error: dr=0!\n");
                    exit(1);
                }
                prf.Npos = (r2-r1)/dr + 1;
                if (prf.Npos > MAX_POSITION) {
                    fprintf(stderr, "Fatal error: Npos too large. Increase MAX_POSITION!\n");
                    exit(1);
                }
                for (i=0; i<prf.Npos; i++) pos[i].x = r1 + (double)i*dr;
            } else if (str[0] == 'S') { /* pos in each shell */
                prf.Npos = prf.Nvol - 1;
                if (prf.Npos > MAX_POSITION) {
                    fprintf(stderr, "Fatal error: Npos too large. Increase MAX_POSITION!\n");
                    exit(1);
                }
                for (i=1; i<prf.Nvol; i++) {
                    pos[i].x = (gSph.Rad[i]+gSph.Rad[i-1])/2.0;
                    pos[i].y = 0.0;
                }
            } else {
                get_dvector(str, d, MAX_POSITION);
                for (i=0; i<prf.Npos; i++) pos[i].x = d[i];
            }
        } else if (n==5) {
            get_dvector(str, d, MAX_POSITION);
            for (i=0; i<prf.Npos; i++) pos[i].y = d[i];
        } else if (n==6) {
            get_dvector(str, d, MAX_POSITION);
            for (i=0; i<prf.Npos; i++) pos[i].beam = d[i];
        } else if (n==7) {
            get_ivector(str, iv, MAX_POSITION);
            for (i=0; i<prf.Npos; i++) {
                pos[i].trans = iv[i];
                if (iv[i] > MAX_RTRANS) {
                    fprintf(stderr, "Fatal error: Selected transition is out of bound.\n");
                    exit(1);
                }
            }
        } else {
            fprintf(stderr, "Skipping line '%s' in file '%s'\n", str, InputFile);
        }
    }
  
    fclose(ifp);
}

void read_prf_file()
{
    int n, l, err=0;
    char prffile[LEN_FILENAME], str[LEN_ROW], Info[LEN_ROW];
    double d[MAX_VOLUME];
    FILE *fp;
  
    void get_dvector();
  
    strcpy(prffile, MCP_DIR);
    strcat(prffile, prf.Molecule);
    strcat(prffile, PRF_SUFFIX);
  
    fp = fopen(prffile, "r");
  
    if (!fp) {
        fprintf(stderr, "Fatal error: Cannot open '%s'\n", prffile);
        exit(1);
    } else {
        printf("Opened file '%s'.\n", prffile);
    }
  
    if (fgets(str, LEN_ROW, fp) != NULL) {
        strcpy(Info, str);
    }
    if (fgets(str, LEN_ROW, fp) != NULL) {
        sscanf(str, "%d %d %d %lf %lf",
               &prf.Ntrans, &prf.Nlevel, &prf.Nvol, &prf.passBand, &gSph.Rmax);
        if (prf.Ntrans > MAX_RTRANS) {
            fprintf(stderr, "Fatal error: Ntrans too large. Increase MAX_RTRANS!\n");
            err = 1;
        }
        if (prf.Nlevel > MAX_LEVEL) {
            fprintf(stderr, "Fatal error: Nlevel too large. Increase MAX_LEVEL!\n");
            err += 2;
        }
        if (prf.Nvol > MAX_VOLUME) {
            fprintf(stderr, "Fatal error: Nvol too large. Increase MAX_VOLUME!\n");
            err += 4;
        }
        if (err) exit(1);
    }
    if (fgets(str, LEN_ROW, fp) != NULL) {
        sscanf(str, "%lf %lf",
               &prf.Tbg, &gSph.Tstar);
    }
    if (fgets(str, LEN_ROW, fp) != NULL) {
        sscanf(str, "%lf %lf %lf",
               &gSph.Vrad0, &gSph.Vrad1, &gSph.Vrad2);
    }
    if (fgets(str, LEN_ROW, fp) != NULL) {
        get_dvector(str, &gSph.Rad[0], MAX_VOLUME);
    }
    if (fgets(str, LEN_ROW, fp) != NULL) {
        get_dvector(str, d, MAX_VOLUME);
        for (n=0; n<prf.Nvol; n++) vol[n].Vdop = d[n];
    }
    for (n=0; n<prf.Ntrans; n++) {
        if (fgets(str, LEN_ROW, fp) != NULL) {
            sscanf(str, "%d %d %lf %lf %lf %lf",  &tra[n].l, &tra[n].u,
                   &tra[n].aconst, &tra[n].st, &tra[n].Gul, &tra[n].Fre);
        }
    }
    if (fgets(str, LEN_ROW, fp) != NULL) {
        sscanf(str, "%d", &prf.overlap);
    }
    if (prf.overlap) {
        for (n=0; n<prf.Ntrans; n++) {
            if (fgets(str, LEN_ROW, fp) != NULL) {
                sscanf(str, "%d %lf %lf", &tra[n].n, &tra[n].v_min, &tra[n].v_max);
                if (tra[n].n == 0) continue;
                for (l=0; l<tra[n].n; l++) {
                    if (fgets(str, LEN_ROW, fp) != NULL) {
                        sscanf(str, "%d %lf", &tra[n].m[l], &tra[n].dv[l]);
                    }
                }
            }
        }
    }
    if (fgets(str, LEN_ROW, fp) != NULL) {
        sscanf(str, "%lf", &prf.dust);
    }
    if (prf.dust) {
        if (fgets(str, LEN_ROW, fp) != NULL) {
            get_dvector(str, d, MAX_VOLUME);
            for (n=0; n<prf.Nvol; n++) vol[n].Tdust = d[n];
        }
        for (n=0; n<prf.Ntrans; n++) {
            if (fgets(str, LEN_ROW, fp) != NULL) {
                get_dvector(str, d, MAX_VOLUME);
                for (l=0; l<prf.Nvol; l++) tra[n].dconst[l] = d[l];
            }
        }
    }

    for (l=0; l<prf.Nlevel; l++) {
        if (fgets(str, LEN_ROW, fp) != NULL) {
            get_dvector(str, d, MAX_VOLUME);
            for (n=0; n<prf.Nvol; n++) vol[n].pop[l] = d[n];
        }
    }
  
    fclose(fp);
}

int is_comment(char *str)
{
    int i;
    char c;
  
    if ((int)strlen(str) == 0) return 1;
  
    i = 0;
    c = str[0];
  
    while (i < (int)strlen(str)-1 && (c == ' ' || c == '\t') ) {
        i++;
        c = str[i];
    }
  
    if (i >= (int)strlen(str)) return 1;  /* Blank line == comment */
  
    if (c == '%' || c == '#' || c == '*' || c == '\0') return 1;
  
    return 0;
}

void get_dvector(char *str, double *d, int size)
{
    int i=0, n;
    char *delims = " ";
    char *token;
  
    token = strtok(str, delims);
  
    while (token != NULL && i < size) {
        d[i] = atof(token);
        token = strtok(NULL, delims);
        i++;
    }
    if (i > 0) {
        for (n=i; n<size; n++) d[n] = d[n-1];
    }
}

void get_ivector(char *str, int *d, int size)
{
    int i=0, n;
    char *delims = " ";
    char *token;
  
    token = strtok(str, delims);
  
    while (token != NULL && i < size) {
        d[i] = atoi(token);
        token = strtok(NULL, delims);
        i++;
    }
    if (i > 0) {
        for (n=i; n<size; n++) d[n] = d[n-1];
    }
}

void make_string(char *s1, char *s2)
{
    int i;
  
    strcpy(s1, s2);
  
    for (i=0; i<(int)strlen(s1); i++) {
        if (s1[i] == '\n') {
            s1[i] = '\0';
            return;
        }
    }
}
