#include <stdio.h>
#include <math.h>

#ifndef PRF_EXTERNAL
#define PRF_EXTERNAL
#endif
#include "prf.h"

#define FANCY_INTEGRATION

double DeltaI[MAX_VEL], Tbri[MAX_VEL], Trad[MAX_VEL];
double S[MAX_VEL], SourceFlux[MAX_VEL];
double Tant[MAX_VEL], Vel[MAX_VEL], Tau0[MAX_VEL];
static double Tau[MAX_VEL], Ibg[MAX_VEL], Istar[MAX_VEL];

void integrate(int k)
{
  int n, i, j, star=0;
  double r, rs, w, w0, w1, p, Itot, Itot1, f1, f2;
  double r1, R1, x, y, b, beam, maxT;
  double *I = NULL;
  double p1, p2;
  POS *P = &pos[k];
  TRANS *T = &tra[P->trans];
  
  void find_bndr();
  void add_con(TRANS *, double *);
  double Bessel_I0(), expBessel_I0();
  
  if (prf.Distance == 0.0) prf.Distance = 1000.0;
      
  r1 = gSph.Rmax/pow(10.0, LOG_PC_CNST)/prf.Distance;
  R1 = 3600.0*RADTODEG*gSph.Rmax/pow(10.0, LOG_PC_CNST)/prf.Distance;
  x = R1 * P->x;
  y = R1 * P->y;
  if (pos[0].beam < 0.0 && pos[1].beam > 0.0) {
    b = -pos[0].beam * (pos[1].beam/T->Fre);
    beam = b/R1;
  } else if (P->beam < 0.0) {
    b = -P->beam;
    beam = b/R1;
  } else {
    b = R1 * P->beam;
    beam = P->beam;
  }

  printf("Using: f=%f GHz (%d)\n", T->Fre, P->trans);
  printf("       Rmax=%9.4e cm corresponds to %.2f\" at D=%.2f pc.\n",
         gSph.Rmax, R1, prf.Distance);
  printf("       x=%f (%.2f\")  y=%f (%.2f\")\n", P->x, x, P->y, y);
  printf("       Beam=%f (%.2f\")\n", beam, b);
  
  if (gSph.Tstar > 0.0 && gSph.Rad[0] > 0.0) star=1;
  
  if (!prf.overlap) {
      T->v_min = -prf.passBand/2.0;
      T->v_max = prf.passBand/2.0;
  }

  prf.DeltaV = (T->v_max - T->v_min)/(double)(prf.Nvel - 1);
  prf.DeltaF = -prf.DeltaV*1.0e3/CLIGHT * T->Fre;
  prf.RefChan = 1.0 - T->v_min/prf.DeltaV;
  prf.RestF = T->Fre;
  
  f1 = prf.RestF - (prf.RefChan - 1.0)*prf.DeltaF;
  f2 = f1 + (prf.Nvel - 1.0)*prf.DeltaF;
  printf("	 v_min=%f  v_max=%f  dv=%f\n",
  	 T->v_min, T->v_max, prf.DeltaV);
  printf("	 f_max=%f  f_0=%f  f_min=%f  df=%f\n",
  	 f1, prf.RestF, f2, prf.DeltaF);
  
  for (i=0; i<prf.Nvel; i++) {
    Tant[i] = 0.0;
    SourceFlux[i] = 0.0;
    Ibg[i] = 0.0;
    Istar[i] = 0.0;
    if (prf.Tbg > 0.0)
      Ibg[i] = T->st/(exp(PLOBO_CNST*T->Fre/prf.Tbg) - 1.0);
    if (gSph.Tstar > 0.0)
      Istar[i] = T->st/(exp(PLOBO_CNST*T->Fre/gSph.Tstar) - 1.0);
    Vel[i] = T->v_min + (double)i * prf.DeltaV;
  }
  
  prf.pDir.x = 0.0;
  prf.pDir.y = 0.0;
  prf.pDir.z = 1.0;
  
  p = sqrt(P->x*P->x + P->y*P->y);
  w0 = 2.0 * ALPHA * exp(-ALPHA*p*p/beam/beam) / beam / beam;
  w1 = 2.0 * PI * r1 * r1;
      /* From W m^-2 Hz^-1 to W cm^-2 (km/s)^-1 */
  w1 *= T->Fre * 1.0e9 / 1.0e4 /(CLIGHT/1.0e3);
  
if (prf.IntMethod == 0) {
  w0 = 2.0 * ALPHA / beam / beam;
  p1 = p - 7.33*beam;
  if (p1 < 0.0) p1 = 0.0;
  p2 = p + 7.33*beam;
  if (p2 > 1.0) p2 = 1.0;
  rs = (p2-p1)/(double)prf.Nrad;
  printf("Beam integration r1,r2 = %f,%f\n", p1, p2);
  for (n=0; n<=prf.Nrad; n++) {
    r = p1 + (double)n * rs;
    if (prf.Nrad < 25) {
        printf(" %d", n);
        fflush(stdout);
    } else if (prf.Nrad < 250) {
        if (n % 10 == 0) {
	    printf(" %d", n);
            fflush(stdout);
	}
    } else if (prf.Nrad < 2500) {
        if (n % 100 == 0) {
	    printf(" %d", n);
            fflush(stdout);
	}
    } else {
        if (n % 1000 == 0) {
	    printf(" %d", n);
            fflush(stdout);
	}
    }
    I = Ibg;
    if (star) {
      prf.pOrg.x = r;
      prf.pOrg.y = 0.0;
      if (r < gSph.Rad[0]) { /* Starting at star */
        prf.pOrg.z = sqrt(gSph.Rad[0]*gSph.Rad[0] - r*r);
        I = Istar;
      } else { /* Starting at cloud edge */
        prf.pOrg.z = -0.9999999 * sqrt(1.0 - r*r);
      }
    } else {
      prf.pOrg.x = r;
      prf.pOrg.y = 0.0;
      prf.pOrg.z = -0.9999999 * sqrt(1.0 - r*r);
    }
    find_bndr();
    if (prf.Nbnd > 0) {
      add_con(T, I);
      w = expBessel_I0(2.0*ALPHA*r*p/beam/beam);
      if (n == 0 || n == prf.Nrad) {
        w *= 1.0 / 3.0;
      } else {
        w *= 2.0 / 3.0 * (double)(1 + (n % 2));
      }
      w *=  w0 * exp(-ALPHA*(r-p)*(r-p)/beam/beam) * r * rs;
      
      maxT = DeltaI[0];
      for (i=0; i<prf.Nvel; i++) {
        Tant[i] += w * DeltaI[i];
	if (DeltaI[i] > maxT) maxT = DeltaI[i];
      }
 /*     printf("r,Tmax= %.4f, %.6f %.4f\n", r, rs, maxT); */
      
    /* Source flux */
      if (p1 == 0.0 && p2 == 1.0) {
        w = w1 * r * rs;
        if (n == 0 || n == prf.Nrad) {
          w *= 1.0 / 3.0;
        } else {
          w *= 2.0 / 3.0 * (double)(1 + (n % 2));
        }
        for (i=0; i<prf.Nvel; i++) SourceFlux[i] += w * S[i];
      }
    }
  }
} else if (prf.IntMethod == 1) {
#ifdef FANCY_INTEGRATION
  for (j=0; j<prf.Nvol; j++) {
  printf(" %d", j);
  fflush(stdout);
  if (j == 0) {
    rs = gSph.Rad[0]/(double)prf.Nrad;
  } else {
    rs = (gSph.Rad[j] - gSph.Rad[j-1])/(double)prf.Nrad;
  }
#else
  rs = 1.0/(double)prf.Nrad;
#endif
  
  for (n=0; n<=prf.Nrad; n++) {
    r = (double)n * rs;
#ifdef FANCY_INTEGRATION
    if (j > 0) r += gSph.Rad[j-1];
#else
    printf(" %d", n);
    fflush(stdout);
#endif
    I = Ibg;
    if (star) {
      prf.pOrg.x = r;
      prf.pOrg.y = 0.0;
      if (r < gSph.Rad[0]) { /* Starting at star */
        prf.pOrg.z = sqrt(gSph.Rad[0]*gSph.Rad[0] - r*r);
        I = Istar;
      } else { /* Starting at cloud edge */
        prf.pOrg.z = -0.9999999 * sqrt(1.0 - r*r);
      }
    } else {
      prf.pOrg.x = r;
      prf.pOrg.y = 0.0;
      prf.pOrg.z = -0.9999999 * sqrt(1.0 - r*r);
    }
    find_bndr();
    if (prf.Nbnd > 0) {
      add_con(T, I);
      if (p == 0.0) {
        w = 1.0;
      } else {
        w = Bessel_I0(2.0*ALPHA*r*p/beam/beam);
      }
      if (n == 0 || n == prf.Nrad) {
        w *= 1.0 / 3.0;
      } else {
        w *= 2.0 / 3.0 * (double)(1 + (n % 2));
      }
      w *=  w0 * exp(-ALPHA*r*r/beam/beam) * r * rs;
      
      maxT = DeltaI[0];
      for (i=0; i<prf.Nvel; i++) {
        Tant[i] += w * DeltaI[i];
	if (DeltaI[i] > maxT) maxT = DeltaI[i];
      }
 /*     printf("r,Tmax= %.4f, %.6f %.4f\n", r, rs, maxT); */
      
    /* Source flux */
      w = w1 * r * rs;
      if (n == 0 || n == prf.Nrad) {
        w *= 1.0 / 3.0;
      } else {
        w *= 2.0 / 3.0 * (double)(1 + (n % 2));
      }
      for (i=0; i<prf.Nvel; i++) SourceFlux[i] += w * S[i];
    }
  }
#ifdef FANCY_INTEGRATION
  }
#endif
}
  printf("\n");
  
  Itot = 0.0;
  Itot1 = 0.0;
  I = Ibg;
  if (star) {
    prf.pOrg.x = p;
    prf.pOrg.y = 0.0;
    if (p < gSph.Rad[0]) { /* Starting at star */
      prf.pOrg.z = sqrt(gSph.Rad[0]*gSph.Rad[0] - p*p);
      I = Istar;
    } else { /* Starting at cloud edge */
      prf.pOrg.z = -0.9999999 * sqrt(1.0 - p*p);
    }
  } else {
    prf.pOrg.x = p;
    prf.pOrg.y = 0.0;
    prf.pOrg.z = -0.9999999 * sqrt(1.0 - p*p);
  }
  find_bndr();
  if (prf.Nbnd > 0) add_con(T, I);
  for (i=0; i<prf.Nvel; i++) {
    if (prf.Nbnd > 0) {
      Tau0[i] = Tau[i];
    } else {
      Tau0[i] = 0.0;
    }
    Itot  += Tant[i] * prf.DeltaV;
    Itot1 += (Tant[i] - (Tant[0] + Tant[prf.Nvel-1])/2.0) * prf.DeltaV;
  }
  P->Itot = Itot;
  printf("Itot= %lf (%lf) K km/s\n", Itot, Itot1);
}

void add_con(TRANS *T, double *I)
{
  int u, l, n, k, i, j, o, ll, uu;
  int olap=0, dust=0, nstep=prf.Nstep;
  double t1, t2, t, ts, arg, vz;
  double dpt, dTau, Snu, Tcnst, Snu_sum, R, oSnu, tau_o, snu0, tau0;
  double d_dTau, d_Snu;
  double o_dpt[MAX_OLAP], o_dTau[MAX_OLAP], o_Snu[MAX_OLAP], o_Tcnst[MAX_OLAP];
  VOLUME *v;
  TRANS *TT;
  
  double VelCos();
  
  l = T->l; u = T->u;
  Tcnst = T->aconst * gSph.Rmax;
  
  if (prf.overlap && T->n > 0) olap = 1;
  if (prf.dust > 0.0) dust = 1;
  
  if (olap) {
    for (o=0; o<T->n; o++)
      o_Tcnst[o] = tra[T->m[o]].aconst * gSph.Rmax;
  }
  
  for (j=0; j<prf.Nvel; j++) {
    DeltaI[j] = I[j];
    Tau[j] = 0.0;
  }
  
  for (k=0; k<prf.Nbnd; k++) {
    n = bnd[k].n;
    if (n <= 0) continue;
    
    if (k == 0) {
      t1 = 0.0;
    } else {
      t1 = bnd[k-1].t;
    }
    t2 = bnd[k].t;
    if (t1 == t2) continue;
    
    ts = (t2 - t1)/(double)(nstep);
    v = &vol[n];
    dpt = Tcnst * (v->pop[l]*T->Gul - v->pop[u])/v->Vdop * ts;
    Snu = T->st/(v->pop[l]/v->pop[u]*T->Gul - 1.0);
    if (dust) {
        d_dTau = T->dconst[n] * ts;
        d_Snu = 0.0;
        if (v->Tdust != 0.0)
            d_Snu = T->st / (exp(PLOBO_CNST*T->Fre/v->Tdust) - 1.0);
    }
    if (olap) {
      for (o=0; o<T->n; o++) {
        TT = &tra[T->m[o]];
        ll = TT->l; uu = TT->u;
        o_dpt[o] = o_Tcnst[o] * (v->pop[ll]*TT->Gul - v->pop[uu])/v->Vdop * ts;
        o_Snu[o] = TT->st/(v->pop[ll]/v->pop[uu]*TT->Gul - 1.0);
        /* if (k == 25) {
            if (o == 0) printf("\n%f depth=%e Snu=%e\n",
                               T->Fre, dpt, Snu);
            printf("%d %f depth=%e Snu=%e\n",
                   T->m[o], TT->Fre, o_dpt[o], o_Snu[o]);
        } */
      }
    }
    for (i=0; i<nstep; i++) {
      t = t1 + (double)i * ts;
      vz = (VelCos(t) + 2.0*VelCos(t + ts/2.0) + VelCos(t + ts))/4.0;
      for (j=0; j<prf.Nvel; j++) {
        arg = (Vel[j] + vz)/v->Vdop;
        tau0 = dTau = dpt * exp(-arg * arg);
        if (olap) {
          snu0 = Snu_sum = Snu * dTau;
          if (dust) {
              Snu_sum += d_Snu * d_dTau;
              dTau += d_dTau;
          }
          oSnu = tau_o = 0.0;
          for (o=0; o<T->n; o++) {
            arg = (Vel[j] + vz - T->dv[o])/v->Vdop;
            o_dTau[o] = o_dpt[o] * exp(-arg * arg);
            oSnu += o_Snu[o] * o_dTau[o];
            Snu_sum += o_Snu[o] * o_dTau[o];
            tau_o += o_dTau[o];
            dTau += o_dTau[o];
          }
          if (fabs(dTau) < 1.0e-5) {
            R = 1.0;
            /* DeltaI[j] += (Snu_sum - dTau * DeltaI[j]); */
          } else if (dTau > 20.0) {
            R = 1.0/dTau;
            /* DeltaI[j] += (Snu_sum - dTau * DeltaI[j])/dTau; */
          } else {
            R = (1.0 - exp(-dTau))/dTau;
            /* DeltaI[j] += (Snu_sum - dTau * DeltaI[j])*(1.0 - exp(-dTau))/dTau; */
          }
          DeltaI[j] += (Snu_sum - DeltaI[j]*dTau) * R;
          /* DeltaI[j] = DeltaI[j]*exp(-dTau) + Snu_sum * R; */
          /* arg = PLOBO_CNST*T->Fre/T->st*(DeltaI[j] - I[j]);
          R = PLOBO_CNST*T->Fre/T->st * Snu_sum * R;
          if (arg < -0.3 && n == 42) {
              printf("T[%d,%d]=%e Snu=%e dSnu=%e oSnu=%e Tau=%e TauD=%e TauO=%e\n",
                     j, n, R, Snu_sum, snu0, oSnu, dTau, tau0, tau_o);
              for (o=0; o<T->n; o++) {
                  R = PLOBO_CNST*T->Fre/T->st * o_Snu[o];
                  arg = PLOBO_CNST*T->Fre/T->st * I[j];
                  printf("%d %e tau=%e  Snu=%e  Ibg=%e\n",
                         o, (Vel[j] + vz - T->dv[o])/v->Vdop, o_dTau[o], R, arg);
              }
          } */
          /* if (dTau != 0.0) {
            DeltaI[j] += (Snu_sum/dTau - DeltaI[j]) * (1.0 - exp(-dTau));
          } */
        } else {
          if (dust) {
            Snu_sum = Snu * dTau + d_Snu * d_dTau;
            dTau += d_dTau;
            if (fabs(dTau) < 1.0e-5) {
              R = 1.0;
            } else if (dTau > 20.0) {
              R = 1.0/dTau;
            } else {
              R = (1.0 - exp(-dTau))/dTau;
            }
            DeltaI[j] += (Snu_sum - DeltaI[j]*dTau) * R;
          } else
            DeltaI[j] += (Snu - DeltaI[j]) * (1.0 - exp(-dTau));
        }
        Tau[j] += dTau;
      }
    }
  }
  
  for (j=0; j<prf.Nvel; j++) {
/* Flux */ 
    S[j] = DeltaI[j];
/* The radiation temperature using the Planck radiation curve */
    if (DeltaI[j] != 0.0)
      Trad[j] = PLOBO_CNST*T->Fre/log(T->st/DeltaI[j] + 1.0);
    else
      Trad[j] = 0.0;
/* The radiation temperature using the Rayleigh-Jeans approximation */
    Tbri[j] = PLOBO_CNST*T->Fre/T->st*DeltaI[j];
/* The antenna temperature (cosmic background subtracted) */
    DeltaI[j] = PLOBO_CNST*T->Fre/T->st*(DeltaI[j] - Ibg[j]);
  }
}

static double p_coeffs[7] = {
      1.0000000e0, 3.5156229e0, 3.0899424e0, 1.2067492e0,
      0.2659732e0, 0.0360768e0, 0.0045813e0};
static double r_coeffs[9] = {
     0.39894228e0, 0.01328592e0, 0.00225319e0, -0.00157565e0,
     0.00916281e0,-0.02057706e0, 0.02635537e0, -0.01647633e0,
     0.00392377e0};

double Bessel_I0(double x)
{
  int i;
  double a, t, i0;

  a = fabs(x);

  if (a == 0.0) {
    return 1.0;
  } else if (a <= 3.75) {
    t = a * a / 3.75 / 3.75;
    i0 = 1.0;
    for (i=1; i<7; i++) {
      i0 += p_coeffs[i] * pow(t, (double)(i));
    }
  } else {
    t = 3.75/a;
    i0 = r_coeffs[0];
    for (i=1; i<9; i++) {
      i0 += r_coeffs[i] * pow(t, (double)(i));
    }
    i0 = i0 * exp(a) / sqrt(a);
  }

  return i0;
}

double expBessel_I0(double x)
{
  int i;
  double a, t, i0;

  a = fabs(x);

  if (a == 0.0) {
    return 1.0;
  } else if (a <= 3.75) {
    t = a * a / 3.75 / 3.75;
    i0 = 1.0;
    for (i=1; i<7; i++) {
      i0 += p_coeffs[i] * pow(t, (double)(i));
    }
    i0 *= exp(-a);
  } else {
    t = 3.75/a;
    i0 = r_coeffs[0];
    for (i=1; i<9; i++) {
      i0 += r_coeffs[i] * pow(t, (double)(i));
    }
    i0 /= sqrt(a);
  }

  return i0;
}
