#define MAX_RTRANS		8300
#define MAX_LEVEL		530
#define MAX_VOLUME		199
#define MAX_POSITION	300
#define MAX_BOUND	(2*MAX_VOLUME)
#define MAX_OLAP		40

#define MAX_VEL			401

#define NSTEPS			32

#define CLIGHT          2.99792458e8    /* c [m/s]                     */
#define TAU_CNST        0.01072068      /* c[km/s] c[cm/s]^2 / 8 pi 10^27    */
#define TAUD_CNST  299792.458
#define PLOBO_CNST      0.04799216      /* h/k                          */
#define PLANCK_CNST     6.626076e-18    /* h 10^9                       */
#define INT_CNST        1.4745e-23      /* 2 h 10^27 / c^2              */
#define LOG_PC_CNST     18.48935063     /* log(3.08*10^18 cm)           */
#define INVCM_TO_K      1.43877
#define K_TO_GHZ        20.83674
#ifndef PI
#define PI              3.141592654
#endif
#define ALPHA			2.772588722
#define RADTODEG	    (180.0/PI)

#define GEO_SPHERE	0
#define GEO_CUBE	1
#define GEO_CLUMPS	2

#ifndef VECTOR
#include "vector.h"
#endif

typedef struct {
  char file[120], prf[120];
  FILE *fp, *pfp;
} LOGS;

typedef struct {
  int Nshell;
  double Rmax;
  double Rad[MAX_VOLUME];
  double Tstar;
  double Vrad0, Vrad1, Vrad2;
} G_SPHERE, *gSphPtr;

typedef struct {
  int Ncube;
  double Lx, Ly, Lz;
  double Tstar;
} G_CUBE, *gCubPtr;

typedef struct {
  int Nclump;
  double Rclump;
  double Tstar;
} G_CLUMP, *gCluPtr;

typedef struct {
  int Iter;
  char Molecule[80], Source[80];
  int Npos;
  double passBand, Tbg;
  int Nlevel, Ntrans;
  int Nvol;
  int geometry;
  gSphPtr gptr;
  CARTESIAN pOrg, pDir;
  int Nbnd;
  int Nrad, Nvel, Nstep;
  double DeltaV, DeltaF, RestF, RefChan, Vlsr, Distance;
  int overlap;
  double dust;
  int IntMethod;
} PRF;

typedef struct {
  double x, y;
  double beam;
  int trans;
  double Itot;
} POS;

typedef struct {
  int n;
  double t;
} BOUND;

typedef struct {
  int u, l;
  double Gul;
  double Fre;
  double aconst;
  double dconst[MAX_VOLUME];
  double st;
  int n;
  int m[MAX_OLAP];
  double dv[MAX_OLAP];
  double v_min, v_max;
} TRANS;

typedef struct {
  double Vdop;
  double Tdust;
  CARTESIAN Vel;
  CARTESIAN Pos;
  CARTESIAN Siz;
  double pop[MAX_LEVEL];
} VOLUME;

#ifndef PRF_EXTERNAL

PRF 		prf;
LOGS 		logs;

TRANS 		tra[MAX_RTRANS];
VOLUME 		vol[MAX_VOLUME];
G_SPHERE	gSph;
G_CUBE		gCub;
G_CLUMP		gClu;
BOUND		bnd[MAX_BOUND];
POS			pos[MAX_POSITION];

#else

extern PRF 			prf;
extern LOGS			logs;

extern TRANS 		tra[MAX_RTRANS];
extern VOLUME 		vol[MAX_VOLUME];
extern G_SPHERE		gSph;
extern G_CUBE		gCub;
extern G_CLUMP		gClu;
extern BOUND		bnd[MAX_BOUND];
extern POS			pos[MAX_POSITION];

#endif
