#include <math.h>
#include <stdio.h>

#ifndef PRF_EXTERNAL
#define PRF_EXTERNAL
#endif

#include "prf.h"

/* #define I16 */
/* #define WHYA */
/* #define ORION */
/* #define STAR */
/* #define STAR_EXP 1.50 */
/* #define FREEINFALL 0.1 */
/* #define O2_INFALL */

/* include defines here so that mistakes are avoided, see also ali/sphere.c */
void report_defines()
{
#ifdef INFALL
    printf("Warning! #define INFALL in sphere.c is used.\n");
#endif
#ifdef FREEINFALL
    printf("Warning! #define FREEINFALL %.3f in sphere.c is used.\n", FREEINFALL);
#endif
#ifdef I16
    printf("Warning! #define I16 in sphere.c is used.\n");
#endif
#ifdef IRC10216
    printf("Warning! #define IRC10216 %.2f in sphere.c is used.\n", IRC10216);
#endif
#ifdef WHYA
    printf("Warning! #define WHYA in sphere.c is used.\n");
#endif
#ifdef ORION
    printf("Warning! #define ORION in sphere.c is used.\n");
#endif
#ifdef STAR
    printf("Warning! #define STAR in sphere.c is used.\n");
#endif
#ifdef STAR_EXP
    printf("Warning! #define STAR_EXP %.2f in sphere.c is used.\n", STAR_EXP);
#endif
}

double VelRad(double r)
{
    double v=0.0;
  
#ifdef FREEINFALL
    double p = gSph.Vrad2;
    double Mgas = 7.045 * (gSph.Vrad1/1000.0) * pow(gSph.Rmax/1.0e18, 3.0) / (1.0 - p/3.0);
    double Vinf = -0.116/sqrt(gSph.Rmax/1.0e18);
    double rmin = gSph.Rad[0];
    double aux = 0.0;
  
    if (r >= rmin && r > 0.0 && r <= 1.0) {
        aux = gSph.Vrad0 + Mgas*(pow(r,3.0-p)- pow(rmin,3.0-p));
        v = FREEINFALL * Vinf * sqrt(aux) * sqrt(1.0/r-1.0);
    }
#else

#ifdef I16  
    if (r >= 0.25) return 0.5;
  
    v = gSph.Vrad0 + gSph.Vrad1 * pow(r/0.25, gSph.Vrad2);
#else
#ifdef WHYA
    if (r > gSph.Vrad1) v = gSph.Vrad0 * sqrt(1.0 - gSph.Vrad1/r);
    if (v < gSph.Vrad2) v = gSph.Vrad2;
#else
#ifdef ORION
    if (r < 0.233) {
        v = 0.0;
    } else {
        v = gSph.Vrad1*r;
    }
#else
#ifdef STAR
    if (r < gSph.Vrad2)
        v = gSph.Vrad1;
    else
#ifdef STAR_EXP
        v = gSph.Vrad1 + (gSph.Vrad0 - gSph.Vrad1)*pow(1.0 - gSph.Vrad2/r, STAR_EXP);
#else
    v = gSph.Vrad1 + (gSph.Vrad0 - gSph.Vrad1)*(1.0 - gSph.Vrad2/r);
#endif
#else
#ifdef O2_INFALL
    if (r <= gSph.Vrad0) {
        v = gSph.Vrad1 * sqrt(r/gSph.Vrad0);
    } else {
        v = gSph.Vrad1;
    }
#else
    if (gSph.Vrad2 == 0.0) {
        if (gSph.Vrad1 < 0.0) {
            if (r < -gSph.Vrad1)
                v = gSph.Vrad0;
            else
                v = 0.0;
        } else {
            if (r >= gSph.Vrad1)
                v = gSph.Vrad0;
            else
                v = 0.0;
        }
    } else {
        v = gSph.Vrad0 + gSph.Vrad1 * pow(r, gSph.Vrad2);
    }
#endif
#endif
#endif
#endif
#endif  
#endif
  
    return v;
}

double dVelRad(double r)
{
    double v;
  
    if (gSph.Vrad2 == 0.0 || gSph.Vrad1 == 0.0) return 0.0;
  
    if (gSph.Vrad2 == 1.0) return (gSph.Vrad1);

    v = gSph.Vrad2 * gSph.Vrad1 * pow(r, gSph.Vrad2 - 1.0);
  
    return v;
}

double VelCos(double t)
{
    double r, vcos;
    CARTESIAN c;
  
    double C_magn(), C_dot(), VelRad();
    CARTESIAN C_line();
  
    c = C_line(prf.pOrg, prf.pDir, t);
    r = C_magn(c);
  
    if (r == 0.0) return 0.0;
  
    vcos = VelRad(r) * C_dot(prf.pDir, c) / r;
  
    return vcos;
}

double dVelCos(double t)
{
    double r, v2, vrad, vcos, dvcos;
    CARTESIAN c;
  
    double C_magn(), C_dot(), VelRad(), VelCos(), dVelRad();
    CARTESIAN C_line();
  
    c = C_line(prf.pOrg, prf.pDir, t);
    r = C_magn(c);
  
    if (r == 0.0) return 0.0;
  
    v2 = C_dot(prf.pDir, prf.pDir);
    vrad = VelRad(t);
    vcos = VelCos(t);
  
    dvcos = vrad*v2/r - vcos/r + dVelRad(t)*pow(vcos/vrad, 2.0);
  
    return dvcos;
}

void find_bndr()
{
    int n, m;
    double a, b, c, t, r, swap;
  
    double C_dot(), C_magn();
    CARTESIAN C_line();
  
    prf.Nbnd = 0;
  
    a = -C_dot(prf.pOrg, prf.pDir);
    b = -C_dot(prf.pOrg, prf.pOrg);
  
    for (n=0; n<prf.Nvol; n++) {
        r = gSph.Rad[n];
        c = a*a + b + r*r;
        if (c <= 0.0) continue;
        for (m=-1; m<=1; m += 2) {
            t = a + (double)m * sqrt(c);
            if (t < 0.0) continue;
            bnd[prf.Nbnd].t = t;
            prf.Nbnd++;
        }
    }
  
    for (n=0; n<prf.Nbnd-1; n++) {
        for (m=n+1; m<prf.Nbnd; m++) {
            if (bnd[n].t > bnd[m].t) {
                swap = bnd[n].t;
                bnd[n].t = bnd[m].t;
                bnd[m].t = swap;
            }
        }
    }
  
    for (m=0; m<prf.Nbnd; m++) {
        if (m == 0) {
            t = 0.0;
        } else {
            t = (bnd[m].t + bnd[m-1].t)/2.0;
        }
        r = C_magn(C_line(prf.pOrg, prf.pDir, t));
        bnd[m].n = -1;
        for (n=0; n<prf.Nvol; n++) {
            if (r < gSph.Rad[n]) {
                bnd[m].n = n;
                break;
            }
        }
    }
}
